/*
 *  MCMC.h
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef MCMC_h__
#define MCMC_h__

#include "Numerics.h"

#include "GTRModel.h"
#include "ProteinModel.h"
#include "Tree.h"
#include "Options.h"
#include "ObservedData.h"

extern Options options;

class MCMC
{
public:
	MCMC(void);
    
    Numerics *num;
    
    // ****************************************************************************************
    // Initialization
    // ****************************************************************************************
	void InitializeTree( ObservedData& mData, Tree& mTree );
    void InitializeTreeAlternate( Tree& mTree );
    void InitializeMixtureModel( MixtureModel& m_Model,
                                const IntMatrix& m_ChangeMatrix);
    
    // ****************************************************************************************
	// Run MCMC moves
    // ****************************************************************************************
  	void RunMarkovChain(void);
    
    /* void RunMarkovChain(
     Tree& mTree,
     GTRModel& mModel,
     IntVector& m_vecSiteAssignment,
     ObservedData& theData);
     
     int RunMarkovChain(
     Tree& mTree,
     ProteinModel& mModel,
     IntVector& m_vecSiteAssignment,
     ObservedData &theData,
     std::ofstream& posteriorPredictiveOut);
     */
    
    // ****************************************************************************************
	// Likelihood calculations
    // ****************************************************************************************
    double CalculateCompleteDataLogLikelihood(
                                              MixtureModel& mModel);
    
    // ****************************************************************************************
	// Samplers: Gibbs update mechanisms
	// ****************************************************************************************
	int Gibbs_updateAllAncestralAndTransientStates_full( MixtureModel& mModel );
    
	int Gibbs_updateSiteAssignments_MixtureComponents( MixtureModel& mModel );
    
    // ****************************************************************************************
	// Tree: wrappers
    // ****************************************************************************************
 	void tree_countChanges( MixtureModel& mModel );
 	void tree_countChanges( MixtureModel& mModel,  int version );
    
    // Data and model
    vector<Tree *> Trees;
    Tree mTree;
    ProteinModel mModel;
    IntVector m_vecSiteAssignment;
    ObservedData theData;
    
    std::ofstream posteriorPredictiveOut;
    
    // Some variables for the running chain
    int currentTree;
    long int numUpdated;
    int resetAfterConjugatePhase;
    double likelihood;
    
    void setPointers( vector<Tree *> &treeInit,
                     int currentTreeInit,
                     ProteinModel& modelInit,
                     IntVector& siteAssignmentInit,
                     ObservedData &dataInit,
                     string ppOutInit  );
    
private:
    long int updateCount;
    int forceCoevolution;
    
    // ****************************************************************************************
	// Likelihood calculations
    // ****************************************************************************************
	double CalculateCompleteDataLogLikelihood(
                                              MixtureModel& mModel,
                                              int version);
	
    // Completely general - wrapper to CalculateCompleteDataLogLikelihood but uses full Pmats along branches
    double CalculateCompleteDataLogLR_overTree(MixtureModel& mModel1,
                                               MixtureModel& mModel2);
	
    // ****************************************************************************************
	// Partial likelihood evaluations
    // ****************************************************************************************
    double CalculateCompleteDataLogLikelihood_Delta_B1u(
                                                        MixtureModel& mModel,
														int theModel, int x, int y);
    
    
    // ****************************************************************************************
	// Samplers: wrappers
    // ****************************************************************************************
    void MCMC_updateRateMatrix( void );
    void MCMC_updateSiteAssignments( void );
    void MCMC_updateRARmodel( void );
    
    // ****************************************************************************************
	// Samplers: mixed methods
    // ****************************************************************************************
	void tree_updateBranchLengths(
                                  int startNode,
                                  MixtureModel& theModel, 
                                  int version, double variance
                                  );
	
	void proposal_rateMatrixAndFreqParameters(
                                              MixtureModel& m_Model,
                                              const double& m_fWindowsize);				//Update GTR model with a specific window size
    
    // Several of the below use Metropolis updates on rate parameters combined with resampling of parameter assignments
    int UpdateRARDistributionParameters(
										ProteinModel& mModel, 
										double updateFraction );
	
	int UpdateRARDistributionParameters(
										ProteinModel& mModel, 
                                        double updateFraction, int whichModel);
    
	int MC_updateRARDistributionParameters_B1u(
											   ProteinModel& mModel, 
											   double updateFraction, int whichModel);
	
    // Redundant
    /*void tree_updateBranchLengths(
                                  int startNode,
                                  GTRModel& mModel, 
                                  int version, double variance
                                  );
    */
    
    // ****************************************************************************************
	// Samplers: slice updaters
	// ****************************************************************************************
    int slice_updateRateMatrix_genNonRev_B1u( PriorDistribution &prior );
    
    // ****************************************************************************************
	// Samplers: conjugate Gibbs
	// ****************************************************************************************
	int cGibbs_updateRateMatrix_genNonRev_B1u( ProteinModel& mModel );
	
    // ****************************************************************************************
	// Samplers: Gibbs update mechanisms
	// ****************************************************************************************
    int Gibbs_updateAllAncestralAndTransientStates_local(
                                                         int& rannode,
                                                         MixtureModel& theModel,
                                                         DoubleMatrix& stateFreqMatrix);
    
    int Gibbs_updateRateParameterAssignments_RAR_B1u(
                                                     ProteinModel& mModel, 
                                                     double updateFraction, int whichModel);
	
    void Gibbs_updateSiteAssignments_RateAssignments( MixtureModel& mModel );
	
	int Gibbs_updateRateParameterAssignments_RAR_DRA_B1u(
                                                         ProteinModel& mModel, 
                                                         double updateFraction, int whichModel);
    
    // ****************************************************************************************
	// Samplers: Metropolis updates
	// ***************************************************************************************
    int UpdateGTRrates_unidimensional_b1u(
										  GTRModel& mModel, 
										  double updateFraction, int whichModel);
	
    // Needed?
    int MC_updateRateMatrixScaling_nonRAR_B1u( ProteinModel& mModel );
    
    // Check this
    int MC_updateStateFreqs_B1u( ProteinModel& mModel );
    
 	int MC_updateASRV_shape(
							ProteinModel& mModel, 
							double proposalVariance);
    
    // Is 'heat' here for relaxation or power posteriors?
	int MC_updateRateMatrix( ProteinModel& mModel  );
    
    // Is 'heat' here for relaxation or power posteriors?
	int MC_updateRateMatrix_nonRAR( ProteinModel& mModel  );
	
    // GTR duplicates - redundant:
    
    // Redundant
	int UpdateGTRrates_unidimensional(
									  GTRModel& mModel, 
									  double updateFraction, int whichModel);
    
	// Redundant
    int MC_updateASRV_shape(
							GTRModel& mModel, 
							double proposalVariance);
	
    // Redundant
	void MC_updateRateMatrix( GTRModel& mModel );
    
    // ****************************************************************************************
	// Tree: operations
    // ****************************************************************************************
    void setAllProbMatricesOnTree( MixtureModel& mModel );
    void setAllProbMatricesOnTree( MixtureModel& mModel, int version );
    
    // ****************************************************************************************
	// Coevolution functions
    // ****************************************************************************************
    void coevo_DataReplacementSimulation(void);
    void coevo_calculateStatistics(void);
    double _calculateMI( int numStates, DoubleVector& marginal1, DoubleVector& marginal2, DoubleVector &jointDwell, double treeLength);
    
    // ****************************************************************************************
	// Utility functions
    // ****************************************************************************************
	void _util_initializeChain(void);
    void _util_initializeRARmodelFromChain(void);
    int  _util_sampleFromCumulative( double *theVector, int length );
    int  _util_sampleFromCumulative_raw( DoubleVector *inVector, int length );
    int  _util_sampleFromCumulative_raw( DoubleVector *inVector, int excludeState, int length );
        

    int  _util_sampleChangeFromCumulative_raw( DoubleVector *inVector, int length, int ignoreState );
    void _util_rescaleForBranchLengths(void);
    void _util_optimizeBranchLengths_capNodes(void);
};



#endif // MCMC_h__
                                                                  
