//
//  NewickTreeParser.cpp
/*
 *  Originally created for David Pollock's CLAMP/PLEX program
 *  Copyright (c) 2004 Dr. David Pollock. All rights reserved.
 
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "NewickTreeParser.h"
//#include "TreeParser.cpp"

#pragma mark -
#pragma mark Delimiters
//---------------------------------------------------------------------------
const char*	NewickTreeParser::cDelimiters					= "(,);";
const char*	NewickTreeParser::cQuotes						= "'";
const char*	NewickTreeParser::cTreeTerminator				= ";";
const char*	NewickTreeParser::cNodeSeparator				= ",";
const char*	NewickTreeParser::cOpenParentNodeDelimiter	= "(";
const char*	NewickTreeParser::cCloseParentNodeDelimiter	= ")";
const char*	NewickTreeParser::cWhitespaceChars			= " \t\n\r";

#pragma mark Error messages
const char*	NewickTreeParser::cTerminatorNotFound			= "';' expected to terminate the tree, but not found";
const char*	NewickTreeParser::cMissingCommaErr			= "',' expected to separate nodes, but not found";
const char*	NewickTreeParser::cMultiNodesAtTopErr			= "Unexpected ',' encountered: Only one node allowed at top level";

#pragma mark -
//---------------------------------------------------------------------------
NewickTreeParser::NewickTreeParser(void) {
	this->SkipWhitespace(true);
	this->SetWhitespaceChars(cWhitespaceChars);
	this->SetDelimiters(cDelimiters);
	this->SetQuoteChars(cQuotes);
	this->RemoveQuotes(true);
}

//---------------------------------------------------------------------------

NewickTreeParser::~NewickTreeParser(void)
{
}

//---------------------------------------------------------------------------

bool
NewickTreeParser::Read(
                       std::istream&	inTreeStream,
                       tree_type&		outTree)
{
	node_list	nodes;
	bool success = ReadNodeList(inTreeStream, true, nodes);
    
	if (success)
	{
		std::string token;
		success = this->ReadToken(inTreeStream, token);
        
		if (success)
		{
			if (token == cTreeTerminator)
			{
				if (nodes.size())
					outTree.SetRoot(nodes[0]);
			}
			else
			{
				success = false;
				this->ReportError(cTerminatorNotFound);
			}
		}
	}
    
	return success;
}

//---------------------------------------------------------------------------

bool
NewickTreeParser::Write(
                        std::ostream&		outTreeStream,
                        const tree_type&	outTree)
{
	return false;
}

//---------------------------------------------------------------------------

bool
NewickTreeParser::ReadNode(
                           std::istream&	inTreeStream,
                           node_type*&		outNode)
{
	node_type*	node = 0;
	
	std::string		token;
	bool success = this->ReadToken(inTreeStream, token);
	if (success)
	{
		if (!this->IsDelimiter(token)) {
			node = this->MakeNode(token);

     /*       size_t found;
            string token2(token);
            
            // Remove branchlength if specified
            found = token.find(":");
            if (found != string::npos) {
                token2 = token.substr(0, int(found) );
            }
            
            // Set the taxon name here
            strcpy( node->GetData().name, token2.c_str() );
      */
       }
		else
		{
			if (token == cOpenParentNodeDelimiter)
			{
				node_list children;
				success = ReadNodeList(inTreeStream, false, children);
				if (success)
				{
					node = new node_type;
					if (node) {
						node->AssignChildren(children);
					}
					else success = false;
                    
					success = this->ReadToken(inTreeStream, token);
					if (success)
						success = (token == cCloseParentNodeDelimiter);
                    
					if (success)
						success = this->ReadToken(inTreeStream, token);
                    
					if (success)
					{

						if (this->IsDelimiter(token)) {
  //                          cout << "   NOT DOING IT with " << token << endl;
							this->PutBackToken(token);
                        } else { 
                            MakeNode(token, node);
                        }
                    }
				}
			}
		}
        
		success = node;
	}
	
	outNode = node;
    
   // exit(-1);
	return success;
}

//---------------------------------------------------------------------------
// A node list is a comma separated list of nodes.

bool
NewickTreeParser::ReadNodeList(
                               std::istream&	inTreeStream,
                               bool			inLimitToOne,
                               node_list&		outNodes)
{
	bool success				= true;
	bool lastTokenWasSeparator	= true;
	
	node_list	nodes;
	
	std::string	token;
	do
	{
		success = this->ReadToken(inTreeStream, token);
		if (success)
		{
			node_type*	node = 0;
			
			if (this->IsDelimiter(token))
			{
				if (token == cOpenParentNodeDelimiter)
				{
					if (lastTokenWasSeparator)
					{
						this->PutBackToken(token);
						success = ReadNode(inTreeStream, node);
						lastTokenWasSeparator = false;
					}
					else
					{
						success = false;
						this->ReportError(cMissingCommaErr);
						break;
					}
				}
				else if (token == cNodeSeparator)
				{
					if (inLimitToOne)
					{
						success = false;
						this->ReportError(cMultiNodesAtTopErr);
						break;
					}
					
					if (lastTokenWasSeparator)
						node = new node_type;
                    
					lastTokenWasSeparator = true;
				}
				else
				{
					this->PutBackToken(token);
					break;
				}
			}
			else
			{
				if (lastTokenWasSeparator)
				{
					this->PutBackToken(token);
					success = ReadNode(inTreeStream, node);
					lastTokenWasSeparator = false;
				}
				else
				{
					success = false;
					this->ReportError(cMissingCommaErr);
					break;
				}
			}
			
			if (success && node)
				nodes.push_back(node);
		}
	} while(success);
	
	if (success)
		outNodes.swap(nodes);
	else
	{
		while(!outNodes.empty())
		{
			delete outNodes.back();
			outNodes.pop_back();
		}
	}
	
	return success;
}
