/*
 *  Numerics.cpp
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Numerics.h"
#include <stdio.h>
#include <iostream>
#include "PLEXApp.h"

#include "dSFMT.h"
#include "dSFMT.c"

#define TINY 1.0e-20;
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

// *****************************************************************************************************************************
// MARK: RNG interface
//       Random number generation interface
// *****************************************************************************************************************************

Numerics::Numerics(void) {
	initLookup(1000000);
    
    // Initialize random number array
    #if defined(__APPLE__) || \
    (defined(__FreeBSD__) && __FreeBSD__ >= 3 && __FreeBSD__ <= 6)
        randArray = (double *)malloc(sizeof(double) * blockSize);
        if (randArray == NULL) {
            printf("can't allocate memory.\n");
            exit(-1);
        }
    #elif defined(_POSIX_C_SOURCE)
        if (posix_memalign((void **)&randArray, 16, sizeof(double) * blockSize) != 0) {
            printf("can't allocate memory.\n");
            exit(-1);
        }
    #elif defined(__GNUC__) && (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 3))
        randArray = memalign(16, sizeof(double) * blockSize);
        if (randArray == NULL) {
            printf("can't allocate memory.\n");
            exit(-1);
        }
    #else // in this case, gcc doesn't support SSE2
        randArray = malloc(sizeof(double) * blockSize);
        if (randArray == NULL) {
            printf("can't allocate memory.\n");
            exit(-1);
        }
    #endif
    
    // Random number generator init
	dsfmt_init_gen_rand(&dsfmt, (int)options.randseed);
	dsfmt_fill_array_close_open(&dsfmt, randArray, blockSize);
}

Numerics& Numerics::getGlobalNumerics(void) {
	static Numerics	globalGenerator;
	return globalGenerator;
}

double Numerics::Random(void) {
	return dsfmt_genrand_close_open(&dsfmt);
}

// *****************************************************************************************************************************
// MARK: Numerics and util functions
// *****************************************************************************************************************************

double Numerics::gammln(double xx) {
	double x,tmp,ser;
	static double cof[6]={76.18009173,-86.50532033,24.01409822,-1.231739516,0.120858003e-2,-0.536382e-5};
	int j;
	
	x=xx-1.0;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.0;
	for (j=0;j<=5;j++) {
		x += 1.0;
		ser += cof[j]/x;
	}
	return -tmp+log(2.50662827465*ser);
}

double Numerics::factorial(int x) {
	// Use the log-factorial function for safety
    return exp(factln(x));
	
    /* 
    int i;
	int fac = 1;
	for (i=2; i<=x; i++) fac *= i;
	return fac;
     */
}

double Numerics::factln(double n) {
//	static double a[101];
	
	if (n < 0){ printf("Err: negative factorial in routine FACTLN"); exit(-1); }
	if (n <= 1) return 0.0;
	else return gammln(n+1.0);
}

// *****************************************************************************************************************************
// MARK: Random variate sampling
// *****************************************************************************************************************************

double Numerics::betaVariate( double a, double b ) {
	// Modified from Radford Neal
	/* BETA GENERATOR. Generates a real number, r, in (0,1), with density
	 proportional to r^(a-1) * (1-r)^(b-1).  Things are fiddled to avoid
	 the end-points, and to make the procedure symmetric between a and b. */
	
	double x, y, r;
	
    x = gammaVariate(a);
    y = gammaVariate(b);
    r = x/(x+y);
	
	return r;
}

double Numerics::gammaVariate( double aa ) {
	
	// Modified from Radford Neal
	/* GAMMA GENERATOR.  Generates a positive real number, r, with density
	 proportional to r^(a-1) * exp(-r).  See Devroye, p. 410 and p. 420.
	 Things are fiddled to avoid ever returning a value that is very near
	 zero. */
	
	double b, c, X, Y, Z, U, V, W;
	
	if (aa<0.00001) {
		X = aa;
	}
	
	else if (aa<=1) {
		U = Random();
		X = gammaVariate(1+aa) * pow(U,1/aa);
	}
	
	else if (aa<1.00001){ 
        X = expVariate(); 

    } else {
		b = aa-1;
		c = 3*aa - 0.75;
		
		for (;;) {
			U = Random();
			V = Random();
	
			W = U*(1-U);
			Y = sqrt(c/W) * (U-0.5);
			X = b+Y;
			if (X>=0)
			{
				Z = 64*W*W*W*V*V;
				
				if (Z <= 1 - 2*Y*Y/X || log(Z) <= 2 * (b*log(X/b) - Y)) break;
			}
		}
	}
	
	return X<1e-30 && X<aa ? (aa<1e-30 ? aa : 1e-30) : X;
}

double Numerics::expVariate (void) {
    // Modified from Radford Neal:
    // EXPONENTIAL GENERATOR.  See Devroye, p. 29.  Written so as to never return exactly zero.
	
    // Assuming rate parameter is 1
    
    return -log(Random());
}

double Numerics::expVariate (double lambda) {
    // Modified from Radford Neal:
    // EXPONENTIAL GENERATOR.  See Devroye, p. 29.  Written so as to never return exactly zero.
	
    // Assuming rate parameter is 1
    
    /*
    ofstream outTest;
    double rannum;
    
    cout << "Outputting rand numbers to disk for lambda of " << lambda << endl;

    outTest.open( "randTest2.txt" );
    for (int ii=0; ii < 1000000; ii++) {
        rannum = -log(Random())/lambda;
        outTest << rannum << endl;
    }
    outTest.close();
    exit(-1);
    */
    
    return -log(Random())/lambda;
}

double Numerics::gaussianVariate(double m, double s)	{
	// Normal random variate generator: mean m, std.dev s
    
	double x1, x2, w, y1;
	static double y2;
	static int use_last = 0;
    
	if (use_last) {
		y1 = y2;
		use_last = 0;
	} else {
		do {
			x1 = 2.0 * Numerics::getGlobalNumerics().Random() - 1.0;
			x2 = 2.0 * Numerics::getGlobalNumerics().Random() - 1.0;
			w = x1 * x1 + x2 * x2;
		} while ( w >= 1.0 );
        
		w = sqrt( (-2.0 * log( w ) ) / w );
		y1 = x1 * w;
		y2 = x2 * w;
		use_last = 1;
	}
    
	return( m + y1 * s );
}

int Numerics::_util_sampleFromCumulative_raw( DoubleVector *inVector, int length ) {
    
    static DoubleVector theVector;
    if (theVector.size() != length) theVector.resize( length, 0.0 );
    
	int i;
	double sum = 0.0; // max = -std::numeric_limits<double>::max( );
	
	for (i=0; i<length; i++) {
        if ((*inVector)[i] >= 0) {
            sum += (*inVector)[i];
        } else {
            //sum -= (*inVector)[i];
        }
    }
	
	for (i=0; i<length; i++) {
        if ((*inVector)[i] >= 0) {
            theVector[i] = (*inVector)[i] / sum;
        } else {
            theVector[i] = 0.0;
        }
        
        if (i>0) theVector[i] += theVector[i-1];
	}
    
//    cout << "Sampling distribution: " << endl;
//     for (i=0; i<length; i++) {
//     cout << "   State " << i << ": " << theVector[i] << endl;
//     }
	
	double rannum = Random();
	
	for (i=0; i<length;i++) {
//        cout << "i is " << i << endl;
		if (rannum <= theVector[i]) return i;
	}
    
//    cout << "ERROR, random num was " << rannum << ", i is " << i << "; eek!" << endl;
//    return i;
    return -1;
}

int Numerics::_util_sampleFromCumulative_raw( DoubleVector *inVector, int excludeState, int length ) {
    
    static DoubleVector theVector;
    if (theVector.size() != length) theVector.resize( length, 0.0 );
    
	int i;
	double sum = 0.0; // max = -std::numeric_limits<double>::max( );
	
	for (i=0; i<length; i++) {
        if (( (*inVector)[i] >= 0) && (i != excludeState) ) {
            sum += (*inVector)[i];
        } else {
            //sum -= (*inVector)[i];
        }
    }
	
    // If no valid states, return the starting state ('exclude state')
    if (sum == 0) {
        cout << "ERROR, rate distribution given starting state " << excludeState << "is zero:" << endl;
        for (int ii=0; ii<length; ii++) {
            cout << ii << ": " << setprecision(10) << (*inVector)[i] << endl;
        }
    }
    //return excludeState;
    
	for (i=0; i<length; i++) {
        if (( (*inVector)[i] >= 0) && (i != excludeState) ) {
            theVector[i] = (*inVector)[i] / sum;
        } else {
            theVector[i] = 0.0;
        }
        
        if (i>0) theVector[i] += theVector[i-1];
	}
    
//    cout << "Sampling distribution: " << endl;
//     for (i=0; i<length; i++) {
//     cout << "   State " << i << ": " << theVector[i] << endl;
//     }
	
	double rannum = Random();
	
    int chosen;
	for (i=0; i<length;i++) {
		if (rannum <= theVector[i]) {
            chosen= i;
            break;
        }
	}
    
    if (i == excludeState) {
        cout << "ERROR! we sampled the wrong state..." << endl;
        exit(-1);
    }
    
    return i;
}


// *****************************************************************************************************************************
// MARK: Function interface simplifiers
// *****************************************************************************************************************************

void ModelPropertiesForUpdate::setUpdateProperties( double theMu, int theX, int theY, int model ) {
    mu = theMu;
    
    x = theX;
    y = theY;
    theModel = model;
}

void ModelPropertiesForUpdate::attachData( DoubleMatrix& _CXYs, DoubleMatrix& _C1s, DoubleMatrix& _rateMatrix, DoubleMatrix& _alphaMatrix, DoubleVector& _stateFreqs, DoubleVector& _siteRates ) {
    CXYs = &_CXYs;
    C1s = &_C1s;
    
    rateMatrix  = &_rateMatrix;
    alphaMatrix = &_alphaMatrix;
    stateFreqs  = &_stateFreqs;
    siteRates   = &_siteRates;        
}

// *****************************************************************************************************************************
// MARK: Slice sampling
// *****************************************************************************************************************************
double Numerics::slice_sampleIndependentRateParameter( ModelPropertiesForUpdate& modelProperties, DoubleVector& averageMoveWidths ) {
    
    // Default to a uniform prior
    static UniformPrior prior;

    return slice_sampleIndependentRateParameter( prior, modelProperties, averageMoveWidths );
}

double Numerics::slice_sampleIndependentRateParameter( PriorDistribution &prior, ModelPropertiesForUpdate& modelProperties, DoubleVector& averageMoveWidths ) {
    // Slice sample: rate parameter when it applies to only a single state exchange
  
    // Expected that the PriorDistribution object already has its parameters set
    
    // Create some objects with parameters for the slice sampler and the model
    static SliceParameters slice;
    static PartialLogLikelihood_individualRateParm logLfxn;
   
    double startingValue;
    double L, R;
    int numSteps;
    
	// Set parameters for slice sampling
    numSteps = options._slice_numGenerationsPerSample;
    slice.w = options._slice_globalWindowMultiple;
    slice.m = options._slice_maximumWindowExpansions;

    // Attach the prior object
    slice.prior = &prior;
 
    // Set starting value - must be unscaled
    if (options.scaleByFrequencies != 1) {
        slice.x0 = modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y];
    } else {
        slice.x0 = modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y];
    }
    
    // Set min and max
    slice.minValue = 0.0;

    // MaxValue needs to be set based on the scaled matrix if scaling
    slice.maxValue = modelProperties.mu - (-modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] - modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y]);

    // Set w from previous average
    if ( averageMoveWidths.mDataArray[ (modelProperties.x * modelProperties.rateMatrix->rows() ) + modelProperties.y ] != 0.0 ) {
       slice.w = averageMoveWidths.mDataArray[ (modelProperties.x * modelProperties.rateMatrix->rows() ) + modelProperties.y ];
       // slice.w = (slice.maxValue - slice.minValue) / 4.0;
        
    } else {
        // Default, start with maxRange / 4
        slice.w = (slice.maxValue - slice.minValue) / 4.0;
    }
    
    // Store starting value for updating the matrices at the end
    startingValue = slice.x0;
    double startingValueAlt = modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y];
    
    // Set the bounds for the likleihood function as well (sets likelihood to floor)
    logLfxn.setBounds( 0.0, slice.maxValue );
    
    // Initialize the logL function with the parameters staying fixed
    logLfxn.setParms( modelProperties );


    // Diagnostic output for testing:
        // Diagnostic:   cout << "DIAG: MLE rate parm is : " << MLE_B1u_instRate((-rateMatrix.mRows.mDataArray[x].mDataArray[x] - slice.x0), C1s.mDataArray[x], CXYs.mRows.mDataArray[x].mDataArray[y], mu) << endl;
        // Diagnostic: ofstream out;
        // Diagnostic: out.open("test.out");
        // Diagnostic: cout << "Delta = " << slice.maxValue << ", CXY = " << CXYs.mRows.mDataArray[x].mDataArray[y] << ", C1 = " << C1s.mDataArray[x] << endl;
        // Diagnostic: numSteps = 50000;
        
        // Diagnostic: Can compare conjugate Gibbs using rbeta(n, CXY+alpha, C1+beta) for alpha, beta hyperparameters on the Beta prior
        //             NOTE however, that conjugate Gibbs sampler puts the prior on lambdaXY/Delta (i.e., lambdaXY/maxValue) 

    
    // Updated value will be 'x1'
	double x1 = 0.0;
    
    // Run the slice sampler for numSteps generations to reduce autocorrelation
    for (int i=0; i<numSteps; i++) {
        _slice_findInterval( L, R, slice, logLfxn );
        x1 = _slice_sampleByShrinkage( L, R, slice, logLfxn );

        slice.x0 = x1;
        // Diagnostic: out << slice.x0 << endl;
    }
    
    // Update the rate (and alpha) matrices
    if ( options.scaleByFrequencies != 1 ) {
        // alpha and rates matrices should be identical
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += startingValue - x1;
        
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1;            
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += startingValue - x1;      

    } else {
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += startingValue - x1;
        
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1 * modelProperties.stateFreqs->mDataArray[modelProperties.y];            
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += (startingValueAlt - x1) * modelProperties.stateFreqs->mDataArray[modelProperties.y];            
    }

    // Return the size of the move
    return ( abs( double( startingValue - x1) ) );
}

double Numerics::slice_sampleReversibleRateParameter( PriorDistribution &prior, ModelPropertiesForUpdate& modelProperties, DoubleVector& averageMoveWidths ) {
    // Slice sample: rate parameter when it applies to a pair of state exchanges
    // Expected that the PriorDistribution object already has its parameters set
    
    // NOTE: if scaleByFrequencies is on, we need the alpha matrix
    double SMALL = 1e-20;
    
    // Create some objects with parameters for the slice sampler and the model
    static SliceParameters slice;
    static PartialLogLikelihood_reversibleRateParm logLfxn;
    
    double startingValue;
    double L, R;
    int numSteps;
    
	// Set parameters for slice sampling
    numSteps = options._slice_numGenerationsPerSample;
    slice.w = options._slice_globalWindowMultiple;
    slice.m = options._slice_maximumWindowExpansions;
    
    // Attach the prior object
    slice.prior = &prior;
    
    // cout << "   DIAGNOSTIC: Updating " << x << " -> " << y << " with slice_sampleReversibleRateParameter()" << endl;
    
    // Set starting value - must be unscaled
    if (options.scaleByFrequencies != 1) {
        slice.x0 = modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y];
    } else {
        slice.x0 = modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y];
    }
    
    // DIAGNOSTIC
    if ( modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] != modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] ) {
        cout << "BEFORE SLICE UPDATE PARAMETERS " << modelProperties.x << " and " << modelProperties.y << " were not equal!  " << modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] << " and " << modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] << endl;
    }
    
    // Set min and max
    slice.minValue = 0.0;
    
    // MaxValue needs to be set based on the scaled matrix if scaling
    slice.maxValue = modelProperties.mu - (-modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] - modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y]);
    
    double maxValue2 = modelProperties.mu - (-modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.y] - modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x]);
    
    // Insure the ceiling is set on the lower of the two rows
    if ( (maxValue2 < slice.maxValue) && (maxValue2 > 0) ) slice.maxValue = maxValue2;
    
    slice.maxValue -= SMALL;

    // If maxValue < 0, mu was too small for the current rate matrix.  This may happen near the beginning of a run during burnin, but never any other time...
    if (slice.maxValue < 0) {
        cout << "   ** Slice sampler failed because mu is too small.  (If this happens more than a few times at the beginning of a chain, there is a problem.)" << endl;
        cout << "       Current " << modelProperties.x << ", " << modelProperties.y << "; Lambda = " << -modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] << "; mu is " << modelProperties.mu << endl;
        cout << "       Current " << modelProperties.y << ", " << modelProperties.x << "; Lambda = " << -modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.y] << "; mu is " << modelProperties.mu << endl;

        // Return from sampler without attempting a sample
        return 0.0;
    }
    
    //cout << "TEST: mu-maxValue-L1 = " << (modelProperties.mu - slice.maxValue - -modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x]) << ", test2 = " << (modelProperties.mu - slice.maxValue - -modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y]) << endl;
    
    // Set w from previous average
    if ( averageMoveWidths.mDataArray[ (modelProperties.x * modelProperties.rateMatrix->rows() ) + modelProperties.y ] != 0.0 ) {
        slice.w = averageMoveWidths.mDataArray[ (modelProperties.x * modelProperties.rateMatrix->rows() ) + modelProperties.y ];
        // slice.w = (slice.maxValue - slice.minValue) / 4.0;
        
    } else {
        // Default, start with maxRange / 4
        slice.w = (slice.maxValue - slice.minValue) / 4.0 ;
       // cout << "   INITIAL W FOR SLICE IS " << slice.w << endl;
       // cout << "   mu is " << modelProperties.mu << " and maxVal is set to " << slice.maxValue << " (with w = " << slice.w << "); Lambdax1 " << -modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] << ", LambdaX2 " << -modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.y] << endl;
    }
    
    
    // Store starting value for updating the matrices at the end
    startingValue = slice.x0;
    double startingAlt  = modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y];
    double startingAlt2  = modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x];
    
    // Set the bounds for the likleihood function as well (sets likelihood to floor)
    logLfxn.setBounds( 0.0, slice.maxValue );
    
    // Initialize the logL function with the parameters staying fixed
    logLfxn.setParms( modelProperties );
    
    
    // Diagnostic output for testing:
    // Diagnostic:   cout << "DIAG: MLE rate parm is : " << MLE_B1u_instRate((-rateMatrix.mRows.mDataArray[x].mDataArray[x] - slice.x0), C1s.mDataArray[x], CXYs.mRows.mDataArray[x].mDataArray[y], mu) << endl;
    // Diagnostic: ofstream out;
    // Diagnostic: out.open("test.out");
    // Diagnostic: cout << "Delta = " << slice.maxValue << ", CXY = " << CXYs.mRows.mDataArray[x].mDataArray[y] << ", C1 = " << C1s.mDataArray[x] << endl;
    // Diagnostic: numSteps = 50000;
    
    // Diagnostic: Can compare conjugate Gibbs using rbeta(n, CXY+alpha, C1+beta) for alpha, beta hyperparameters on the Beta prior
    //             NOTE however, that conjugate Gibbs sampler puts the prior on lambdaXY/Delta (i.e., lambdaXY/maxValue) 
    
    
    // Updated value will be 'x1'
	double x1 = 0.0;
    
    // Run the slice sampler for numSteps generations to reduce autocorrelation
    for (int i=0; i<numSteps; i++) {
        // Something in this function screws things up
        _slice_findInterval( L, R, slice, logLfxn );

        x1 = _slice_sampleByShrinkage( L, R, slice, logLfxn );
        
        slice.x0 = x1;
        // Diagnostic: out << slice.x0 << endl;
    }

    // Update the rate (and alpha) matrices
    if ( options.scaleByFrequencies != 1 ) {
        // alpha and rates matrices should be identical
        //cout << "NOT doing scale by freqs..." << endl;
        
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += startingValue - x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] = x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.y] += startingValue - x1;

        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1;            
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += startingValue - x1;      
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] = x1;            
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.y] += startingValue - x1;      

    } else {
       // cout << "Doing scale by freqs..." << endl;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += startingValue - x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] = x1;
        modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.y] += startingValue - x1;
        
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] = x1 * modelProperties.stateFreqs->mDataArray[modelProperties.y];            
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.x] += startingAlt - (x1 * modelProperties.stateFreqs->mDataArray[modelProperties.y]);            
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] = x1 * modelProperties.stateFreqs->mDataArray[modelProperties.x];            
        modelProperties.rateMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.y] += startingAlt2 - (x1 * modelProperties.stateFreqs->mDataArray[modelProperties.x]);            

    }
    
    // DIAGNOSTIC
    if ( modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] != modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] ) {
        cout << "AFTER SLICE UPDATE PARAMETERS " << modelProperties.x << " and " << modelProperties.y << " were not equal!  " << modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.y].mDataArray[modelProperties.x] << " and " << modelProperties.alphaMatrix->mRows.mDataArray[modelProperties.x].mDataArray[modelProperties.y] << endl;
        exit(-1);
    }

    
   // [ARGH!] options.scaleByFrequencies = 0;

    
    // Return the size of the move
    return ( abs( double( startingValue - x1) ) );
     
}

double averageCountsPerSet_fi;

void Numerics::_slice_findInterval( double &L, double &R, SliceParameters &slice, PartialLogLikelihood &f ) {
	// Find the interval I=(L,R) using the 'stepping out' procedure of Neal (2003)
    
    // NOTE: To replicate prior used for conjugate Gibbs sampler, remember that prior is on lambdaXY/slice.maxValue (not lambdaXY)
    //       This tends to over-emphasize the prior when rates are large.
	
    // x0 is the starting value of x
	// f(x) is assumed to be the LOG likelihood
    
    const double SMALL = 0.0; // 1e-30;
    
	double U, V, J, K;
    
    static double aveCount = 0.0;
    static int callCount=0;
    int count=0;

    // Sample a vertical level at which to draw the slice
    slice.y = ( f.logL( slice.x0 ) + ( slice.prior->evaluateLogPDF( slice.x0 ) ) ) - expVariate();
    
    count++;
    
//    return; 
//    
//    // WTF?!  BLOCKING BELOW HERE CAUSES PROBLEM...
//    
//    if ( slice.w < 0 ) {
//        cout << "ERROR, slice.w became <0! " << slice.w << "; min " << slice.minValue << ", max " << slice.maxValue << endl;
//        exit(-1);
//    }
    
    U = Random(); 
    L = slice.x0 - slice.w * U;
    R = L + slice.w;
    
    if ( L <= slice.minValue ) { L = slice.minValue + SMALL; R = L + slice.w; } //cout << "     Set L to min" << endl;}
    if ( R >= slice.maxValue ) { L = slice.maxValue - SMALL - slice.w; R = L + slice.w; } //  cout << "  Set R to max" << endl;}
    
	V = Random(); 
	J = floor( slice.m * V );
	K = (slice.m - 1) - J;
    
	while ( (J > 0) && (slice.y < ( f.logL( L ) + ( slice.prior->evaluateLogPDF( L ) ) ) ) ) {
        count++;
        L = L - slice.w;
        if (L <= slice.minValue) L = slice.minValue + SMALL;
		J = J-1;
	}
    
	while ( (K > 0) && (slice.y < ( f.logL( R ) + ( slice.prior->evaluateLogPDF( R ) ) ) ) ) {
        count++;
        R = R + slice.w;
        if (R >= slice.maxValue) R = slice.maxValue - SMALL;
		K = K-1;
	}
    
    callCount++;
    aveCount += count;
        
    if ( (callCount % 20) == 0 ) {
       // cout << "  ** Running find-interval average is: " << (aveCount / 20) << endl;
        averageCountsPerSet_fi = aveCount;
        aveCount = 0;
    }        

}

double Numerics::_slice_sampleByShrinkage( double L, double R, SliceParameters &slice, PartialLogLikelihood &f ) {
	// Sample a new value by Neal's shrinkage method (Neal, 2003)
    
	double startVal = slice.x0;
    double returnVal = slice.x0;
	double Lp = L;
	double Rp = R;
	double U;
    
    static double aveCount = 0.0;
    static int callCount=0;
    int count=0;
    
    int rejectCount = 0;
	while (1) {
		U = Random(); 
		returnVal = Lp + U * (Rp - Lp);
        
        if ( ( returnVal >= slice.minValue) && (returnVal <= slice.maxValue) ) {
            count++;
            if ( slice.y < ( f.logL( returnVal ) + ( slice.prior->evaluateLogPDF( (returnVal ) ) ) ) ) {
                break;
            }
        }
        
		if ( returnVal < slice.x0 ) {
			Lp = returnVal;
		} else {
			Rp = returnVal;
		}
        
        rejectCount++;
        if ( (abs(Lp - Rp) <= 1e-10) || (rejectCount >1000) ) { cout << "       [ INTERVAL SHRUNK TO " << abs( Lp - Rp ) << " WITH REJECTCOUNT = " << rejectCount << "; min " << slice.minValue << ", max " << slice.maxValue << ", w " << slice.w << "; L " << Lp << ", R " << Rp << endl; returnVal = startVal; break; }
	};
    
   // cout << "       Slice sampler used " << count << " posterior evals..." << endl;
    
    callCount++;
    aveCount += count;
    
    if ( (callCount % 20) == 0 ) {
      //  cout << "  ** Running sampler average is: " << ( (aveCount / 20)) << endl;
      //  cout << "       ** Total = " <<  (aveCount / 20) + (averageCountsPerSet_fi/20) << endl;
        aveCount = 0;
    }        
    
    return returnVal;
}

// *****************************************************************************************************************************
// MARK: Sub. prob. calcs for branch
// *****************************************************************************************************************************

double Numerics::calculateSubstitutionProbabilityBranch_B1u( const DoubleMatrix &theMatrix, double t, int x, int y) {
	// This version gets the optimal mu...
	double mu = 0;
	for (int i=0; i<theMatrix.rows(); i++) {
        if ( (-1.0 * theMatrix.mRows.mDataArray[i].mDataArray[i]) > mu ) {
            mu = -1.0 * theMatrix.mRows.mDataArray[i].mDataArray[i];
        }
    }
    
	double returnVal;
	if ( x != y ) {
		returnVal = theMatrix.mRows.mDataArray[x].mDataArray[y] * t / ( 1 + mu * t );
	} else {
		returnVal = (1 + ( mu + theMatrix.mRows.mDataArray[x].mDataArray[x]) * t) / ( 1 + mu * t );
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }
	
	return returnVal;
}	

double Numerics::calculateSubstitutionProbabilityBranch_B1u( const DoubleMatrix &theMatrix, double t, int x, int y, double mu ) {
	return calculateSubstitutionProbabilityBranch_B1u(theMatrix, t, x, y, mu, 1.0);
}

double Numerics::calculateSubstitutionProbabilityBranch_B1u( const DoubleMatrix &theMatrix, double t, int x, int y, double mu, double rateScaleFactor ) {
    // This version scales the rates by a scale-factor
	double returnVal;
	
	if ( x != y ) {
		returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * rateScaleFactor) * t / ( 1 + mu * t );
	} else {
		returnVal = (1 + ( mu + (theMatrix.mRows.mDataArray[x].mDataArray[x] * rateScaleFactor)) * t) / ( 1 + mu * t );
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }
	return returnVal;
}	

// *** THE BELOW SUBSTITUTION PROBABILITY CALCULATIONS ARE DESCRIBED IN Krishnan et al. (2004) AND ARE NO LONGER FULLY SUPPORTED ***

double Numerics::calculateSubstitutionProbabilityBranch_B1( const DoubleMatrix &theMatrix, double t, int x, int y ) {
	double returnVal;
    
	if (x != y) {
		if (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y] > 0.0001 || theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[x].mDataArray[x] > 0.0001) {
			returnVal = ( theMatrix.mRows.mDataArray[x].mDataArray[y] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) - exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t)) / (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y]) );
		} else {
			returnVal = ( theMatrix.mRows.mDataArray[x].mDataArray[y] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) * t ) );
		}
	} else {
		returnVal = ( exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) );
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }
	return returnVal;
}	

double Numerics::calculateSubstitutionProbabilityBranch_B2( const DoubleMatrix &theMatrix, double t, int x, int y, int z ) {
	double returnVal;
	
	if (x == y || y == z) {	// only one substitution or no substitution
		return 0.0;
	} else {
		if (theMatrix.mRows.mDataArray[y].mDataArray[y]- theMatrix.mRows.mDataArray[z].mDataArray[z] < 0.0001 || theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[y].mDataArray[y] < 0.0001) {
            // when rate[y] == rate[z]
			if (theMatrix.mRows.mDataArray[x].mDataArray[x]- theMatrix.mRows.mDataArray[z].mDataArray[z] < 0.0001 || theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[x].mDataArray[x] < 0.0001)
                // when rate[x] = rate[z]
				returnVal = ( theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t) * t * t * 0.5 );
			else
                // when rate[x] != rate[z]
				returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t) * (exponential((theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z]) * t) - 1 - t * (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z])) / pow(theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z], 2) );
		} else {
			// when rate[y] != rate[z]
			if (theMatrix.mRows.mDataArray[x].mDataArray[x]- theMatrix.mRows.mDataArray[y].mDataArray[y] < 0.0001 || theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[x].mDataArray[x] < 0.0001)
				// when rate[x] == rate[y]
				returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * t * exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t) / (theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z]) -theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t)-exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t))/((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z])*(theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z])));
			else {
				// rate[x] != rate[y]
				if ((theMatrix.mRows.mDataArray[x].mDataArray[x]- theMatrix.mRows.mDataArray[z].mDataArray[z] < 0.0001 || theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[x].mDataArray[x] < 0.0001))
					// rate[x] = rate[z]
					returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * (exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t)-exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t))/((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z])*(theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y])) - theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * t * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t) / (theMatrix.mRows.mDataArray[y].mDataArray[y] -theMatrix.mRows.mDataArray[z].mDataArray[z]));
				else
					// rate[x] != rate[z]
					returnVal = (theMatrix.mRows.mDataArray[x].mDataArray[y] * theMatrix.mRows.mDataArray[y].mDataArray[z] * ((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z]) * exponential(theMatrix.mRows.mDataArray[x].mDataArray[x] * t) + (theMatrix.mRows.mDataArray[z].mDataArray[z] - theMatrix.mRows.mDataArray[x].mDataArray[x]) * exponential(theMatrix.mRows.mDataArray[y].mDataArray[y] * t) + (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y]) * exponential(theMatrix.mRows.mDataArray[z].mDataArray[z] * t))/((theMatrix.mRows.mDataArray[y].mDataArray[y] - theMatrix.mRows.mDataArray[z].mDataArray[z]) * (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[y].mDataArray[y]) * (theMatrix.mRows.mDataArray[x].mDataArray[x] - theMatrix.mRows.mDataArray[z].mDataArray[z])));
			}
		}
	}
	
	if (returnVal <= 0) { returnVal = 1e-70; }
	return returnVal;
}	

// *****************************************************************************************************************************
// MARK: Partial B1u likelihood functions
// *****************************************************************************************************************************
double Numerics::b1u_likelihoodContribution_rateParameter( double lambdaXY, double LambdaX, double C1, double CXY, double mu) {
//    cout << "called b1u_likelihoodContribution_rateParameter().  Parms are lXY " << lambdaXY << " Lx " << LambdaX << ", C1 " << C1 << ", CXY " << CXY << ", mu " << mu << endl;

    if (options.allowASRV) {
        cout << "       ERROR: wrong partial likelihood function called when ASRV is active b1u_likelihoodContribution_rateParameter()" << endl;
        exit(-1);
    }
    return C1 * logarithm( (mu - LambdaX) ) + CXY * logarithm( lambdaXY );
}

double Numerics::b1u_likelihoodContribution_rateParameterOnly( double lambdaXY, double CXY ) {
	return CXY * logarithm( lambdaXY );
}

double Numerics::b1u_likelihoodContribution_rateAwayOnly( int x, int theModel, double LambdaX, DoubleMatrix* C1s, double mu, DoubleVector* siteRates ) {
    // The interface for this function got a little messy due to integration of rate categories.  Make this simpler later.
    double theValue = 0.0;
    
    if (options.allowASRV) {
        // Over all rate categories
        for (int i=0; i < options.numCatASRV; i++) {
            // Add up the effect on all rate categories, given the virtual sub counts per category (and the site rates)
            theValue += b1u_likelihoodContribution_rateAwayOnly( ( LambdaX * (*siteRates).mDataArray[ i ] ), (*C1s).mRows.mDataArray[ returnModelIndex( theModel, i, 0) ].mDataArray[ x ], mu );
        }
    } else {
        theValue = b1u_likelihoodContribution_rateAwayOnly( LambdaX, (*C1s).mRows.mDataArray[ theModel ].mDataArray[ x ], mu );
    }
    
    return theValue;
}

double Numerics::b1u_likelihoodContribution_rateAwayOnly( double LambdaX, double C1, double mu) {
    return C1 * logarithm( (mu - LambdaX) );
}

// *****************************************************************************************************************************
// MARK: B1u MLEs and analytical variances for certain models
// *****************************************************************************************************************************

double Numerics::MLE_B1u_branchlength( double C1, double C0, double mu) {
    // C1 is the # of Cxx(virtual) + Cxy changes along the branch; C0 is the number of Cxx (no sub) events
	return ( C1 / (C0 * mu) );
}

double Numerics::MLE_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu) {
	return ( (CXY * (mu - LambdaXsp)) / (CXY + C1) );
}

double Numerics::Variance_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu) {
	// variance at the MLE...
	return ( sqrt( ( ( C1 * pow( (mu - LambdaXsp), 2)) / ( pow( (C1+CXY),3)))));
}

double Numerics::Variance_B1u_GivenLambda( double lambdaXY, double LambdaX, double C1, double CXY, double mu) {
	// asymptotic variance is:
	return sqrt( 1.0 / ( (CXY / pow(lambdaXY,2)) + ( C1 / pow( (mu - LambdaX - lambdaXY), 2 ) ) ) );
}

// *****************************************************************************************************************************
// MARK: Proposal evaluation functions
//       (Metropolis moves, with likelihood heating for 'power-posterior' marginal likelihood calc.)
// *****************************************************************************************************************************

int Numerics::evaluateProposal_logL( double logLikRatio ) {
	return evaluateProposal_logL( 1.0, (1.0 - logLikRatio), 1.0, 1.0, 1.0, 1.0 );
}

int Numerics::evaluateProposal_logL( double newLik, double lastLik ) {
	return evaluateProposal_logL( newLik, lastLik, 1.0, 1.0, 1.0, 1.0 );
}

int Numerics::evaluateProposal_logL( double newLik, double lastLik, double likHeat, double posteriorHeat ) {
	return evaluateProposal_logL( newLik, lastLik, 1.0, 1.0, likHeat, posteriorHeat );
}

int Numerics::evaluateProposal_logL( double newLik, double lastLik, double proposalRatio, double priorRatio, double likHeat, double posteriorHeat ) {
	// Expecting the likelihoods to be input as log-likelihoods
	double acceptanceCutoff;
	
	if (options.__globalLikelihoodTemp != 1.0) {
		likHeat = options.__globalLikelihoodTemp;
	}
	
	if (likHeat != 1.0) {
		acceptanceCutoff = exponential( ( likHeat * (newLik - lastLik) ) );
	} else {
		acceptanceCutoff = exponential( (newLik - lastLik) );
	}	
	
	acceptanceCutoff *= priorRatio * proposalRatio;
    
	if (posteriorHeat != 1.0) {
		double tempval = acceptanceCutoff;
		acceptanceCutoff = pow( tempval, posteriorHeat);
	}
	
	if ( acceptanceCutoff >= 1 ) {
		return 1;
	} else {
		double u = Random();
        
		if (u < acceptanceCutoff ) {
			return 1;
		} else {
			return 0;
		}	
	}
}

// *****************************************************************************************************************************
// MARK: Linear systems
//       (solve for stationary frequencies
// *****************************************************************************************************************************

int Numerics::NR_ludcmp(double *a, int n, int *indx, double *d) {
	// Adapted from Numerical Recipes
	int i,imax,j,k;
	double big,dum,sum,temp;
	static double *vv;
    static int vvAlloc=0;
    
    imax = 0;
	
    if (vvAlloc == 0) {
        vv = new double[n+1];
        vvAlloc = 1;
    }
    
//	vv=vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++) {
		//	std::cout << "fabs of a[" << i << ", " << j << "] = " << fabs(a[i*n+j]) << endl;
			if ((temp=fabs(a[i*n+j])) > big) big=temp;
		}
		if (big == 0.0) { std::cout << "Singular matrix in routine ludcmp: error doing LU decomposition" << endl; return -1;} //exit(-1); }
		vv[i]=1.0/big;
	}
	
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i*n+j];
			for (k=1;k<i;k++) sum -= a[i*n+k]*a[k*n+j];
			a[i*n+j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i*n+j];
			for (k=1;k<j;k++)
				sum -= a[i*n+k]*a[k*n+j];
			a[i*n+j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}

		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax*n+k];
				a[imax*n+k]=a[j*n+k];
				a[j*n+k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
	
		indx[j]=imax;
		if (a[j*n+j] == 0.0) a[j*n+j]=TINY;
		if (j != n) {
			dum=1.0/(a[j*n+j]);
			for (i=j+1;i<=n;i++) a[i*n+j] *= dum;
		}
	}
	
	//delete [] vv;
    return 1;
}

void Numerics::NR_lubksb(double *a, int n, int *indx, double *b) {
	// Adapted from Numerical Recipes
	int i,ii=1,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i*n+j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i*n+j]*b[j];
		b[i]=sum/a[i*n+i];
	}
}

// *****************************************************************************************************************************
// MARK: LUT functions
//       Lookup-table functions: logarithm() and exponential()
// *****************************************************************************************************************************
void Numerics::initLookup(int theN) {
	n = theN;
	logFloor = log(1e-70);
	logLookupTable = new double[n];
	expLookupTable = new double[n];

	precomputeLogTable(logLookupTable);
	precomputeExpTable(expLookupTable);
	theMode = 1; //setLUTmode(1);
}

void Numerics::setLUTmode(int val) {
	theMode = val;
}

void Numerics::precomputeLogTable(double *LUT) { 
	min = 0.0000000001;
	max = 5.000000000;
	incr = (max-min)/n;
	
	for(int i=0;i<n;i++) { 
		LUT[i] = log((min + (i*incr))); //save the log value 
	}
} 

void Numerics::precomputeExpTable(double *LUT) { 
	expMin = -5.00000;
	expMax = 5.000000;
	expIncr = (expMax-expMin)/n;
	
	for(int i=0;i<n;i++) { 
		LUT[i] = exp((expMin + (i*expIncr))); //save the log value 
	}
} 

double Numerics::logarithm(double val) { 
    // Allow 'theMode' to override LUT usage
    if (theMode==0) { return log(val); }
    
    // Minimum value
	if (val == 1e-70) { return logFloor;}
    
    // If requested value is out of table range, calcuate it with standard function
	if ((val > max) || (val < min)) { return log(val); }
	
    static int x1,x2;
	static double y1,y2, theX1, theX2;
	x1 = static_cast <int>( (val-min)/incr );
	x2 = x1+1;
	theX1 = min + (x1*incr);
	theX2 = min + (x2*incr);
	
    // Use linear interpolation between LUT entries
	y1 = logLookupTable[x1]; y2 = logLookupTable[x2];
	return y1 + (val-theX1)*(y2-y1)/(theX2-theX1);
} 

double Numerics::exponential(double val) { 
    // Allow 'theMode' to override LUT usage
	if (theMode==0) { return exp(val); }
    
    // If requested value is out of table range, calculate it normally
    if ((val > expMax) || (val < expMin)) { return exp(val); }
    
    static int x1,x2;
	static double y1,y2, theX1, theX2;
	x1 = static_cast <int>( (val-expMin)/expIncr );
	x2 = x1+1;
	theX1 = expMin + (x1*expIncr);
	theX2 = expMin + (x2*expIncr);
	
    // Use linear interpolation between LUT entries
    y1 = expLookupTable[x1]; y2 = expLookupTable[x2];
	return y1 + (val-theX1)*(y2-y1)/(theX2-theX1);
} 


// *****************************************************************************************************************************
// MARK: Quantile functions
//       Imported: Empirical quantiles functions (Modified from GSL)
// *****************************************************************************************************************************

void Numerics::simpleSort(double *list, int arraySize) {
	for (int i = 0; i < arraySize ; i++) {
		int potentialSmallest = i; //assume this is smallest
		
		//look over remaining elements to find smallest.
		for (int j = i + 1; j < arraySize; j++) {
			if (list[j] < list[potentialSmallest]) {
				//remember index for latter swap.
				potentialSmallest = j;
			}
		}
		
		//swap smallest remaining element
		double temp = list[i];
		list[i] = list[potentialSmallest];
		list[potentialSmallest] = temp;
	}
}

double Numerics::gsl_stats_quantile_from_sorted_data(const double sorted_data[],const size_t stride,const size_t n,const double f) {
	const double index = f * (n - 1) ;
	const size_t lhs = (int)index ;
	const double delta = index - lhs ;
	double result;
	
	if (n == 0)
		return 0.0 ;
	
	if (lhs == n - 1)
    {
		result = sorted_data[lhs * stride] ;
    }
	else
    {
		result = (1 - delta) * sorted_data[lhs * stride] + delta * sorted_data[(lhs + 1) * stride] ;
    }
	
	return result ;
}

// *****************************************************************************************************************************
// MARK: Matrix exp()
//       Imported functions for matrix exponentiation (modified from PAML)
// *****************************************************************************************************************************
int Numerics::PAML_PMatUVRoot (DoubleMatrix *P, double t, int n, double U[], double V[], double Root[]) {
    /* P(t) = U * exp{Root*t} * V */
    int i,j,k;
    double expt, uexpt;
    double smallp = 0;
    if (t<0) { t = 0.0; }
    
	for (i=0;i<n;i++) {
		for (j=0;j<n;j++) {
			P->mRows.mDataArray[i].mDataArray[j] = 0.0;
		}
	}
	
    if (t<1e-100) {
		for (i=0;i<n;i++) {
			P->mRows.mDataArray[i].mDataArray[i] = 1.0;
		}
		return(0);
	}
	
    
    for (k=0; k<n; k++) {
        for (i=0,expt=exp(t*Root[k]); i<n; i++) {
            for (j=0,uexpt=U[i*n+k]*expt; j<n; j++) {
                P->mRows.mDataArray[i].mDataArray[j] += uexpt*V[k*n+j];
            }
        }
    }
    
    for(i=0;i<n;i++) {
        for(j=0;j<n;j++) {
            if ( P->mRows.mDataArray[i].mDataArray[j]<smallp) P->mRows.mDataArray[i].mDataArray[j]=0;
        }
    }
    return (0);
}

int Numerics::PAML_eigenQREV (double Q[], double pi[], int n, 
               double Root[], double U[], double V[]) {
/* 
   This finds the eigen solution of the rate matrix Q for a time-reversible 
   Markov process, using the algorithm for a real symmetric matrix.
   Rate matrix Q = S * diag{pi} = U * diag{Root} * V, 
   where S is symmetrical, all elements of pi are positive, and U*V = I.
   space[n] is for storing sqrt(pi).

   [U 0] [Q_0 0] [U^-1 0]    [Root  0]
   [0 I] [0   0] [0    I]  = [0     0]

   Ziheng Yang, 25 December 2001 (ref is CME/eigenQ.pdf)
*/

    int i,j, inew, jnew, nnew, status;
    static double *pi_sqrt;
    double small=1e-100;
    static int allocated = 0;
    
	if (allocated==0) {
		pi_sqrt = new double[n];
		allocated=1;
	}
    for(j=0,nnew=0; j<n; j++)
        if(pi[j]>small)
            pi_sqrt[nnew++]=sqrt(pi[j]);
    
    /* store in U the symmetrical matrix S = sqrt(D) * Q * sqrt(-D) */
    
    if(nnew==n) {
        for(i=0; i<n; i++)
            for(j=0,U[i*n+i] = Q[i*n+i]; j<i; j++)
                U[i*n+j] = U[j*n+i] = (Q[i*n+j] * pi_sqrt[i]/pi_sqrt[j]);
        
        status=PAML_eigenRealSym(U, n, Root, V);
        for(i=0;i<n;i++) for(j=0;j<n;j++)  V[i*n+j] = U[j*n+i] * pi_sqrt[j];
        for(i=0;i<n;i++) for(j=0;j<n;j++)  U[i*n+j] /= pi_sqrt[i];
    }
    else {
        for(i=0,inew=0; i<n; i++) {
            if(pi[i]>small) {
                for(j=0,jnew=0; j<i; j++) 
                    if(pi[j]>small) {
                        U[inew*nnew+jnew] = U[jnew*nnew+inew] 
                        = Q[i*n+j] * pi_sqrt[inew]/pi_sqrt[jnew];
                        jnew++;
                    }
                U[inew*nnew+inew] = Q[i*n+i];
                inew++;
            }
        }
        
        status=PAML_eigenRealSym(U, nnew, Root, V);
        
        for(i=n-1,inew=nnew-1; i>=0; i--)   /* construct Root */
            Root[i] = (pi[i]>small ? Root[inew--] : 0);
        for(i=n-1,inew=nnew-1; i>=0; i--) {  /* construct V */
            if(pi[i]>small) {
                for(j=n-1,jnew=nnew-1; j>=0; j--)
                    if(pi[j]>small) {
                        V[i*n+j] = U[jnew*nnew+inew]*pi_sqrt[jnew];
                        jnew--;
                    }
                    else 
                        V[i*n+j] = (i==j);
                inew--;
            }
            else 
                for(j=0; j<n; j++)  V[i*n+j] = (i==j);
        }
        for(i=n-1,inew=nnew-1; i>=0; i--) {  /* construct U */
            if(pi[i]>small) {
                for(j=n-1,jnew=nnew-1;j>=0;j--)
                    if(pi[j]>small) {
                        U[i*n+j] = U[inew*nnew+jnew]/pi_sqrt[inew];
                        jnew--;
                    }
                    else 
                        U[i*n+j] = (i==j);
                inew--;
            }
            else 
                for(j=0;j<n;j++)
                    U[i*n+j] = (i==j);
        }
    }
    
    /*   This routine works on P(t) as well as Q. */
    /*
     if(fabs(Root[0])>1e-10 && noisy) printf("Root[0] = %.5e\n",Root[0]);
     Root[0]=0; 
     */
    return(status);
}

int Numerics::PAML_eigenRealSym(double A[], int n, double Root[], double work[]) {
/* This finds the eigen solution of a real symmetrical matrix A[n*n].  In return, 
   A has the right vectors and Root has the eigenvalues. 
   work[n] is the working space.
   The matrix is first reduced to a tridiagonal matrix using HouseholderRealSym(), 
   and then using the QL algorithm with implicit shifts.  

   Adapted from routine tqli in Numerical Recipes in C, with reference to LAPACK
   Ziheng Yang, 23 May 2001
*/
   int status=0;
   PAML_HouseholderRealSym(A, n, Root, work);
   status=PAML_EigenTridagQLImplicit(Root, work, n, A);
   PAML_EigenSort(Root, A, n);

   return(status);
}

void Numerics::PAML_EigenSort(double d[], double U[], int n) {
/* this sorts the eigen values d[] and rearrange the (right) eigen vectors U[] */
   int k,j,i;
   double p;

   for (i=0;i<n-1;i++) {
      p=d[k=i];
      for (j=i+1;j<n;j++)
         if (d[j] >= p) p=d[k=j];
      if (k != i) {
         d[k]=d[i];
         d[i]=p;
         for (j=0;j<n;j++) {
            p=U[j*n+i];
            U[j*n+i]=U[j*n+k];
            U[j*n+k]=p;
         }
      }
   }
}

void Numerics::PAML_HouseholderRealSym(double a[], int n, double d[], double e[]) {
    /* This uses HouseholderRealSym transformation to reduce a real symmetrical matrix 
     a[n*n] into a tridiagonal matrix represented by d and e.
     d[] is the diagonal (eigends), and e[] the off-diagonal.
     */
    int m,k,j,i;
    double scale,hh,h,g,f;
    
    for (i=n-1;i>=1;i--) {
        m=i-1;
        h=scale=0;
        if (m > 0) {
            for (k=0;k<=m;k++)
                scale += fabs(a[i*n+k]);
            if (scale == 0)
                e[i]=a[i*n+m];
            else {
                for (k=0;k<=m;k++) {
                    a[i*n+k] /= scale;
                    h += a[i*n+k]*a[i*n+k];
                }
                f=a[i*n+m];
                g=(f >= 0 ? -sqrt(h) : sqrt(h));
                e[i]=scale*g;
                h -= f*g;
                a[i*n+m]=f-g;
                f=0;
                for (j=0;j<=m;j++) {
                    a[j*n+i]=a[i*n+j]/h;
                    g=0;
                    for (k=0;k<=j;k++)
                        g += a[j*n+k]*a[i*n+k];
                    for (k=j+1;k<=m;k++)
                        g += a[k*n+j]*a[i*n+k];
                    e[j]=g/h;
                    f += e[j]*a[i*n+j];
                }
                hh=f/(h*2);
                for (j=0;j<=m;j++) {
                    f=a[i*n+j];
                    e[j]=g=e[j]-hh*f;
                    for (k=0;k<=j;k++)
                        a[j*n+k] -= (f*e[k]+g*a[i*n+k]);
                }
            }
        } 
        else
            e[i]=a[i*n+m];
        d[i]=h;
    }
    d[0]=e[0]=0;
    
    /* Get eigenvectors */
    for (i=0;i<n;i++) {
        m=i-1;
        if (d[i]) {
            for (j=0;j<=m;j++) {
                g=0;
                for (k=0;k<=m;k++)
                    g += a[i*n+k]*a[k*n+j];
                for (k=0;k<=m;k++)
                    a[k*n+j] -= g*a[k*n+i];
            }
        }
        d[i]=a[i*n+i];
        a[i*n+i]=1;
        for (j=0;j<=m;j++) a[j*n+i]=a[i*n+j]=0;
    }
}

int Numerics::PAML_EigenTridagQLImplicit(double d[], double e[], int n, double z[]) {
/* This finds the eigen solution of a tridiagonal matrix represented by d and e.  
   d[] is the diagonal (eigenvalues), e[] is the off-diagonal
   z[n*n]: as input should have the identity matrix to get the eigen solution of the 
   tridiagonal matrix, or the output from HouseholderRealSym() to get the 
   eigen solution to the original real symmetric matrix.
   z[n*n]: has the orthogonal matrix as output

   Adapted from routine tqli in Numerical Recipes in C, with reference to
   LAPACK fortran code.
   Ziheng Yang, May 2001
*/
   int m,j,iter,niter=30, status=0, i,k;
   double s,r,p,g,f,dd,c,b, aa,bb;

   for (i=1;i<n;i++) e[i-1]=e[i];  e[n-1]=0;
   for (j=0;j<n;j++) {
      iter=0;
      do {
         for (m=j;m<n-1;m++) {
            dd=fabs(d[m])+fabs(d[m+1]);
            if (fabs(e[m])+dd == dd) break;  /* ??? */
         }
         if (m != j) {
            if (iter++ == niter) {
               status=-1;
               break;
            }
            g=(d[j+1]-d[j])/(2*e[j]);

            /* r=pythag(g,1); */

            if((aa=fabs(g))>1)  r=aa*sqrt(1+1/(g*g));
            else                r=sqrt(1+g*g);

            g=d[m]-d[j]+e[j]/(g+SIGN(r,g));
            s=c=1;
            p=0;
            for (i=m-1;i>=j;i--) {
               f=s*e[i];
               b=c*e[i];

               /*  r=pythag(f,g);  */
               aa=fabs(f); bb=fabs(g);
               if(aa>bb)       { bb/=aa;  r=aa*sqrt(1+bb*bb); }
               else if(bb==0)             r=0;
               else            { aa/=bb;  r=bb*sqrt(1+aa*aa); }

               e[i+1]=r;
               if (r == 0) {
                  d[i+1] -= p;
                  e[m]=0;
                  break;
               }
               s=f/r;
               c=g/r;
               g=d[i+1]-p;
               r=(d[i]-g)*s+2*c*b;
               d[i+1]=g+(p=s*r);
               g=c*r-b;
               for (k=0;k<n;k++) {
                  f=z[k*n+i+1];
                  z[k*n+i+1]=s*z[k*n+i]+c*f;
                  z[k*n+i]=c*z[k*n+i]-s*f;
               }
            }
            if (r == 0 && i >= j) continue;
            d[j]-=p; e[j]=g; e[m]=0;
         }
      } while (m != j);
   }
   return(status);
}

// *****************************************************************************************************************************
// MARK: Misc. functions
// *****************************************************************************************************************************

int Numerics::returnModelIndex( int modelassignment, int rateCatAssignment, int discreteBranchLengthIndex ) {
	/* if ( options.numCatASRV < 1 ) {
	 std::cout << “ERROR! numCatASRV must be >= 1 always” << endl;
	 exit(-1);
	 }
	 
	 if ( options.numDiscreteBLs < 1 ) {
	 std::cout << “ERROR! numCatASRV must be >= 1 always” << endl;
	 exit(-1);
	 } */
	
	return ( modelassignment * ( options.numCatASRV * options.numDiscreteBLs ) + (discreteBranchLengthIndex * options.numCatASRV) + rateCatAssignment );
}


