/*
 *  Numerics.h
 /*
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef Numerics_h__
#define Numerics_h__

#include <cmath>
#include "CLAMPTypes.h"
#include "PartialLogLikelihood.h"
#include "PriorDistributions.h"

#define	HAVE_SSE2
#define _XOPEN_SOURCE 600
#include "dSFMT.h"

// Forward declaration
class PartialLogLikelihood;

// To simplify some function interfaces
class ModelPropertiesForUpdate {
public:
    int x, y, theModel;
    double mu;
    DoubleMatrix* CXYs;
    DoubleMatrix* C1s;
    
    DoubleMatrix* rateMatrix;
    DoubleMatrix* alphaMatrix;
    
    DoubleVector* stateFreqs;
    DoubleVector* siteRates;
    
    void setUpdateProperties( double mu, int theX, int theY, int model );
    void attachData( DoubleMatrix& CXYs, DoubleMatrix& C1s, DoubleMatrix& rateMatrix, DoubleMatrix& alphaMatrix, DoubleVector& stateFreqs, DoubleVector& siteRates );
    
};

// Main numerics interface
class Numerics {
public:
    Numerics(void);
    
    // Provide a way to instantiate a single numerics object globally
    static Numerics& getGlobalNumerics(void);
    
    // Random number generation
    double Random(void);
    
    // Factorials
    double factorial(int x);
    double factln(double n);
    double gammln(double xx);
    
    // Random variate samplers
    double betaVariate( double a, double b );	
    double gammaVariate( double a );
    double expVariate(void);
    double expVariate (double lambda);
    double gaussianVariate(double m, double s);
    
    // Sample from a discrete distribution
    int _util_sampleFromCumulative_raw( DoubleVector *inVector, int length );
    int _util_sampleFromCumulative_raw( DoubleVector *inVector, int excludeState, int length );
    
    // Slice sampling parameters
    class SliceParameters {
    public:
        double minValue;
        double maxValue;
        double x0, y;
        double w;
        int m;
        
        PriorDistribution *prior;
    };
    

    // Slice sampling functions
    double slice_sampleValueBySlicer( int numSteps, double x0, double (*f)(double), double w, int m );
    void   slice_sampleArrayBySlicer( int n, double x0, double array[], double (*f)(double) );

    double   slice_sampleIndependentRateParameter( PriorDistribution &prior, ModelPropertiesForUpdate& modelProperties, DoubleVector& averageMoveWidths );
    double   slice_sampleIndependentRateParameter( ModelPropertiesForUpdate& modelProperties, DoubleVector& averageMoveWidths );
   
    double   slice_sampleReversibleRateParameter( PriorDistribution &prior, ModelPropertiesForUpdate& modelProperties, DoubleVector& averageMoveWidths );
    
    void slice_sampleReversibleRateParameter( PriorDistribution &prior, int x, int y, double mu, DoubleMatrix& CXYs, DoubleVector& C1s, DoubleMatrix& rateMatrix, DoubleMatrix& alphaMatrix, DoubleVector& siteRates, DoubleVector &stateFreqs );
    
    // Lookup-table based functions
    void   setLUTmode(int val);
    void   precomputeLogTable(double *logLookupTable);
    void   precomputeExpTable(double *expLookupTable);
    double logarithm( double val);
    double exponential(double y);
	
    // Quantile summary (function imported from GSL)
    void   simpleSort( double *list, int listSize);
    double gsl_stats_quantile_from_sorted_data(const double sorted_data[],const size_t stride,const size_t n,const double f);
    
    // B1u probability calculations
    double calculateSubstitutionProbabilityBranch_B1u( const DoubleMatrix &theMatrix, double t, int x, int y);
    double calculateSubstitutionProbabilityBranch_B1u( const DoubleMatrix &theMatrix, double t, int x, int y, double mu );
    double calculateSubstitutionProbabilityBranch_B1u( const DoubleMatrix &theMatrix, double t, int x, int y, double mu, double rateScaleFactor );
    double calculateSubstitutionProbabilityBranch_B1( const DoubleMatrix &theMatrix, double t, int x, int y ) ;
    double calculateSubstitutionProbabilityBranch_B2( const DoubleMatrix &theMatrix, double t, int x, int y, int z ) ;
	
    // B1u likelihood calculation for a single rate parameter and its LambdaX
    double b1u_likelihoodContribution_rateParameter( double lambdaXY, double LambdaX, double C1, double CXY, double mu );
    double b1u_likelihoodContribution_rateParameterOnly( double lambdaXY, double CXY );
    double b1u_likelihoodContribution_rateAwayOnly( double LambdaX, double C1, double mu );
    double b1u_likelihoodContribution_rateAwayOnly( int x, int theModel, double LambdaX, DoubleMatrix* C1s, double mu, DoubleVector* siteRates );
   
    // B1u analytical MLEs and variances
    double MLE_B1u_branchlength( double C1, double C0, double mu);
    double MLE_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu);
    double Variance_B1u_instRate( double LambdaXsp, double C1, double CXY, double mu);
    double Variance_B1u_GivenLambda( double lambdaXY, double LambdaXsp, double C1, double CXY, double mu);
    
    // Proposal evaluation functions (Metropolis)
    int evaluateProposal_logL( double logLikRatio );
    int evaluateProposal_logL( double newLik, double lastLik );
    int evaluateProposal_logL( double newLik, double lastLik, double likHeat, double posteriorHeat );
    int evaluateProposal_logL( double newLik, double lastLik, double proposalRatio, double priorRatio, double likHeat, double posteriorHeat );
	
    // Misc. numeric functions (e.g., for solving linear systems to find stationary freqs)
    int  NR_ludcmp(double *a, int n, int *indx, double *d);
    void NR_lubksb(double *a, int n, int *indx, double b[]);
    void NR_dfpmin(double p[], int n, double gtol, int *iter, double *fret,double(*func)(double [], int np), void (*dfunc)(double [], int, double []));
    void NR_lnsrch(int n, double xold[], double fold, double g[], double p[], double x[],double *f, double stpmax, int *check, double (*func)(double [], int n));
    
    // Functions adapted from PAML for matrix exponentiation
    int  PAML_gradient(int n, double x[], double f0, double g[], double (*fun)(double x[],int n), double space[], int Central);
    void PAML_EigenSort(double d[], double U[], int n);
    int  PAML_PMatUVRoot (DoubleMatrix *P, double t, int n, double U[], double V[], double Root[]);
    int  PAML_eigenQREV (double Q[], double pi[], int n, double Root[], double U[], double V[]);
    
    // Return model index given some model options
    int returnModelIndex( int modelassignment, int rateCatAssignment, int discreteBranchLengthIndex );
    
public:
    int theMode;

private:
    // Mostly private variables for lookup-tables
    int n;
    double min, max, incr;
    double expMin, expMax, expIncr;
    double *logLookupTable;
    double *expLookupTable;
    double logFloor;
    void initLookup(int theN);		
    
    // For caching random number generation
    double *randArray;
    static const int blockSize = 10000000;
    
    // The random number generator
    dsfmt_t dsfmt;
    
    // Private member functions for slice sampling
    void   _slice_findInterval( double &L, double &R, SliceParameters &slice, PartialLogLikelihood &f );
    double _slice_sampleByShrinkage( double L, double R, SliceParameters &slice, PartialLogLikelihood &f );
    
    // Functions adapted from PAML for matrix exponentiation
    int PAML_eigenRealSym(double A[], int n, double Root[], double work[]);
    void PAML_HouseholderRealSym(double a[], int n, double d[], double e[]);
    int PAML_EigenTridagQLImplicit(double d[], double e[], int n, double z[]);
};

#endif