//
/*  ObservedData.cpp
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "ObservedData.h"

ObservedData::ObservedData() {
    // Defaults
    numSites = numTaxa = 0;
    tupleSize = 1;
    numStates = 0;
    
    _currentState = 0;
    _overlappingTuples = false; // no overlapping tuples by default (good for codon models)
    _getStatesFromData = true;  // by default get the state index from the datafile
    _gapsAsCharacters = false;  // do not model gaps as states by default
}

void ObservedData::setNumStates(int val) {
    numStates = val;
}

void ObservedData::setTupleSize(int val) {
    tupleSize = val;
}

int ObservedData::getAlignmentLength(void) {
    return theStates[0].size();
}

void ObservedData::resetAlignmentLength( int val ) {
    // Resize the number of columns in the data matrix
    // NOTE: this is not guaranteed to preserve existing data
    // (only useful when about to simulate a new set of data)
    
    for (int ii=0; ii < getNumSequences(); ii++) {
        theStates[ii].resize( val );
    }
}

int ObservedData::getNumSequences(void) {
    return taxonNames.size();
}

int ObservedData::getNumStates(void) {
    return numStates;
}

string ObservedData::getState(int which) {
	return reverseStateIndex[ which ];
}

string& ObservedData::getTaxonName_ref(int which) {
    return taxonNames[ which ];
}

std::vector<int>& ObservedData::getSequence_ref(int which) {
    return theStates[ which ];
}

bool ObservedData::isAnyGapPresent(int site) {
    return anyGapPresent[site];
}

// --- Internal functions -------------------------------------

bool ObservedData::_hasGap( string theTuple ) {
    // Check if the current tuple has a gap...
    if ( ( theTuple.find('-') == string::npos ) && ( theTuple.find('?') == string::npos ) ) {
        return false;
    } else {
        return true;
    }
}

void ObservedData::_removeAllGapColumns( void ) {
    // This function ALSO sets realDatasetColumn and reverseRealDatasetColumn even if gaps are being modelled...
    int i;
    int origCol, newCol;
    bool removedGaps = false;
    
    cout << "\t-Removing gaps..." << endl;
    
    vector<int> *currentSeq;
    
    // Get number of gapped positions from first sequence
    int gapCount = 0;
    currentSeq = &theStates.at(0);
    for (i=0; i<currentSeq->size(); i++) {
        if ( _excludedPositions[i] ) gapCount++;
    }
        
    // theStates contains all of the sequences...
    for (i=0; i<theStates.size(); i++) {
        currentSeq = &theStates.at(i);
        origCol = newCol = 0;
        
        if (i==0) { 
            cout << "\t- Processed alignment has " << currentSeq->size() << " columns after gapped-column removal (if applicable)" << endl;
        }
        
        cout << "\t\t- Processing sequence " << i+1 << " of " << theStates.size()+1 << "..." << endl;
        
         // Initialize coordinates
        if (i==0) {
            for (std::vector<int>::iterator j = currentSeq->begin(); j != currentSeq->end(); j++, origCol++) {
                if (! _gapsAsCharacters ) {
                    if ( ! _excludedPositions[origCol] ) {
                        // If position not removed, set
                        reverseRealDatasetColumn[origCol] = newCol;
                        realDatasetColumn[newCol++] = origCol;
                    }
                } else {
                    reverseRealDatasetColumn[origCol] = newCol++;
                }
            }
        }
        
        int numRemoved = 0;
        
        
        // This version is INCREDIBLY SLOW!
        // Just create a new vector with the sequence minus gaps
        /*
        // Now erase columns (last column first, so we don't change the numbering as we delete)...
        bool notdone = true;
        origCol = currentSeq->size();
  
        for (std::vector<int>::iterator j = currentSeq->end(); notdone; j--, origCol--) {
            // If gaps are being deleted...
            if (! _gapsAsCharacters ) {
                if ( _excludedPositions[origCol] ) {
                    // Remove
                    removedGaps = true;
                    currentSeq->erase(j);
                    numRemoved++;
                }
            }
        
            // I hate iterators
            if (j == currentSeq->begin()) notdone = false;
        }
        */
        
        int newSize = currentSeq->size() - gapCount;
        
        // New version: construct an entirely new vector
        static vector<int> tempSeq;
        tempSeq.clear();
        if (tempSeq.size() != newSize ) tempSeq.resize( newSize );
      
        // Add
        int curPos = 0;
        for (int j=0; j < currentSeq->size(); j++) {
            // If gaps are being deleted...
            if (! _gapsAsCharacters ) {
                if ( _excludedPositions[j] == 0) {
                    tempSeq.at( curPos ) = ( currentSeq->at( j ) );
                    curPos++;
                } else {
                    // Remove
                    removedGaps = true;
                    numRemoved++;
                }
            }
        }

        // Copy back
        currentSeq->resize( tempSeq.size() );
        for (int j=0; j < currentSeq->size(); j++) {
            currentSeq->at(j) = tempSeq.at(j);
        }

        // Done?
    }
    
    // If we removed gaps, reduce the number of observed states...
    /*
    if ( removedGaps && _getStatesFromData ) {
        numStates--;
        cout << "[ After removing the gap character as a state, the model has " << numStates << " states. ]" << endl << endl;
    }
    */
}

// --- Primary functions -------------------------------------

void ObservedData::usePresetStateIndex(int which) {
    _getStatesFromData = false;
    
    // Confirmed: preset relative rate matrices were mapped to this order
    static const char aas[21] = { 'M','T','N','I','R','K','S','H','P','L','V','E','F','D','A','W','G','C','Q','Y' };
    
    static const char nts[ 5] = { 'T', 'C', 'A', 'G' };
    
    numStates = which;
    const char *whichPreset;
    
    if (numStates == 4) {
        whichPreset = nts;
    } else if (numStates == 20) {
        whichPreset = aas;
    }
    
    stringstream ss;
    for (int i=0; i < numStates; i++) {
        ss.str("");
        ss << whichPreset[i];
        stateIndex[ ss.str() ] = i;
        reverseStateIndex[ i ] = ss.str();
    }
}


void ObservedData::readData( string fileName ) {
    // Read and encode the sequence data from a FASTA formatted file...
    int seqNum = 0;
    
    int totalCount = 0;
    
    ifstream infile( fileName.c_str() );
    
    static int lengthEstablished = 0;
    int        theLength = 0;
    
    static vector<int> currentSeqVector;
    string line;
    string currentSeq;
    
    currentSeq.clear();
    currentSeqVector.clear();
    _excludedPositions.clear();
    
    while (getline( infile, line )) {
            line.erase( std::remove( line.begin(), line.end(), ' ' ), line.end() );

            // Skip blank lines
            if ( line == "" ) { continue; }
            
            if (line.at(0) == '>') {
                line.erase( std::remove( line.begin(), line.end(), '>' ), line.end() );
                
                // New sequence
                taxonNames.push_back( line );
                 numTaxa++;
                
                // If we have a seq, add it...
                if (currentSeq.length() != 0) {
                    currentSeqVector.clear();
                    
                    if ( lengthEstablished == 0 ) {
                        // After reading the first sequence, initialize our global vector
                        theLength = currentSeq.length();
                        anyGapPresent.resize( theLength );
                        for (int ii=0; ii < theLength; ii++) anyGapPresent[ii] = false;
                    }
                    
                    encodeSequence( currentSeq, currentSeqVector );
                    theStates.push_back( currentSeqVector );
                    
                    if (options.useEmpiricalStationary) {
                        if (empiricalFreqs.empty()) {
                            empiricalFreqs.resize(numStates, 0.0);
                        }
                        
                        for (int ii=0; ii < currentSeqVector.size(); ii++) {
                            empiricalFreqs.mDataArray[ currentSeqVector[ ii ] ]++;
                            totalCount++;
                        }
                        
                    }
                    // Set the number of sites from the first sequence
                    if (seqNum == 0) {
                        numSites = currentSeqVector.size();
                    }
                    
                    currentSeq.clear();
                    seqNum++;
                }
                
            } else {
                line.erase( std::remove( line.begin(), line.end(), '\n' ), line.end() );
                currentSeq = currentSeq + line;
            }

        if (infile.eof() ) { break; } 
    }
    
    // Encode the last sequence...
    if (currentSeq.length() != 0) {
        currentSeqVector.clear();

        encodeSequence( currentSeq, currentSeqVector );
  
        if (options.useEmpiricalStationary) {
            for (int ii=0; ii < currentSeqVector.size(); ii++) {
                empiricalFreqs.mDataArray[ currentSeqVector[ ii ] ]++;
                totalCount++;
            }
        }
        
        theStates.push_back( currentSeqVector );
        currentSeq.clear();
        seqNum++;
    }
  
    if (options.useEmpiricalStationary) {
        // This currently uses all data, including columns containing gaps
        // Missing data is not accomodated
        
        for (int ii=0; ii < numStates; ii++) {
            empiricalFreqs.mDataArray[ ii ] /= (totalCount * 1.0);
        }
    }
    
    if ( options.gapsAsMissing == 0) {
        _removeAllGapColumns();
    }
    
    infile.close();
}

void ObservedData::encodeSequence( string characters, vector<int> &encodedStates ) {
    // Encode a sequence to a list of states (integers) using the stateIndex
    // - if  _getStatesFromData, the index is built incrementally
    // - if  _overlappingTuples, all tuples of a given size are built and returned
    // - if !_overlappingTuples, states are harvested as expected (use this for codon models)
    // - an index of gap containing states is updated while processing; need to remove gap columns at the end
    
    int i;
    string currentTuple;
    int incr;
    
    // Set the increment (varies when getting tuples, etc.)
    if ( (tupleSize == 1) || (_overlappingTuples) ) {
        incr = 1;
    } else {
        incr = tupleSize;
    }
    
    encodedStates.clear();
    // For each position...
    for (i=0; i< characters.length(); i+=incr) {
        currentTuple = characters.substr( i, tupleSize );
        
        // If the character state is non-gap (or gaps being modelled) or gaps to be treated as missing...
        if ( !_hasGap( currentTuple ) || (_gapsAsCharacters && !_hasGap( currentTuple ) ) || ( _hasGap( currentTuple) && ( options.gapsAsMissing == 1 ) ) ) {
            
            // Add state to the state index if appropriate...
            if (_getStatesFromData && ( !_hasGap( currentTuple) && !_gapsAsCharacters ) ) {
                if ( stateIndex.find( currentTuple ) == stateIndex.end() ) {
                  // [diagnostic]  cout << "   ** ADDING STATE TO INDEX: " << currentTuple << " = " << _currentState << " **" << endl;
                    stateIndex[ currentTuple ] = _currentState;
                    reverseStateIndex[ _currentState++ ] = currentTuple;
                    numStates++;
                }
            }
            
            if ( _hasGap( currentTuple ) && (options.gapsAsMissing == 1) ) {
                // WARNING: for models with multiple states this sets any tuple with a gap equal to 100% missing
                // (for codon models etc., should instead integrate over individual positions)
                encodedStates.push_back( -1 );
                anyGapPresent[i] = true;
            } else {
                encodedStates.push_back( stateIndex[ currentTuple ] );
            }
        } else if ( !_gapsAsCharacters ) {
            // Mark gapped position for removal
            _excludedPositions[ i ] = true;
            
            // Mark the gapped position in the character string
            encodedStates.push_back( -1 );
        }
    }

    // Debugging code:
    /*
    cout << "Just encoded s equence: " << characters << endl;
    cout << "  as: ";
    for (i=0; i<encodedStates.size(); i++) {
        cout << encodedStates.at(i) << " ";
    } cout << endl;
   */ 
}
