// PLEX.cpp
/*
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// ********************************************************************************************************************************
// Current Version: 0.94
// This is an experimental beta release.

// Please note that this code has a variety of vestigial features from several earlier versions of related code from our lab.
// As a result it may appear to be quite a mess in places.  This is only an illusion: Don't Panic!

// Citation for the program is:
//      de Koning APJ, Gu W, Castoe TA, and DD Pollock (2012).  Phylogenetics,
//      Likelihood, Evolution, and Complexity (PLEX).  Bioinformatics (submitted).

// Citation for the core methods is:
//      de Koning APJ, Gu W, and DD Pollock (2009).  Rapid likelihood analysis on large
//      phylogenies using partial sampling of substitution histories.  Mol. Biol. Evol.

// With the publication of several related manuscripts, new features will be added, and old, experimental code purged.

// For any questions please contact Jason (Jason.de.Koning@gmail.com)
// ********************************************************************************************************************************

#include "ReadInputs.h"
#include "PLEXApp.h"
#include "ObservedData.h"
#include "GTRModel.h"
#include "ProteinModel.h"
#include "MCMC.h"
#include "Numerics.h"
#include "Options.h"
#include "NewickTreeParser.h"
#include "NodeDistanceMaker.h"
#include "Tree.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>
#include <string>
#include <fstream>
#include <iomanip>

#include <sys/param.h>
#include <sys/times.h>
#include <sys/types.h>

using std::cerr;
using std::endl;
using std::ofstream;
using std::setw;
using std::setprecision;
using std::showpos;
using std::noshowpos;
using std::fixed;

// Globally persistent objects
Options options;
Numerics num;

// Global link
Tree *treeLink;

using namespace std;

//---------------------------------------------------------------------------
bool GetTree(
             ifstream&		inTreeStream,
             Tree&	outTree) {
    
	NewickTreeParser	parser;
	parser.SetNodeMaker(new NodeDistanceMaker);
	
	Tree	tree;
    
	bool success = parser.Read(inTreeStream, tree);
	
	if (success)
	{
		tree.MakeRootless();
		tree.swap(outTree);
	}
	
	return success;
} 

//---------------------------------------------------------------------------
void GetTree(
             const char*		m_strSequencefile,
             Tree&	outTree)
{
    const string cFileOpenErr = "Error opening file: ";
    const string cFileReadErr = "Error reading file: ";
    
    PLEXApp* m_oPLEXApp = new(PLEXApp);
	ifstream treeFile(m_strSequencefile);
    
	if (treeFile.is_open())
	{
		//m_oPLEXApp->ReportError( "  * Processing tree file...", false );
		
		if (!GetTree(treeFile, outTree))
			m_oPLEXApp->ReportError(cFileReadErr + m_strSequencefile);
	}
	else 
		m_oPLEXApp->ReportError(cFileOpenErr + m_strSequencefile);
    
//	cerr << endl;
	
	delete m_oPLEXApp;
}


int main( int argc, char *argv[]) {
	int	returnCode = 0;
	
    cout << "*************************************************************************************************" << endl;
    cout << "  PLEX: Phylogenetics, Likelihood, Evolution, and Complexity" << endl;
    cout << "  by A.P. Jason de Koning <jason.de.koning@gmail.com>" << endl;
    cout << endl;
    cout << "  with David D. Pollock. Early contributions by Wanjun Gu." << endl;
    cout << "  Current versions and documentation: http://jasondk.org or http://www.EvolutionaryGenomics.com" << endl;
    cout << "*************************************************************************************************" << endl;
    cout << "[ version 0.92; May 2012 ]" << endl;
    cout << endl;
    
    
    // Set the number of threads
    if (options.numThreads == -1) {
        int nthreads, tid;
       
	/* 
        #pragma omp parallel shared( nthreads ) private( tid )
        {
            tid = omp_get_thread_num();
            
            if (tid == 0) {
                nthreads = omp_get_num_threads();
            }
        }
        
        std::cout << "INITIALIZING NUMBER OF THREADS TO " << nthreads << "." << std::endl;
	*/
	
	options.numThreads = nthreads;
    }
    
    
	// JdK, Sept 18/07: for execution timing...
	time_t startTime;
	startTime = time(NULL);
	
	try {
		int m_nInitseed,  programTask;
		programTask = 0;
        
		ReadInputs* readInputs = new(ReadInputs);
		
		if ( !readInputs->ReadControlfile(CONTROLFILE) )
			exit(1);
		
		// *********************************************
		// Test the Randomize function
		// *********************************************
		PLEXApp* m_oPLEXApp = new(PLEXApp);
		
        m_nInitseed = options.randseed;
       
        //		options.randseed = m_nInitseed;
		programTask = options.task;
		
        // Only one task defined so far...
		if ( programTask == 0) {
            
            options.__currentGen = 0;
            
			// ************************************************************************************************************
			//              Read Control file
			// ************************************************************************************************************
			//if (!readInputs->ReadControlfile(CONTROLFILE) ) {
			//	throw(1);
			//}
			
			// ************************************************************************************************************
            //            Read Input files                
			// ************************************************************************************************************
            ObservedData sequenceData;
            
            // Use preset state index if using fixed model
            if ( options.fixedRates > -1 ) {
                if (options.datatype == 2) {
                    sequenceData.usePresetStateIndex( 20 );
                } else if (options.datatype == 1) {
                    sequenceData.usePresetStateIndex( 4 );
                } else {
                    cout << "ERROR: unknown datatype specified (" << options.datatype << ")" << endl;
                }
            }
           
            string fname( options.m_strSeqfile );
            sequenceData.readData( fname );
            
            int m_nAlignlength = sequenceData.getAlignmentLength();
	
 			// ************************************************************************************************************
			// Read the tree file to get the tree topology
			// ************************************************************************************************************
			
            std::vector<Tree *> Trees;
            Tree tempT;
            //Tree  Tree;
			treeLink = &tempT; //Tree;
            
            // Sampling trees
            if (options.sampleTrees) {
                for (int ii=0; ii<options.treeFiles.size(); ii++) {
                    cout << "Processing tree file #" << ii << ": " << options.treeFiles[ii] << endl;
                    
                    GetTree( options.m_strTreefile,  tempT );
                    
                    Trees.push_back( &tempT );
                }
                
                // exit(1);
            } else {
                
                GetTree( options.m_strTreefile,  tempT );
                Trees.push_back( &tempT );
            }
            //Tree = tempT;
            
			
            //			GetTree( options.m_strTreefile,  *Trees.at(0) );
			
            // We set the # of branches globally here.  (Alternative topologies better have the same #!)
			int m_nLeaves =  (*Trees.at(0)).leaves();
			int m_nInternodes =  (*Trees.at(0)).internalnodes();
			int m_nBranches = options.m_nNumBranches = (*Trees.at(0)).branches();
            (*Trees.at(0)).numRegularNodes = m_nInternodes + m_nLeaves;
			
			// ************************************************************************************************************
			// Optionally subdivide long branches so a maximum length is not exceeded
			// ************************************************************************************************************
            options.__initialNumBranches = m_nBranches;
			std::cout << "\t- Before processing tree: leaves = " << m_nLeaves << ", internal nodes = " << m_nInternodes << ", branches = " << m_nBranches << endl;
			double loss = 0.0;
			
            // DIAGNOSTIC OUTPUT...
            // ofstream outTreeStream2(options.m_strTreeOutfile);
            // outTreeStream2 << "tree gen" << 0 << " = ";
            // (*Trees.at(0)).TreeOutput_PrintBranchLengths( outTreeStream2 );
            
            // Further intializations
            // if (options.allowASRV == 0) options.numCatASRV = 1;
            
			MCMC* m_oMCMC = new(MCMC);
            
            // Initialize sequences, etc. for every tree
            
            for (int ii=0; ii < options.treeFiles.size(); ii++) {
                m_oMCMC->InitializeTree( sequenceData, (*Trees.at(ii)) );
                
                (*Trees.at(ii)).root()->_mixSequences( (*Trees.at(ii)).root()->GetData().sequence, (*Trees.at(ii)).root()->Left()->GetData().sequence, (*Trees.at(ii)).root()->Right()->GetData().sequence );
                // (*Trees.at(ii)).root()->GetData().sequence.Mix(& (*Trees.at(ii)).root()->Left()->GetData().sequence, & (*Trees.at(ii)).root()->Right()->GetData().sequence);
                
                if (options.m_nMethod != 3 && options.m_nMethod != 4  && options.m_nMethod != 5) {
                    // Subdivide all branches so each segment is <= maxBL...
                    
                    // Don't allow subdivision initially...
                    // options.capNumTreeNodes = 1;
                    
                    // SKIPPING SUBDIVISION HERE B/C DEPENDS ON RATE MATRIX
                    // Don't subdivide the tree until burned in a little...
                    // (*Trees.at(ii)).subdivideLongBranches( (*Trees.at(ii)).root(), options.maxBL);
                }
                
                if (options.m_nMethod == 3 || options.m_nMethod == 5) {
                    // Use a branchlength discrete approximation...
                    (*Trees.at(ii)).root()->_mixSequences( (*Trees.at(ii)).root()->GetData().sequence, (*Trees.at(ii)).root()->Left()->GetData().sequence, (*Trees.at(ii)).root()->Right()->GetData().sequence );
                    
                    loss =  (*Trees.at(ii)).discreteBranchLengthApproximation( (*Trees.at(ii)).root());
                } else {
                    (*Trees.at(ii)).MakeRootless();
                }

                if (options.m_nMethod == 3 || (options.m_nMethod == 5)) { std::cout << endl << "TOTAL BRANCHLENGTH TRUNCATION ERROR (RMSE): " << sqrt(loss/m_nBranches) << endl << endl; }
                
                // Shouldn't need this
                m_oMCMC->InitializeTreeAlternate( (*Trees.at(ii)) );
                
                m_nLeaves =  (*Trees.at(ii)).leaves();
                m_nInternodes =  (*Trees.at(ii)).internalnodes();
                m_nBranches =  (*Trees.at(ii)).branches();
                
                std::cout << "\t- After processing tree: leaves = " << m_nLeaves << ", internal nodes = " << m_nInternodes << ", branches = " << m_nBranches << endl;
                std::cout << "\t- Initial branch-length number inflation (i.e., B > B'): " << setprecision(3) << (double)((double)m_nBranches/(double)options.__initialNumBranches) << "X" << " given maximum segment size of " << options.maxBL << endl;
                std::cout << "\t- Complete." << endl << endl;
                // ************************************************************************************************************
              }
			
            // Tree output should be updated to deal with mult trees
			ofstream outTreeStream(options.m_strTreeOutfile);
			ofstream outSubsStream(options.m_strSubOutfile);
			ofstream outSequenceStream(options.m_strSeqOutfile);
			
             // ** [STUFF BELOW HERE NEEDS ADAPTATION TO MULTIPLE TREES] **
            
            int currentTree = 0;
            IntMatrix  TreeChanges;
            // ************************************************************************************************************
            // Count changes on tree -> matrix
            // ************************************************************************************************************
            // count state changes of each branch, and sum over to get the whole tree changes, use this to initialize the model
            (*Trees.at(currentTree)).CountChangesAlongEveryBranch( sequenceData );			
            (*Trees.at(currentTree)).SumChangesOverTree( TreeChanges );
            
            // ************************************************************************************************************
            
            
			// ************************************************************************************************************
			// Model definition and setup
			// ************************************************************************************************************
			
			MixtureModel *m_gtrModel;
			
            
            // Diagnostic check:
            /*
            cout << "       [ Number of states is " << sequenceData.getNumStates() << " ]" << endl;
            for (int ii=0; ii < sequenceData.getNumStates(); ii++ ) {
                cout << "State ii " << sequenceData.getState( ii ) << endl;
            }
            */
           
            cout << "Initializing model, tree, and MCMC objects..." << endl;
            
			if ( (options.m_nDatatype ==2) ) { //(options.doRAR == 1) {
				// Use new protein-model class...
				m_gtrModel = new(ProteinModel);
				
				// Need to provide a vector of initial values for gamma RAR...		
				double *gammaShapes = new double[options.m_nNumModels];
				for (int i=0; i<options.m_nNumModels;i++) {
					gammaShapes[i] = 1.0; // 0.42; // * (Numerics::getGlobalNumerics().Random() * 5.0);
				}
				
				// Basic options...
				(*(ProteinModel *)m_gtrModel).setReversible(0);
				(*(ProteinModel *)m_gtrModel).setFixedFreqs(1);
				(*(ProteinModel *)m_gtrModel).SetNumModels( options.m_nNumModels );
				
				if (options.fixedRates != -1) {
					// For predefined-matrix stuff...
					//	std::cout << "NON-GAMMA MODEL DEFINED..." << endl; exit(1);
					(*(ProteinModel *)m_gtrModel).setRARdistributed(0);
					(*(ProteinModel *)m_gtrModel).setPredefinedMatrix(options.fixedRates);
					(*(ProteinModel *)m_gtrModel).setScaleMatrix(1);
					(*(ProteinModel *)m_gtrModel).setReversible(1);
					(*(ProteinModel *)m_gtrModel).initRateScale(0.03); //0.03  //0.034493);
				} else if (options.doRAR) {
					//	std::cout << "GAMMA MODEL DEFINED..." << endl; exit(1);
					(*(ProteinModel *)m_gtrModel).setRARdistributed(1);
					(*(ProteinModel *)m_gtrModel).setScaleMatrix(0);
				} 
				
				if (options.freeFreqs == 1) {
					(*(ProteinModel *)m_gtrModel).setFixedFreqs(0);					
				}
				
				// For gamma-distributed RAR models...
				
				(*(ProteinModel *)m_gtrModel).setIndependentFractionPrior(options.DRAprior);
				(*(ProteinModel *)m_gtrModel).setDependentRates(options.DRA);
				//				(*(ProteinModel *)m_gtrModel).randomInitRateAssignments();
                
				if ( (*(ProteinModel *)m_gtrModel).RARdistributedRates ) (*(ProteinModel *)m_gtrModel).initGamma3D(options.m_nNumModels, options.numCat, gammaShapes);
				m_oMCMC->InitializeMixtureModel( (*m_gtrModel), TreeChanges );
				
				//	if ( (*(ProteinModel *)m_gtrModel).RARdistributedRates ) (*(ProteinModel *)m_gtrModel).setDistributionParmsFromCounts(  Trees.at(0) );
				
			} else {	// GTR model
				m_gtrModel = new(GTRModel);
                
				m_oMCMC->InitializeMixtureModel( (*m_gtrModel), TreeChanges );
			}
			

			// ************************************************************************************
			// Initialize the ASRV distribution if necessary...
			// ************************************************************************************
			//if ( ( (*Trees.at(0)).root()->GetData().siteCats == NULL) ) {
			// std::cout << "CREATING SITE CATEGORY DATA AT ROOT NODE..." << endl;
            (*Trees.at(currentTree)).root()->GetData().siteCats = new SiteCategory( m_nAlignlength );
			//}
			// ************************************************************************************		
			
			ofstream outCountStream(options.m_strCountOutfile);
			
			if (options.Debugging) {
				if ( (options.m_nDatatype != 2)) {
					(*(GTRModel *)m_gtrModel).PrintCountMatrix(outCountStream, sequenceData);
				} else {
					(*(ProteinModel *)m_gtrModel).PrintCountMatrix(outCountStream, sequenceData);
				}
			}
			ofstream outRatematrixStream(options.m_strMatrixOutfile);
			
			// Jason
			ofstream outSiteAssignmentStream("siteAssignments");
			ofstream outMixtureDependencyStream("mixtureDependencies");
			ofstream outRatesAcrossRatesStream("RARparameters");
//			ofstream posteriorPredictiveStream("posteriorPredictive");
			ofstream freqStream("stateFrequencies");
			ofstream statFreqStream("stationaryFrequencies");
			
            // Print header
			outRatematrixStream << setw(6) << "Gen " << "\tmu\trateSum\ttreeDepth\t"; //treeDepth\t";
            m_oPLEXApp->PrintSequenceStatesLine(outRatematrixStream, sequenceData );	//print the header for rates matrix			
			
            statFreqStream << "gen\tmodel\tmu\t";
            m_oPLEXApp->PrintSequenceStates(statFreqStream, sequenceData );

            freqStream << "gen\tmodel\tmu\t";
            m_oPLEXApp->PrintSequenceStates(freqStream, sequenceData );
            
			// ************************************************************************************************************
			// Run mixture model MCMC 
			// ************************************************************************************************************
			// Force PLEX to use the standard mode
			int m_nMixtureScheme = 1;
            
            int tempCalcProbDist = (*Trees.at(currentTree)).root()->getCalculateFullProbabilityDistribution();
            (*Trees.at(currentTree)).root()->setCalculateFullProbabilityDistribution(1); //override
            (*Trees.at(currentTree)).setprobmatrixalongbranch( (*m_gtrModel), (*m_gtrModel).GetFreqModels() );
            (*Trees.at(currentTree)).root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
			
			if (m_nMixtureScheme == 1) {
				// ************************************************************************************************************
				// Run the Markov chain and save/output state data
				// ************************************************************************************************************
				IntVector m_vecSiteAssignment;
				
				// Doing this for the temporal model, temporarily...
				m_vecSiteAssignment.resize( m_nAlignlength, 0 );
                
                // Prepare MCMC object by pointing it to our data
                m_oMCMC->setPointers( Trees, currentTree, (*(ProteinModel *)m_gtrModel), m_vecSiteAssignment, sequenceData, string("posteriorPredictive") );

                // Initialize empirical frequency distribution
                if (options.useEmpiricalStationary == 1) {
                    if ( m_oMCMC->mModel.returnStationaryFrequencies().empty()) {
                        m_oMCMC->mModel.returnStationaryFrequencies().resize( m_oMCMC->mModel.GetNumModels(), 0 );
                        
                        for (int jj=0; jj<(*m_gtrModel).GetNumModels(); jj++) {
                            if ( m_oMCMC->mModel.returnStationaryFrequencies().mRows.mDataArray[jj].empty()) {
                                m_oMCMC->mModel.returnStationaryFrequencies().mRows.mDataArray[jj].resize( m_oMCMC->mModel.GetNumStates(), 0 );
                            }
                            
                            for (int ii=0; ii < (*m_gtrModel).GetNumStates(); ii++) {
                                m_oMCMC->mModel.returnStationaryFrequencies().mRows.mDataArray[jj].mDataArray[ii] = sequenceData.empiricalFreqs[ii];
                                cout << "   Set state " << ii << " to empirical freq " << sequenceData.empiricalFreqs[ii] << endl;
                                
                            }
                        }
                        
                        //                    (*m_gtrModel).returnStationaryFrequencies().mRows.mDataArray[0].mDataArray[ii] = 0.05; //sequenceData.empiricalFreqs[ii];
                    }
                }
                
                /*
                m_oMCMC->Gibbs_updateAllAncestralAndTransientStates_full( m_oMCMC->mModel, m_oMCMC->theData );
                
				//(*Trees.at(currentTree)).TreeOutput_PrintSequences( outSequenceStream, m_oMCMC->theData );
                //exit(-1);
                
                m_oMCMC->tree_countChanges( m_oMCMC->mModel, m_oMCMC->theData );
				m_oMCMC->Gibbs_updateSiteAssignments_MixtureComponents( m_oMCMC->mModel, m_oMCMC->theData );
				
                */
                
				bool m_bModel, m_bState, m_bAssignment;
				double logLikelihood, m_fMaxLnLikelihood;
				
				m_fMaxLnLikelihood = logLikelihood = 0.0;
				
				ofstream outLikelihoodStream(options.m_strLikelihoodOutfile);
				outLikelihoodStream << "Count\tUpdates\tLnLikelihood\tIntegratedLnLikelihood\tMaxLnLikelihood" << setw(0) << endl;
				
				int m_nUnacceptedModel = 0;
				int m_nUpdateNode = 0;
				int m_nAcceptModel = 0;
				int m_nUpdateAssignment = 0;
				int SampleSequence = 0;
				int maxSampleSequence = 10000;
				
				StringMatrix InternalNodeSeq(m_nInternodes, maxSampleSequence);
				StringVector AncestralNodeSeq(m_nInternodes);
				IntVector NodeID(m_nInternodes);
				
				ofstream ASRVoutfile("siteRateParameters");
				ofstream siteRatesOutfile("siteRates");
				
				siteRatesOutfile << "Gen\t";
				for (int jj=0; jj< m_oMCMC->mTree.root()->GetData().siteCats->siteAssignments.size(); jj++) {
					if (jj != ( m_oMCMC->mTree.root()->GetData().siteCats->siteAssignments.size() - 1) ) {
						siteRatesOutfile << "Site " << (jj+1) << "\t";
					} else {
						siteRatesOutfile << "Site " << (jj+1) << endl;
					}
				}
				
				long int totalUpdates = 0;
                
                cout << "\t- Complete." << endl;
                cout << endl << "*************************************************************************************************" << endl;
                cout << "[ MCMC kernel awake! ]" << endl << endl;
                
                // ************************************************************************************************************
                // MAIN MCMC LOOP HERE...
                // ************************************************************************************************************
                for (unsigned int i = 0; i < options.m_nGenerations; i++) {
					if (options.suppressFileOutput == 0) if ( (i % 1000) == 0 ) { std::cout << "	Generation " << i << endl; }

                    options.treeSampleFreq = 0.002;
                    // To simplify the updater functions, topology-switch proposals run separately from the main RunMarkovChain() functions
                    double u = Numerics::getGlobalNumerics().Random();

                    if ( (options.sampleTrees) && ( u < options.treeSampleFreq ) && (i > options.numGensToReallocateProposals ) ) {
                        i -= options.numGensToReallocateProposals;
                   //     updateCount -= options.numGensToReallocateProposals;
                        
                        // Select among topologies
                        cout << "       [ SAMPLING TOPOLOGY ]" << endl;
                        
                        double starting = m_oMCMC->mTree.incompleteDataLnL;
                        
                        // a) propose which
                        int lastTree = currentTree;
                        currentTree = ( Numerics::getGlobalNumerics().Random() * options.treeFiles.size() ) ;
                        
                        m_oMCMC->currentTree = currentTree;
                        m_oMCMC->mTree = (*m_oMCMC->Trees[currentTree]);
                        
                        cout << "       starting lnL = " << starting << "; chosen tree is " << currentTree << endl;
                        
                        // b) do full marginal ancstate updater; count subs
                        m_oMCMC->Gibbs_updateAllAncestralAndTransientStates_full( m_oMCMC->mModel );
                        m_oMCMC->tree_countChanges( m_oMCMC->mModel );
                        m_oMCMC->Gibbs_updateSiteAssignments_MixtureComponents( m_oMCMC->mModel );

                        cout << "       [ PROPOSAL REALLOCATION STEP ]" << endl;
                        // c) run standard updater for some generations
                        for (int jj=0; jj < options.numGensToReallocateProposals; jj++) {
                            m_oMCMC->RunMarkovChain();
                            totalUpdates += m_oMCMC->numUpdated;
                        }                        
                        // d) accept/reject
                        m_oMCMC->Gibbs_updateAllAncestralAndTransientStates_full( m_oMCMC->mModel );
                        m_oMCMC->tree_countChanges( m_oMCMC->mModel );
                        double ending =  m_oMCMC->mTree.incompleteDataLnL;

                        // compare incompletelnl
                        cout << "       Proposal reallocation gave - new: " << ending << "; old: " << starting << endl;
                                                
                    } else {
                        // Standard updater
                        m_oMCMC->RunMarkovChain();
                        totalUpdates += m_oMCMC->numUpdated;
                    }
                    
                    // Output and update count stuff is below:
					if ((options.suppressFileOutput == 1)  && (i % 1000 == 0)) std::cout << "	REAL GEN " << i << "..." << endl;
					if (options.suppressFileOutput == 1) continue;
					
					if (m_bState) {
						m_nUpdateNode++;
					} else if (m_bModel) {
						m_nAcceptModel++;
					} else if (m_bAssignment) {
						m_nUpdateAssignment++;
					} else
						m_nUnacceptedModel++;
					
					if ( i % options.outputFrequency == 0 ) { 			
						//output the count matrix of total changes in different states
                        TreeChanges.fill(0);
                        m_oMCMC->mTree.SumChangesOverTree( TreeChanges);
						
                        m_oPLEXApp->PrintMatrix(outCountStream, TreeChanges, m_oMCMC->theData );
						
                        if ( m_oMCMC->mTree.root()->GetData().tempTreeDepth.empty() ) {
                            m_oMCMC->mTree.root()->GetData().tempTreeDepth.resize( m_oMCMC->mModel.GetNumModels(), 0.0 );
                        }
                        
                        //DoubleVector* treeDepth = &m_oMCMC->mTree.getRoot()->GetData().tempTreeDepth;
                        
						outRatematrixStream << setw(2) << i << "\t" << setprecision(5) << m_oMCMC->mModel.returnUniformizationConstants()->mDataArray[0] << "\t" << setprecision(5) << ( m_oMCMC->mModel.getRateSum( m_oMCMC->mTree ) / m_oMCMC->mModel.GetNumModels() ) << "\t" << m_oMCMC->mTree.getRoot()->GetData().treeLength << "\t";
                    
						if (options.m_nDatatype == 2) {
							m_oMCMC->mModel.PrintRateParameters(outRatesAcrossRatesStream);	// only applies to protein models so far
						}
						m_oMCMC->mModel.PrintRateMatrix3D(outRatematrixStream, m_oMCMC->theData);
                        
                        if (options.freeFreqs) {
                            freqStream << i << "\t";
                            m_oMCMC->mModel.PrintFrequencyMatrix(freqStream, m_oMCMC->theData);
                        }
                        
                        // Stationary frequency output
                        if ( !( (options.freeFreqs) && (options.reversibleModel) ) ) {

                            // Force recalculation of the stationary freqs when output is required...
                            if (options.useEmpiricalStationary == 0) m_oMCMC->mModel.forceUpdateStationaryFreqs();
                            
                            statFreqStream << i << "\t";
                            m_oMCMC->mModel.PrintStationaryFreqMatrix(statFreqStream, m_oMCMC->theData);
                        }
                        
                        if (options.m_nNumModels > 1) m_oMCMC->mModel.PrintModelAssignments(outSiteAssignmentStream, m_oMCMC->m_vecSiteAssignment);
						
						if (options.doRAR == 1) {
							if (m_oMCMC->mModel.getDependentRates() == 1) m_oMCMC->mModel.PrintDependentRates(outMixtureDependencyStream);
						}
						
							
						// Output the likelihood with current rate matrix and count changes:				
						// CHECK THIS: SHOULDN'T NEED TO RECALCULATE THE COMPLETE_DATA LOG-L HERE
						
                        logLikelihood =  m_oMCMC->mTree.completeDataLnL;
						
						if (m_nAcceptModel <= 1)
							m_fMaxLnLikelihood = logLikelihood;
						else {
							if (m_fMaxLnLikelihood < logLikelihood)
								m_fMaxLnLikelihood = logLikelihood;
						}
						
						outLikelihoodStream << i << "\t" << setw(11) << totalUpdates << "\t" << setprecision(4) << fixed << showpos << logLikelihood << "\t" <<  m_oMCMC->mTree.incompleteDataLnL << "\t" << m_fMaxLnLikelihood << setprecision(0) << setw(0) << noshowpos << endl;
						
  						// Output tree with branchlengths...
						if ( (i==0) || ( (options.updatebls) &&  (options.outputTrees) && ( (i % options.outputFrequency) == 0) ) ) {
                            outTreeStream << "tree gen" << i << " = ";
                            m_oMCMC->mTree.TreeOutput_PrintBranchLengths( outTreeStream );
                        }

                        // Output substitutions...
                        if (options.outputSubs) {
                            if ( (i % options.outputSubFrequency) == 0 ) {
                                // Output complete sequences at the same freq as substitutions...   
                                if (options.outputSequences) m_oMCMC->mTree.TreeOutput_PrintSequences( outSequenceStream, m_oMCMC->theData );
                                outSequenceStream << "//" << endl;
                                
                                if (options.outputSubType == 0) {
                                    // Output subs on tree (end point only)
                                    m_oMCMC->mTree.TreeOutput_PrintBranchChangeList( outSubsStream, m_oMCMC->theData );
                                } else if ( options.outputSubType == 1) {
                                    // Output subs in table (end point only)
                                    m_oMCMC->mTree.TreeOutput_PrintBranchChangeListTable( outSubsStream, m_oMCMC->theData );
                                    outSubsStream << "//" << endl;
                                } else if ( options.outputSubType == 2) {
                                    // Output subs in table (summed over branch)
                                    m_oMCMC->mTree.TreeOutput_PrintBranchChangeListTable_Sum( outSubsStream, m_oMCMC->theData );
                                    outSubsStream << "//" << endl;
                                } else if ( options.outputSubType == 3) {
                                    m_oMCMC->mTree.TreeOutput_PrintBranchChange_SiteSpecificList( outSubsStream, m_oMCMC->theData );
                                    outSubsStream << "//" << endl;
                                } else if ( options.outputSubType == 4) {
                                    m_oMCMC->mTree.TreeOutput_PrintBranchChange_SiteSpecificList_sum( outSubsStream, m_oMCMC->theData );
                                    outSubsStream << "//" << endl;
                                }
                            }
                        }
                        
						// Any type of update...
						if (options.allowASRV) {
							if ( i % options.outputFrequency == 0 ) {
								ASRVoutfile << i << "\tGamma shape:\t" <<  m_oMCMC->mTree.root()->GetData().siteCats->getShapeParm() << "\tRates:\t";
								for (int ii=0; ii<options.numCatASRV; ii++) {
									ASRVoutfile <<  m_oMCMC->mTree.root()->GetData().siteCats->getRate(ii) << "\t";
								}
								
								if (options.outputSiteRates) {
									ASRVoutfile << "Num sites per class:\t";
									static int *tempCounts;
									static int countAlloc = 0;
									if (countAlloc == 0) {
										tempCounts = new int[options.numCatASRV];
										countAlloc = 1;
									}
									for (int ii=0; ii<options.numCatASRV; ii++) {
										tempCounts[ii] = 0;
									}
									
									siteRatesOutfile << i << "\t";
									for (int jj=0; jj< m_oMCMC->mTree.root()->GetData().siteCats->siteAssignments.size(); jj++) {
										tempCounts[  m_oMCMC->mTree.root()->GetData().siteCats->getAssignment(jj) ]++;
										siteRatesOutfile <<  m_oMCMC->mTree.root()->GetData().siteCats->getRate(  m_oMCMC->mTree.root()->GetData().siteCats->getAssignment(jj) ) << "\t";
									} siteRatesOutfile << endl;
									
									for (int ii=0; ii<options.numCatASRV; ii++) {
										ASRVoutfile << tempCounts[ii] << "\t";
									} ASRVoutfile << endl;
								} else {
									ASRVoutfile << endl;
								}
								
							}
						}
					}			
				}
                
				if (options.Debugging)
				{
					outLikelihoodStream << flush;
					
					//std::cout << setw(40) << "Number of Updated Node" << setw(40) << "Number of Accepted Proposal Model" << setw(40) << "Number of Unaccepted Proposal Model" << setw(0) << endl;
					//std::cout << setw(40) << m_nUpdateNode << setw(40) << m_nAcceptModel << setw(40) << m_nUnacceptedModel << setw(0) << endl;
					std::cout << flush;
				}
			}
			
		}
    }
    
	
	// ************************************************************************************************************
	// Error handling and reporting
	// ************************************************************************************************************
    
	catch(int newreturnCode) {
		returnCode = newreturnCode;
		
		cerr << endl << endl
		<< "**********************************" <<endl
		<< " ------------- ERROR ------------ " << endl
		<< endl <<endl;
		
		if (newreturnCode == 1)
			cerr << "error " << returnCode << ": " 
			<< "Open controlfile error" << endl;
		else if (newreturnCode == 2)
			cerr << "error " << returnCode << ": " 
			<< "Read sequence file error" << endl;
		else if (newreturnCode == 3)
			cerr << "error " << returnCode << ": " 
			<< "Read position file error" << endl;
		else if (newreturnCode == 4)
			cerr << "error " << returnCode << ": " 
			<< "Read OL file error" << endl;
		else if (newreturnCode == 5)
			cerr << "error " << returnCode << ": " 
			<< "Read genome size file error" << endl;
		else if (newreturnCode == 6)
			cerr << "error " << returnCode << ": " 
			<< "Not all sequence position information inputted" << endl;
		else if (newreturnCode == 7)
			cerr << "error " << returnCode << ": " 
			<< "Not all OL or size information inputted" << endl;
		else if (newreturnCode == 8)
			cerr << "error " << returnCode << ": " 
			<< "Read tree file error" << endl;
		else if (newreturnCode == 9)
			cerr << "error " << returnCode << ": " 
			<< "Method to calculate probability is error" << endl;
		
		cerr << endl << endl
		<< "**********************************" <<endl
		<< endl <<endl <<endl;
		
		exit(returnCode);
	}
	
	catch(...) {
		cerr << "**********************" << endl;
		cerr << "Unknown error" << endl;
		returnCode = -1;
		exit(returnCode);
	}
	
	// ************************************************************************************************************
	// Calculation of timetaken
	// ************************************************************************************************************
	
	time_t theT;
	theT = time(NULL)-startTime;
	int h,m,s;
	h=theT/3600; m=(theT%3600)/60; s=theT-(theT/60)*60;
	printf("\n\nTime elapsed: ");
	if(h)  printf("%d:%02d:%02d", h,m,s);
	else   printf("%2d:%02d", m,s); 
	ofstream outTimeStream("timetakenfile");
	if (h) outTimeStream << h << ":" << m << ":" << s << endl;
	else outTimeStream << m << ":" << s << endl;
   	printf("\n"); 
	// ************************************************************************************************************
	
	return returnCode;
}
