/* PartialLogLikelihood.cpp
PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
version 0.94 - beta version  July, 2012

Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
jason.de.koning@gmail.com http://jasondk.org
wanjungu@gmail.com
David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com

This program was created and distributed by the
David Pollock laboratory, University of Colorado School of Medicine
copyright © 2012 David Pollock
All rights reserved.

http://www.evolutionarygenomics.com
Dept. Biochemistry & Molecular Genetics
12801 17th Ave., MS 8101
PO Box 6511
Aurora, CO 80045

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <limits>
#include "PartialLogLikelihood.h"
#include "Numerics.h"
#include "Options.h"

extern Options options;

// *****************************************************************************************************************************
//  Interface to callable partial likelihood functions that use the Numerics:: functions
// *****************************************************************************************************************************

PartialLogLikelihood_individualRateParm::PartialLogLikelihood_individualRateParm( void ) { }

PartialLogLikelihood::PartialLogLikelihood( void ) {
    num = &Numerics::getGlobalNumerics();
}

PartialLogLikelihood::PartialLogLikelihood( int xInit, int yInit, double muInit, double C1init, double CXYinit, DoubleMatrix& rateMatrix, DoubleMatrix& alphaMatrix, DoubleVector& stateFreqsInit  ) {
    x = xInit;
    y = yInit;
    mu = muInit;
    C1 = C1init;
    CXY = CXYinit;

    rates = &rateMatrix;
    alpha = &alphaMatrix;
    stateFreqs = &stateFreqsInit;
    
    num = &Numerics::getGlobalNumerics();
}

void PartialLogLikelihood::setParms( ModelPropertiesForUpdate& modelProperties ) {
    x        = modelProperties.x;
    y        = modelProperties.y;
    mu       = modelProperties.mu;
    theModel = modelProperties.theModel;
    
    C1s      = modelProperties.C1s;
    CXYs     = modelProperties.CXYs;
    
    CXY      = modelProperties.CXYs->mRows.mDataArray[ x ]. mDataArray[ y ];
    C1       = modelProperties.C1s->mRows.mDataArray[ theModel ]. mDataArray[ x ];
    
    rates       = modelProperties.rateMatrix;
    alpha       = modelProperties.alphaMatrix;
    stateFreqs  = modelProperties.stateFreqs;
    siteRates   = modelProperties.siteRates; 
}

void PartialLogLikelihood::setBounds( double theMin, double theMax ) {
    max = theMax;
    min = theMin;
}

void PartialLogLikelihood::setModel( int model ) {
    theModel = model;
}

double PartialLogLikelihood_individualRateParm::logL( double x1 )  {
    // Expects: argument x1 is the rate parameter (unscaled by frequency if applicable)
    double theValue = 0.0;
    double xInitial;
    
    xInitial = (*rates).mRows.mDataArray[x].mDataArray[y];
    
    if (x1 < min ) return -std::numeric_limits<double>::max();
    if (x1 > max ) return -std::numeric_limits<double>::max();
   
    if (options.scaleByFrequencies != 1) {
        if ( options.allowASRV ) {
            theValue += num->b1u_likelihoodContribution_rateParameterOnly( x1, CXY );
            theValue += num->b1u_likelihoodContribution_rateAwayOnly( x, theModel, (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + x1), C1s, mu, siteRates );
            return theValue;
        } else {
            return num->b1u_likelihoodContribution_rateParameter( x1, (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + x1), C1, CXY, mu);
        }
    } else {
        if ( options.allowASRV ) {
            theValue += num->b1u_likelihoodContribution_rateParameterOnly( (x1 * (*stateFreqs)[y]), CXY );
            theValue += num->b1u_likelihoodContribution_rateAwayOnly( x, theModel, (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + (x1 * (*stateFreqs)[y])), C1s, mu, siteRates );
            return theValue;
        } else {
            return num->b1u_likelihoodContribution_rateParameter( (x1 * (*stateFreqs)[y]), (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + (x1 * (*stateFreqs)[y])), C1, CXY, mu);        
        }
    }
}

PartialLogLikelihood_reversibleRateParm::PartialLogLikelihood_reversibleRateParm( int xInit, int yInit, double muInit, double C1init, double CXYinit, double C1init_a, double CXYinit_a, DoubleMatrix& rateMatrixInit, DoubleMatrix& alphaMatrixInit, DoubleVector& stateFreqsInit ) {
 
    x = xInit;
    y = yInit;
    mu = muInit;
    C1 = C1init;
    CXY = CXYinit;
    C1a = C1init_a;
    CXYa = CXYinit_a;

    rates = &rateMatrixInit;
    alpha = &alphaMatrixInit;
    stateFreqs = &stateFreqsInit;
    
    num = &Numerics::getGlobalNumerics();
}

double PartialLogLikelihood_reversibleRateParm::logL( double x1 )  {
    // Expects: argument x1 is the rate parameter (unscaled by frequency if applicable)
    
    double xInitial;
    double yInitial;
    
    if (x1 < min ) return -std::numeric_limits<double>::max();
    if (x1 > max ) return -std::numeric_limits<double>::max();
    
    /*
    if ( options.scaleByFrequencies ) {
        // Need to keep the frequencies fixed and use the unscaled 
        xInitial = (*rates).mRows.mDataArray[x].mDataArray[y];
        yInitial = (*rates).mRows.mDataArray[y].mDataArray[x];
        
        return ( num->b1u_likelihoodContribution_rateParameter( (x1 * (*stateFreqs)[ y ]), (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + (x1 * (*stateFreqs)[ y ])), C1,  CXY,  mu) + 
                 num->b1u_likelihoodContribution_rateParameter( (x1 * (*stateFreqs)[ x ]), (-(*rates).mRows.mDataArray[y].mDataArray[y] -yInitial + (x1 * (*stateFreqs)[ x ])), C1a, CXYa, mu) );
    } else {
        xInitial = (*rates).mRows.mDataArray[x].mDataArray[y];
        yInitial = (*rates).mRows.mDataArray[y].mDataArray[x];
        
        return ( num->b1u_likelihoodContribution_rateParameter( x1, (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + x1), C1,  CXY,  mu) + 
                num->b1u_likelihoodContribution_rateParameter(  x1, (-(*rates).mRows.mDataArray[y].mDataArray[y] -yInitial + x1), C1a, CXYa, mu) );

    }
     */
    double theValue = 0.0;
    
    xInitial = (*rates).mRows.mDataArray[x].mDataArray[y];
    yInitial = (*rates).mRows.mDataArray[y].mDataArray[x];

    if (options.scaleByFrequencies != 1) {
        if ( options.allowASRV ) {
            theValue += num->b1u_likelihoodContribution_rateParameterOnly( x1, CXYs->mRows.mDataArray[x].mDataArray[y] );
            theValue += num->b1u_likelihoodContribution_rateAwayOnly( x, theModel, (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + x1), C1s, mu, siteRates );

            theValue += num->b1u_likelihoodContribution_rateParameterOnly( x1, CXYs->mRows.mDataArray[y].mDataArray[x] );
            theValue += num->b1u_likelihoodContribution_rateAwayOnly( y, theModel, (-(*rates).mRows.mDataArray[y].mDataArray[y] -yInitial + x1), C1s, mu, siteRates );
        } else {
            theValue += num->b1u_likelihoodContribution_rateParameter( x1, (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + x1), C1s->mRows.mDataArray[theModel].mDataArray[x], CXYs->mRows.mDataArray[x].mDataArray[y], mu);

            theValue += num->b1u_likelihoodContribution_rateParameter( x1, (-(*rates).mRows.mDataArray[y].mDataArray[y] -yInitial + x1), C1s->mRows.mDataArray[theModel].mDataArray[y], CXYs->mRows.mDataArray[y].mDataArray[x], mu);
        }
    } else {
        if ( options.allowASRV ) {
            theValue += num->b1u_likelihoodContribution_rateParameterOnly( (x1 * (*stateFreqs)[y]), CXYs->mRows.mDataArray[x].mDataArray[y] );
            theValue += num->b1u_likelihoodContribution_rateAwayOnly( x, theModel, (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + (x1 * (*stateFreqs)[y])), C1s, mu, siteRates );

            theValue += num->b1u_likelihoodContribution_rateParameterOnly( (x1 * (*stateFreqs)[x]), CXYs->mRows.mDataArray[y].mDataArray[x] );
            theValue += num->b1u_likelihoodContribution_rateAwayOnly( y, theModel, (-(*rates).mRows.mDataArray[y].mDataArray[y] -yInitial + (x1 * (*stateFreqs)[x])), C1s, mu, siteRates );
        } else {
            theValue += num->b1u_likelihoodContribution_rateParameter( (x1 * (*stateFreqs)[y]), (-(*rates).mRows.mDataArray[x].mDataArray[x] -xInitial + (x1 * (*stateFreqs)[y])), C1s->mRows.mDataArray[theModel].mDataArray[x], CXYs->mRows.mDataArray[x].mDataArray[y], mu);

            theValue += num->b1u_likelihoodContribution_rateParameter( (x1 * (*stateFreqs)[x]), (-(*rates).mRows.mDataArray[y].mDataArray[y] -yInitial + (x1 * (*stateFreqs)[x])), C1s->mRows.mDataArray[theModel].mDataArray[y], CXYs->mRows.mDataArray[y].mDataArray[x], mu);
        }
    }

    return theValue;
}

PartialLogLikelihood_reversibleRateParm::PartialLogLikelihood_reversibleRateParm( void ) {
    num = &Numerics::getGlobalNumerics();
}
