/*
 *  PriorDistributions.cpp Created for PLEXlite by A.P. Jason de Koning on 1/25/12.
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <vector>
#include "PriorDistributions.h"
#include "prob.h"
#include "math.h"
#include "Numerics.h"

PriorDistribution::PriorDistribution(void) {
    num = &Numerics::getGlobalNumerics();
}

void PriorDistribution::setParameters( std::vector<double> &parms ) {
    hyperParameters.resize( parms.size() );
    
    for (int i=0; i<parms.size(); i++) {
        hyperParameters.at(i) = parms.at(i);
    }
}

double PriorDistribution::evaluateLogPDF( double x ) {
    return num->logarithm( evaluatePDF( x ) );
}

// Prior distributions and pdf evaluation functions
UniformPrior::UniformPrior(void) {
    // Parameters are a,b
    hyperParameters.resize( 0 );
}

double UniformPrior::evaluatePDF( double x, std::vector<double> &parms ) {
    return 1.0;
}

double UniformPrior::evaluatePDF( double x ) {
    return 1.0;
}

double UniformPrior::evaluateLogPDF( double x ) {
    return 0.0;
}

BetaPrior::BetaPrior(void) {
    // Parameters are a,b
    hyperParameters.resize( 2 );
}

BetaPrior::BetaPrior(double a, double b) {
    // Parameters are a,b
    hyperParameters.resize( 2 );
    
    hyperParameters[0] = a;
    hyperParameters[1] = b;
}

double BetaPrior::evaluatePDF( double x, std::vector<double> &parms ) {
    // Ignore the normalization constant, fast version
    double prob = pow( x, (parms[0] - 1)) * pow((1 - x), (parms[1]-1));
     
    // double prob = beta_pdf( x, parms[0], parms[1] );
    return prob <= 0.0 ? 1e-30 : prob;
}

double BetaPrior::evaluatePDF( double x ) {
    // Ignore the normalization constant, fast version
    double prob = pow( x, (hyperParameters[0] - 1)) * pow((1 - x), (hyperParameters[1]-1));
    
    // Full version (including normalization constant) - slow
    //double prob = beta_pdf( x, hyperParameters[0], hyperParameters[1] );
    
    return prob <= 0.0 ? 1e-30 : prob;
}

double BetaPrior::evaluateLogPDF( double x ) {
    // Fast version already logged
    double prob = (hyperParameters[0]-1) * num->logarithm(x) + (hyperParameters[1]-1) * num->logarithm(1-x);

    return prob;
}

GammaPrior::GammaPrior(void) {
    // Parameters are alpha, beta (shape and scale)
    hyperParameters.resize( 2 );
}

GammaPrior::GammaPrior(double a, double b) {
    // Parameters are a,b
    hyperParameters.resize( 2 );
    
    hyperParameters[0] = a;
    hyperParameters[1] = b;
}

double GammaPrior::evaluatePDF( double x, std::vector<double> &parms ) {
    // We use shape (alpha) and scale (beta) parameterizations
    double prob = gamma_pdf( x, 0.0, 1.0 / parms[1], parms[0] );

    return prob <= 0.0 ? 1e-30 : prob;
}

double GammaPrior::evaluatePDF( double x ) {
    double prob = gamma_pdf( x, 0.0, 1.0 / hyperParameters[1], hyperParameters[0] );

    return prob <= 0.0 ? 1e-30 : prob;
}

ExpPrior::ExpPrior(void) {
    // Parameters are lambda
    hyperParameters.resize( 1 );
}

ExpPrior::ExpPrior(double lambda) {
    // Parameters are lambda
    hyperParameters.resize( 1 );
    
    hyperParameters[0] = lambda;
}

double ExpPrior::evaluatePDF( double x, std::vector<double> &parms ) {
    // parameter b is beta = 1/lambda
    double prob = exponential_pdf( x, 0, (1.0 / parms[0]) );

    return prob <= 0.0 ? 1e-30 : prob;
}

double ExpPrior::evaluatePDF( double x ) {
    // parameter b is beta = 1/lambda
    double prob = exponential_pdf( x, 0, (1.0 / hyperParameters[0]) );

    return prob <= 0.0 ? 1e-30 : prob;
}

double ExpPrior::evaluateLogPDF( double x ) {
    double prob = -hyperParameters[0] * x + num->logarithm(x);
    return prob;
}

GaussianPrior::GaussianPrior(void) {
    // Parameters are mu, sigma
    hyperParameters.resize( 2 );
}

GaussianPrior::GaussianPrior( double mu, double sigma ) {
    // Parameters are mu, sigma
    hyperParameters.resize( 2 );
    
    hyperParameters[0] = mu;
    hyperParameters[1] = sigma;
}

double GaussianPrior::evaluatePDF( double x, std::vector<double> &parms ) {
    double prob = normal_pdf( x, parms[0], parms[1] );

    return prob <= 0.0 ? 1e-30 : prob;
}

double GaussianPrior::evaluatePDF( double x ) {
    double prob = normal_pdf( x, hyperParameters[0], hyperParameters[1] );

    return prob <= 0.0 ? 1e-30 : prob;
}
