/*
 *  RecycleMatrix.h
 *  Originally created by Chip Jarred for David Pollock's CLAMP/PLEX program
 *  Copyright (c) 2004 Dr. David Pollock. All rights reserved.
 
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *
 * recycle the memory cost for matrix, not used in current release
 *
 */

#ifndef RecycleMatrix_h__
#define RecycleMatrix_h__

#include <algorithm>
#include "KeyedRecyclery.h"
#include "Matrix.h"

#pragma mark -
//---------------------------------------------------------------------------
template <typename DataType>
class RecycleMatrix : public AbstractMatrix<DataType>
{
	public:
			typedef	typename AbstractMatrix<DataType>::RowType		RowType;
			typedef	typename AbstractMatrix<DataType>::RowVector	RowVector;
			typedef	typename AbstractMatrix<DataType>::size_type	size_type;
			
											RecycleMatrix(void);
											RecycleMatrix(
												size_type			inRows,
												size_type			inColumns,
												const DataType&		inValue = DataType());
											RecycleMatrix(
												const RecycleMatrix&		inRecycleMatrix);
												
	virtual									~RecycleMatrix(void);
												
	virtual	void							fill(
												const DataType&		inVal);
			
	virtual	void							resize(
												size_type			inRows,
												size_type			inColumns,
												const DataType&		inVal = DataType());
												
	virtual	const RowType&					operator[](
												size_type			inRowIndex) const;
	virtual	RowType&						operator[](
												size_type			inRowIndex);
												
			const RecycleMatrix&			operator=(
												const RecycleMatrix&		inRecycleMatrix);
			const RecycleMatrix&			operator=(
												DataType**			inRecycleMatrix);
												
	virtual	void							SwapRows(
												size_type			inRow1,
												size_type			inRow2);
	virtual	void							SwapColumns(
												size_type			inColumn1,
												size_type			inColumn2);
												
			void							swap(
												RecycleMatrix&				inRecycleMatrix);
															
	private:
		typedef typename Matrix<DataType>::Initializer Initializer;
		
				Matrix<DataType>*						mMatrix;
				Initializer								mInitializer;
				
		typedef KeyedRecyclery<Matrix<DataType>, Initializer>	RecycleryType;
		static	RecycleryType							sRecyclery;
};

template <typename DataType>
typename RecycleMatrix<DataType>::RecycleryType	RecycleMatrix<DataType>::sRecyclery;

#pragma mark -
//---------------------------------------------------------------------------
template <typename DataType>
RecycleMatrix<DataType>::RecycleMatrix(void)
	:	AbstractMatrix<DataType>(),
		mMatrix(0),
		mInitializer(0, 0, DataType())
{
	mMatrix = sRecyclery.New(mInitializer);
}

//---------------------------------------------------------------------------
template <typename DataType>
RecycleMatrix<DataType>::RecycleMatrix(
	size_type		inRows,
	size_type		inColumns,
	const DataType&	inValue)
	:	AbstractMatrix<DataType>(inRows, inColumns, inValue),
		mMatrix(0),
		mInitializer(inRows, inColumns, inValue)
{
	mMatrix = sRecyclery.New(mInitializer);
}

//---------------------------------------------------------------------------
template <typename DataType>
RecycleMatrix<DataType>::RecycleMatrix(
	const RecycleMatrix&	inRecycleMatrix)
	:	AbstractMatrix<DataType>(inRecycleMatrix),
		mMatrix(0),
		mInitializer(inRecycleMatrix.mInitializer)
{
	mMatrix = sRecyclery.New(mInitializer);
	*mMatrix = *(inRecycleVector.mMatrix);
}

//---------------------------------------------------------------------------
template <typename DataType>
RecycleMatrix<DataType>::~RecycleMatrix(void)
{
	sRecyclery.Dispose(mInitializer, mMatrix);
}

//---------------------------------------------------------------------------
template <typename DataType>
void
RecycleMatrix<DataType>::fill(
	const DataType&	inVal)
{
	mMatrix->fill(inVal);
}

//---------------------------------------------------------------------------
template <typename DataType>
void
RecycleMatrix<DataType>::resize(
	size_type		inRows,
	size_type		inColumns,
	const DataType&	inVal)
{
	mMatrix->resize(inRows, inColumns, inVal);
	mInitializer.fRows	= inRows;
	mInitializer.fCols	= inColumns;
	mInitializer.fInitValue	= inVal;
	AbstractMatrix<DataType>::resize(inRows, inColumns, inVal);
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
const typename RecycleMatrix<DataType>::RowType&
RecycleMatrix<DataType>::operator[](
	size_type	inRowIndex) const
{
	return (*mMatrix)[inRowIndex];
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
typename RecycleMatrix<DataType>::RowType&
RecycleMatrix<DataType>::operator[](
	size_type	inRowIndex)
{
	return (*mMatrix)[inRowIndex];
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
const RecycleMatrix<DataType>&
RecycleMatrix<DataType>::operator=(
	const RecycleMatrix&	inRecycleMatrix)
{
	RecycleMatrix	temp = inRecycleMatrix;
	
	swap(temp);
	return *this;
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
void
RecycleMatrix<DataType>::SwapRows(
	size_type	inRow1,
	size_type	inRow2)
{
	mMatrix->SwapRows(inRow1, inRow2);
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
void
RecycleMatrix<DataType>::SwapColumns(
	size_type	inColumn1,
	size_type	inColumn2)
{
	mMatrix->SwapColumns(inColumn1, inColumn2);
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
void
RecycleMatrix<DataType>::swap(
	RecycleMatrix&	inRecycleMatrix)
{
	std::swap(inRecycleMatrix.mInitializer, mInitializer);
	std::swap(inRecycleMatrix.mMatrix, mMatrix);
	AbstractMatrix<DataType>::swap(inRecycleMatrix);
}

//---------------------------------------------------------------------------
#endif // RecycleMatrix_h__
