/*
 *  RecycleVector.h
 *  Originally created by Chip Jarred for David Pollock's CLAMP/PLEX program
 *  Copyright (c) 2004 Dr. David Pollock. All rights reserved.
 
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *
 * recycle the memory cost for vector, not used in current release
 *
 */

#ifndef RecycleVector_h__
#define RecycleVector_h__

#include <algorithm>
#include <cassert>

#include "KeyedRecyclery.h"
#include "Vector.h"

//---------------------------------------------------------------------------
template <typename DataType>
class RecycleVector : public AbstractVector<DataType>
{
	public:
		typedef typename AbstractVector<DataType>::size_type	size_type;
		
											RecycleVector(void);
											RecycleVector(
												size_type		inNumItems,
												const DataType&	inVal = DataType());
											RecycleVector(
												const RecycleVector&	inRecycleVector);
		virtual								~RecycleVector(void);
											
				const RecycleVector&				operator=(
												const RecycleVector&	inRecycleVector);
												
		virtual	void						resize(
												size_type		inNewSize,
												const DataType&	inVal = DataType());
		virtual	void						fill(
												const DataType&	inVal);
												
		virtual	DataType&					operator[](
												size_type		inIndex);
		virtual	const DataType&				operator[](
												size_type		inIndex) const;
												
				void						swap(
												RecycleVector&			ioRecycleVector);
	private:
		typedef typename Vector<DataType>::Initializer Initializer;
		
				Vector<DataType>*						mVector;
				Initializer								mInitializer;
				
		typedef KeyedRecyclery<Vector<DataType>, Initializer>	RecycleryType;
		static	RecycleryType							sRecyclery;
};

template <typename DataType>
typename RecycleVector<DataType>::RecycleryType	RecycleVector<DataType>::sRecyclery;

#pragma mark -
//---------------------------------------------------------------------------
template <typename DataType>
RecycleVector<DataType>::RecycleVector(void)
	:	AbstractVector<DataType>(),
		mVector(0),
		mInitializer(0, DataType())
{
	mVector = sRecyclery.New(mInitializer);
}

//---------------------------------------------------------------------------
template <typename DataType>
RecycleVector<DataType>::RecycleVector(
	size_type		inNumItems,
	const DataType&	inVal)
	:	AbstractVector<DataType>(inNumItems, inVal),
		mVector(0),
		mInitializer(inNumItems, inVal)
{
	mVector = sRecyclery.New(mInitializer);
}

//---------------------------------------------------------------------------
template <typename DataType>
RecycleVector<DataType>::RecycleVector(
	const RecycleVector&	inRecycleVector)
	:	AbstractVector<DataType>(inRecycleVector),
		mVector(0),
		mInitializer(inRecycleVector.mInitializer)
{
	mVector = sRecyclery.New(mInitializer);
	*mVector = *(inRecycleVector.mVector);
}

//---------------------------------------------------------------------------
template <typename DataType>
RecycleVector<DataType>::~RecycleVector(void)
{
	sRecyclery.Dispose(mInitializer, mVector);
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
const RecycleVector<DataType>&
RecycleVector<DataType>::operator=(
	const RecycleVector&	inRecycleVector)
{
	RecycleVector	temp(inRecycleVector);
	swap(temp);
	return *this;
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
DataType&
RecycleVector<DataType>::operator[](
	size_type	inIndex)
{
	return (*mVector)[inIndex];
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
const DataType&
RecycleVector<DataType>::operator[](
	size_type	inIndex) const
{
	return (*mVector)[inIndex];
}

//---------------------------------------------------------------------------
template <typename DataType>
void
RecycleVector<DataType>::resize(
	size_type		inNewSize,
	const DataType&	inVal)
{
	mVector->resize(inNewSize, inVal);
	mInitializer.fNumItems	= inNewSize;
	mInitializer.fInitValue	= inVal;
	AbstractVector<DataType>::resize(inNewSize, inVal);
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
void
RecycleVector<DataType>::fill(
	const DataType&	inVal)
{
	mVector->fill(inVal);
}

//---------------------------------------------------------------------------
template <typename DataType>
inline
void
RecycleVector<DataType>::swap(
	RecycleVector&	inRecycleVector)
{
	AbstractVector<DataType>::swap(inRecycleVector);
	std::swap(inRecycleVector.mVector, mVector);
	std::swap(inRecycleVector.mInitializer, mInitializer);
}



#endif // RecycleVector_h__
