/*
 *  SiteCategory.cpp
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Options.h"
#include "SiteCategory.h"
#include "dcdflib.h"
#include "Numerics.h"


extern Options options;

SiteCategory::SiteCategory(int numSites) {
	if (options.allowASRV == 1) {
        std::cout << "\t- Initializing gamma-distribution for rate variation.  Shape parameter is: " << options.gammaShapeASRV << std::endl;
	}
	
	siteAssignments.resize( numSites, 0 );
	backupAssignments.resize( numSites, 0 );
	setGammaDistribution(options.gammaShapeASRV);
}

int SiteCategory::setGammaDistribution(double shapeParm) {
	gammaShapeParameter_ASRV = shapeParm;
	return setGammaDistribution();
}

int SiteCategory::setGammaDistribution(void) {
    // Added a check to make sure that rate categories differ from each other
    // (This shouldn't strictly be required, but for some reason it appears to be.  Undiagnosed issue.)
    
	// Code adapted from PIDAMEBL, A.P. Jason de Koning
	int which, status;
	double p,q,x,a,b, bound;
	
    // SET ALL THE MIXTURE COMPONENTS TO USE THE SAME GAMMA DIST (#0)...
    if ( siteRateDistribution.empty() ) {
		siteRateDistribution.resize( options.numCatASRV, 1.0 );
	}

	double sum = 0.0;
    x = 0;
	for (int k=0; k<options.numCatASRV; k++) {
		a = b = gammaShapeParameter_ASRV;
		//a = gammaShapeParameter[0];
		//b = gammaScaleParameter[0];
		
		which = 2;                                      // (calculate P and Q)
		p = (2.0*k+1.0) / (2.0 * options.numCatASRV);	// set region to integrate over
		q = (1-p);
		
		// Do the inverseCDF search for gamma distributions...
		cdfgam( &which, &p, &q, &x, &a, &b, &status, &bound);
		
		if ( x < 1e-10 ) { x = 1e-10; }
		if ( x > 100   ) { x = 100; }
		
		siteRateDistribution.mDataArray[k] = x;
		sum += x;
	}
	
	// INSTEAD OF SCALING TO SUM TO ONE FOLLOWING YANG, WE SCALE SO THAT THE MAXIMUM RATE MODIFIER IS ONE
    // ( THIS SIMPLIFIES DEALING WITH THE UNIFORMIZATION )
	sum = x;
	
    int isUniform = 0;
    
	for (int k=0; k<options.numCatASRV; k++) {
		siteRateDistribution.mDataArray[k] /= sum; 
    //	siteRateDistribution.mDataArray[k] *= options.numCatASRV/sum; 
	//	std::cout << "			Gamma dist cat " << k << ": " << siteRateDistribution.mDataArray[k] << std::endl;
        if (k>0) if ( siteRateDistribution.mDataArray[k] == siteRateDistribution.mDataArray[k-1] ) isUniform++;
    }

	// Set the largest rate (global variable)
	options.__largestRate = siteRateDistribution.mDataArray[ (options.numCatASRV - 1)];
    
/*
    if (options.allowASRV) {
        std::cout << "******************************************************************************************************" << std::endl;
        std::cout << "OVERRIDING RATES FOR TESTING:" << std::endl;

        // MAX RATE AT 1 SIMPLIFIES UNIFORMIZATION CONSTANT
        
        siteRateDistribution.mDataArray[0] = 0.1;
        siteRateDistribution.mDataArray[1] = 1.0;
        options.__largestRate = siteRateDistribution.mDataArray[ 1 ];
       
        for (int i=0; i<2000; i++) {
            if (i < 1000) {
                siteAssignments.mDataArray[i] = 0;
            } else {
                siteAssignments.mDataArray[i] = 1;
            }
        }
        std::cout << "******************************************************************************************************" << std::endl;
    }
 */
    return isUniform;
    
    
}
