/*
 *  MCMC.cpp
 PLEX - Phylogenetics, Likelihood, Evolution, and Complexity
 version 0.94 - beta version July, 2012
 
 Authors: A.P. Jason de Koning, Wanjun Gu and David Pollock
 jason.de.koning@gmail.com http://jasondk.org
 wanjungu@gmail.com
 David.Pollock@UCDenver.edu  http://www.evolutionarygenomics.com
 
 This program was created and distributed by the
 David Pollock laboratory, University of Colorado School of Medicine
 copyright © 2012 David Pollock
 All rights reserved.
 
 http://www.evolutionarygenomics.com
 Dept. Biochemistry & Molecular Genetics
 12801 17th Ave., MS 8101
 PO Box 6511
 Aurora, CO 80045
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vector>
#include <fstream>
#include "MCMC.h"
#include "SiteCategory.h"
#include "PriorDistributions.h"

#include <time.h> // for timing

using std::cerr;
using std::endl;

// TO DO:
// - DRA update with B1u - plug-in new function
// - make sure all samplers respect the global likelihood temperature
// - go back to a single uniformization constant across models

// - remove the redundant GTR vs ProteinModel functions; make a master class for all models

// ************************************************************************//
// EmptySequenceDetector is old weird stuff; get rid of this
// ************************************************************************//
class EmptySequenceDetector {
public:
    bool	operator()(
                       const TreeNodeData_&	inData) const;
};

bool EmptySequenceDetector::operator()( const TreeNodeData_& inData) const{
    return inData.sequence.empty();
}

// ************************************************************************//
// MARK: Initialization
// ************************************************************************//
MCMC::MCMC(void) {
    // Initials
    updateCount = 1;
    resetAfterConjugatePhase = 0;
    
	// Set a pointer to the global numerics object
	num = &Numerics::getGlobalNumerics();
}

void MCMC::InitializeMixtureModel(MixtureModel& m_Model,
								  const IntMatrix& m_ChangeMatrix) {
	
    m_Model.SetNumStates( theData.getNumStates() );
    m_Model.SetNumModels(options.m_nNumModels);
	m_Model.SetFixedModels(options.m_bFixedFreq);
	m_Model.SetUpdateWindowsize(options.m_fUpdatewindow);
	m_Model.InitializeParameter(m_ChangeMatrix);
	
	if (!options.capUniformizationRate) m_Model.updateUniformization();
}

void MCMC::InitializeTree(ObservedData& mData,  Tree& mTree ) {
	
	// Initialize the tree with the sequences, sequence names
	// mTree.for_each(SequenceAssigner(m_vecSequences, m_vecSeqname));
    
    mTree.assignSequencesToLeaves( mData );
	mTree.RemoveFreeLeaves(EmptySequenceDetector());
	mTree.setinternalnodes( mData.getAlignmentLength() );
	
	mTree.root()->propogateRootLink(mTree.root());
    
}

void MCMC::InitializeTreeAlternate(Tree& mTree) {
	
	// Propogate a link to the root node
	mTree.root()->propogateRootLink(mTree.root());
}

// ****************************************************************************************
// MARK: Run MCMC moves
// ****************************************************************************************
/*
 void MCMC::RunMarkovChain(
 Tree& mTree,
 GTRModel& mModel,
 IntVector& m_vecSiteAssignment,
 ObservedData& theData) {
 
 // MESSY UPDATER FUNCTION FOR GTR ANALYSIS; approximately what was used in de Koning et al (2009)
 
 static int modelCount = 0;
 static int assignCount = 0;
 static int shapeCount = 0;
 static int ancstateCount = 0;
 static int blCount = 0;
 
 double rannum1 = num->Random();
 int numModels = mModel.GetNumModels();
 
 if (updateCount == 0) {
 // std::cout << "Performing initial marginal reconstruction (sample)..." << endl;
 ancstateCount++;
 
 if (options.m_nMethod == 6) {
 mModel.updateUniformization();
 int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
 mTree.root()->setCalculateFullProbabilityDistribution(1);
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
 } else {
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 }
 
 mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1 );
 
 // Now reset uniformization constant to something bigger...
 if (options.m_nMethod == 6) mModel.enlargeUniformization();
 
 if (options.m_nMethod == 6) {
 mModel.setUniformization(0, 1.0);
 }
 
 if ( options.allowASRV && (options.numCatASRV > 1) ) {
 if (options.m_nMethod != 6) {
 tree_countChanges( mModel, theData, m_vecSiteAssignment );
 } else {
 tree_countChanges( mTree,mModel, theData, m_vecSiteAssignment );	
 }
 
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 Gibbs_updateSiteAssignments_RateAssignments(mModel, theData, m_vecSiteAssignment);	
 }
 
 if (options.m_nMethod != 6) {
 tree_countChanges( mModel, theData, m_vecSiteAssignment );
 } else {
 tree_countChanges( mModel, theData, m_vecSiteAssignment );	
 }
 
 
 }
 else if ( ((updateCount % options.outputIncompleteLogL) == 0) && (options.outputIncompleteLogL > 0) ) {
 ancstateCount++;
 
 //        cout << "UNI CONSTANT IS: " << mModel.returnUniformizationConstants()->mDataArray[0] << endl;
 
 int tempCalcProbDist2 = mTree.root()->getCalculateFullProbabilityDistribution();
 mTree.root()->setCalculateFullProbabilityDistribution(1);
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist2);
 
 // [DIAGNOSTIC CHECK:] cout << "Alternative lnL says: " << mTree.root()->incompleteDataLikelihood( mTree.root(), mModel.returnStateFrequencies(), mTree.GetModelFrequencyVector(), m_vecSiteAssignment, theData ) << endl;
 
 // Force a full ancestral state update every so many gens...
 mModel.updateUniformization();
 int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
 mTree.root()->setCalculateFullProbabilityDistribution(1);
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
 
 mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1 );
 
 // Now reset uniformization constant to something bigger...
 if (options.m_nMethod == 6) mModel.enlargeUniformization();
 
 // Don't need to do this every time, do we?
 tree_countChanges( mModel, theData, m_vecSiteAssignment );
 }
 
 updateCount++;
 
 double ancstateProb;
 ancstateProb = options.ancStateUpdateFreq;
 if (updateCount < 2000) { 
 ancstateProb = 0.99; 
 }
 
 // ********************************************************************************************************************************
 // Output new parameter file 'posteriorRateMatrix'
 // ********************************************************************************************************************************
 if ( (updateCount % options.outputFrequency == 0) ) {
 static int fileOpen = 0;
 static ofstream posteriorMeanMatrix;
 if (fileOpen == 0) {
 posteriorMeanMatrix.open( "posteriorRateMatrix", fstream::out);
 fileOpen = 1;
 }
 
 // output the posterior mean rates
 
 // output SUM over: rate * siteRate * siteRateFraction
 static DoubleVector siteRateFraction;
 if (siteRateFraction.empty()) { siteRateFraction.resize( options.numCatASRV, 0 ); }
 else siteRateFraction.fill(0.0);
 
 for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
 siteRateFraction.mDataArray[ mTree.root()->GetData().siteCats->getAssignment( ii ) ]++;
 }
 
 for (int jj=0; jj<options.numCatASRV; jj++) {
 siteRateFraction.mDataArray[jj] /= m_vecSiteAssignment.size();
 }
 
 static int firstOutput = 1;
 if (firstOutput == 1) {
 posteriorMeanMatrix << "TC	TA	TG	CA	CG";
 if (options.allowASRV) {
 posteriorMeanMatrix << "	Shape";
 } 
 if (options.updatebls) {
 posteriorMeanMatrix << "	TreeLength" << endl;
 }
 std::cout << endl;
 firstOutput = 0;
 }
 
 static double treeLen;
 if ( mTree.root()->GetData().treeLength != 0 ) {
 treeLen =  mTree.root()->GetData().treeLength;
 }
 
 for (int mm=0; mm<mModel.GetNumModels(); mm++) {
 if (mModel.GetNumModels() > 1) posteriorMeanMatrix << "Model class " << mm << ":" << endl;
 int iPairs[6] = { 1, 0, 1, 0, 2};
 int jPairs[6] = { 3, 1, 2, 3, 3}; 
 for (int ii=0; ii<5; ii++) {
 double outputRate = 0.0;
 //for (int kk=0; kk<options.numCatASRV; kk++) {
 //	outputRate += mModel.returnAlphaMatrix3D()->mRows.mDataArray[mm].mRows.mDataArray[iPairs[ii]].mDataArray[jPairs[ii]] * mTree.root()->GetData().siteCats->getRate( kk ) * siteRateFraction.mDataArray[kk];
 //}
 outputRate = mModel.returnAlphaMatrix3D()->mRows.mDataArray[mm].mRows.mDataArray[iPairs[ii]].mDataArray[jPairs[ii]] / mModel.returnAlphaMatrix3D()->mRows.mDataArray[mm].mRows.mDataArray[0].mDataArray[2];
 posteriorMeanMatrix << outputRate;
 if (ii < 4) posteriorMeanMatrix << "\t";
 } 
 if (options.allowASRV) {
 posteriorMeanMatrix << "\t" << mTree.root()->GetData().siteCats->getShapeParm();
 }
 if (options.updatebls) {
 posteriorMeanMatrix << "\t" << treeLen;
 }
 posteriorMeanMatrix << endl;
 } 
 }
 // ********************************************************************************************************************************
 
 
 if (rannum1 < ancstateProb)	{ 
 double rannum2 = num->Random();
 
 // Branch-length updater
 if ( (options.updatebls == 1) && (num->Random() < options.brlenupdatefreq)) {
 blCount++;
 bool m_bUpdate;
 int rannode = (int)( num->Random() * options.m_nNumBranches);
 
 // Fully calculate all substitution prob matrices...
 int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
 
 // Version 0: full branch-specific likelihood
 // Version 1: conjugate gibbs on bl's
 // Version 2: simple change in lik (b1u)
 // Version 3: MLE for bl (b1u)
 
 int version = 0;
 double var = 0.2;
 if (updateCount < 5000) { var = 1.0; }
 
 if (options.m_nMethod == 6) {
 // Conjugate Gibbs sampler seems to be working very well...
 if (updateCount < 5000) { version = 1; }
 else { version = 1; }
 }
 
 if ( version == 0 ) {
 mTree.root()->setCalculateFullProbabilityDistribution(1);
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 }
 
 tree_updateBranchLengths( rannode, mModel, theData, m_vecSiteAssignment, version, var );
 
 // Now reset the flag...
 mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
 
 } else if ( (rannum2 < 0.99) || ( (numModels <=1) && ( (options.allowASRV == 0) || (options.numCatASRV <=1) ) ) )		//update model
 {
 m_bStateUpdate = false;
 m_bAssignment = false;
 
 static int capped = 0;
 if (options.m_nMethod == 6) {
 if ( (updateCount >= 5000) && (capped == 0) ) {
 std::cout << "              [ CAPPING THE UNIFORMIZATION CONSTANT and the NUMBER OF TREE NODES ]" << endl;
 options.capUniformizationRate = 1;
 
 // This should be averaged across model classes...
 double theRateFactor = mModel.getRateSum(); // / mModel.GetNumModels();
 
 // std::cout << "Got rate sum of " << theRateFactor << endl;
 
 // mModel.scaleDiscreteRates( theRateFactor ) ;
 
 double mu = mModel.returnUniformizationConstants()->mDataArray[0] / theRateFactor;
 
 // Change uniformization and ratescale for Qs
 mModel.setUniformization(0, mu ); //(mu / theRateFactor) );
 
 mModel.scaleRateMatrices( theRateFactor );
 
 // Transform bls
 mTree.root()->TransformBranchLength_B1u( mTree.root(), ( 1.0 / theRateFactor), mu );
 
 // Need to re-subdivide the tree...
 mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
 
 options.capNumTreeNodes = true;
 std::cout << "		[ --- FIXING UNIFORMIZATION CONSTANT AT " << (*mModel.returnUniformizationConstants())[0] << " --- ]" << endl;
 capped = 1;
 }
 }
 
 // SHOULD NOT DO UNIDIMENSIONAL PROPOSALS IF WE'RE NORMALIZING THE RATE MATRIX...
 if ( (updateCount > 5000) && (num->Random() <= 0.95) && ( options.m_nMethod == 6) && (1==2) ) { // ( m_nMethod != 4)
 // The unidimensional mechanism for B1u (updates many rate parameters one-by-one)...
 
 if (options.m_nMethod == 6) {
 UpdateGTRrates_unidimensional_b1u( mModel, theData, m_vecSiteAssignment, 1.0, -1 );
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(), 1);			
 } else {
 // This version should work with any method...
 UpdateGTRrates_unidimensional( mModel, theData, m_vecSiteAssignment, 1.0, -1 );
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(), 1);			
 }
 
 } else {
 MC_updateRateMatrix(mModel, theData, m_vecSiteAssignment);
 }
 modelCount++;
 
 // Call the gamma shape updater for ASRV if appropriate...
 // 1% of 95% of parameter updates
 if ( options.allowASRV && (options.numCatASRV > 1) && options.updateASRVparms && (num->Random() < 0.05) ) {
 shapeCount++;
 double var;
 var = 0.01;
 MC_updateASRV_shape(mModel, theData, m_vecSiteAssignment, var);
 
 }
 else		// update assignment
 {
 // 1% of parameter updates
 m_bModelUpdate = true;
 m_bAssignment = true;
 m_bStateUpdate = false;
 
 
 if ( options.allowASRV && (options.numCatASRV > 1) ) {
 assignCount++;
 if (options.m_nMethod != 6) {
 int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
 mTree.root()->setCalculateFullProbabilityDistribution(1);
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
 }			
 
 Gibbs_updateSiteAssignments_RateAssignments(mModel, theData, m_vecSiteAssignment);	
 }
 
 if ( numModels >= 1) {
 Gibbs_updateSiteAssignments_MixtureComponents(m_vecSiteAssignment, mModel, theData );	
 }
 }
 }
 } else			// update internal nodes
 {
 ancstateCount++;
 m_bModelUpdate = true;
 m_bAssignment = false;
 m_bStateUpdate = true;
 
 //---------------------------------
 // About to update internal states 
 //---------------------------------
 
 // 'Full' marginal ancestral state sampler...
 if (  (options.m_nMethod == 3) || (options.m_nMethod == 5) || (options.m_nMethod == 4) || ( options.forceFullMarginalUpdate == 1 ) ) { //==6 && (updateCount<2000) ) ) ) { //&& (updateCount < 2000) ) ) { // || ( ( m_nMethod == 6) && (num->Random() < 0.25) ) ) { // || (options.m_nMethod == 6)
 if (options.m_nMethod == 6) mModel.updateUniformization();
 
 int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
 mTree.root()->setCalculateFullProbabilityDistribution(1);
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
 
 //		mModel.updateStationaryFreqs();
 //	mTree.AssignAncestralStatesByMarginalProbability(options.m_nMethod, m_vecStateIndex, numModels, mModel.returnStationaryFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates, 1);
 mTree.AssignAncestralStatesByMarginalProbability( mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies(), 1 );
 
 } else {
 //  cout << "** DOING LOCAL UPDATER **" << endl;
 // Original 'local' updater...
 
 if (options.m_nMethod == 6) mModel.updateUniformization();
 //[regular method]
 int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
 mTree.root()->setCalculateFullProbabilityDistribution(1);
 mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
 mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
 
 int rannode = (int)( num->Random() * mTree.numRegularNodes);
 //			std::cout << "Doing ancestral state sample form node " << rannode << "..." << endl;
 //[regular]
 Gibbs_updateAllAncestralAndTransientStates_local(  rannode, mModel, theData, m_vecSiteAssignment, mModel.returnStateFrequencies() );
 //[N4 method]
 // RunMCMCGibbsInternalStatesByN4(rannode, mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates, mModel.returnRateMatrix3D(), mModel.returnUniformizationConstants());
 
 
 }
 
 // Now reset uniformization constant to something bigger...
 if (options.m_nMethod == 6) mModel.enlargeUniformization();
 
 // Don't need to do this every time, do we?
 tree_countChanges( mModel, theData, m_vecSiteAssignment );
 }
 
 
 double logLikelihood1 = CalculateCompleteDataLogLikelihood(mModel, m_vecSiteAssignment);
 //[this one]
 //	std::cout << " ***** Complete data log-likelihood:  	" << logLikelihood1 << endl;
 mTree.completeDataLnL = logLikelihood1;
 
 m_bModelUpdate = true;
 if ( (updateCount % 10000) == 0) std::cout << "	*** Ancstate updates: " << ancstateCount << ", model updates: " << modelCount << ", bl counts: " << blCount << ", assign counts: " << assignCount << ", shape update counts: " << shapeCount << endl;
 }
 
 */

void MCMC::setPointers(vector<Tree *> &treeInit,
                       int currentTreeInit,
                       ProteinModel& modelInit,
                       IntVector& siteAssignmentInit,
                       ObservedData &dataInit,
                       string ppOutInit ) {
    
    mModel = modelInit;
    m_vecSiteAssignment = siteAssignmentInit;
    theData = dataInit;
    Trees = treeInit;
    currentTree = currentTreeInit;
    mTree = (*Trees[currentTree]);
    posteriorPredictiveOut.open( ppOutInit.c_str() );
}

void MCMC::RunMarkovChain( void ) {
    // TODO: Clean this up!
    
    int numModels = mModel.GetNumModels();
	bool success = false;
    
	static int realNumGens = 0;
	static int lastUpdateType = -1;
	static int blCount = 0;
    
	likelihood = 0;         // for keeping track of lnL
    numUpdated = 0;         // only count # of updates made per generation
    forceCoevolution = 0;   // allow data replacement simulation to force an immediate coevol calc.
    
    // if (updateCount > 500 ) cout << "   MU is " << mModel.returnUniformizationConstants()->mDataArray[0] << endl;
    
	// **************************************************************************************************
	// Set some of the move type probabilities...
	// **************************************************************************************************
	double ancestralStateProb = 0.999;
	double mixtureAssignmentProb = 0.85;
	
    // Do more aggressive updates early in the chain (during burn-in)
	if (updateCount > 200) ancestralStateProb = options.ancStateUpdateFreq;
	if (!mModel.getRARdistributed()) ancestralStateProb = options.ancStateUpdateFreq;
    
    // Force branch-length updates to be infrequent during burnin
    // if (updateCount < 5000 && options.allowASRV) ancestralStateProb = 0.975;
	
	// **************************************************************************************************
	// Initialize chain and run posterior statistics
	// **************************************************************************************************
	if (updateCount == 1) _util_initializeChain();
    
// Testing:
//    if (updateCount >= 6000) {
//
//        // Prep by making sure counts are up to date with model
//        tree_countChanges(mModel);
//        
//        for (int iii=0; iii < 100; iii++) {
//
//            cout << "Updating rate matrix #" << iii << "..." << endl;
//            MCMC_updateRateMatrix();
//            
//            cout << "Updating stationary frequencies..." << endl;
//            mModel.forceUpdateStationaryFreqs();
//            
//            cout << "Doing ancestral state update (full)..." << endl;
//            Gibbs_updateAllAncestralAndTransientStates_full( mModel );
//
//            cout << "Updating change counts # " << iii << "..." << endl;
//            tree_countChanges(mModel);
//        
//        }
//        
//        updateCount = 7000;
//        coevo_calculateStatistics();
//        cout << "DONE." << endl;
//        exit(-1);
//    }
    
 	// **************************************************************************************************
	// Do full marginal ancestral state updater if requested
	// **************************************************************************************************
    if ( updateCount % options.outputIncompleteLogL == 0 ) {
        numUpdated += Gibbs_updateAllAncestralAndTransientStates_full(  mModel );
    
		// Run the data replacement simulation (if applicable)
		coevo_DataReplacementSimulation();
		
		// Run the coevolutionary statistics (if applicable)
		coevo_calculateStatistics();
	}
    
	// **************************************************************************************************
	// Calculate the compete-data logL every outputFrequency generations
	// **************************************************************************************************
    if ( (updateCount % options.outputFrequency) == 0 ) {
        likelihood = CalculateCompleteDataLogLikelihood(mModel);
        mTree.completeDataLnL = likelihood;
    }

	// **************************************************************************************************
	// Main loop
	// **************************************************************************************************
	while (!success) { 
            // **************************************************************************************************
            // Cap the uniformization constant after burn-in ends
            // NOTE: am experimenting with overriding this whenever ancestral states get updated...
			// **************************************************************************************************
			if (options.m_nMethod == 6) {
                
                // Equilibrate mu if not updating branchlengths...
                
                // SKIPPING THE EQUILIBRATION STEP; SEEMS TO BE MESSING WITH THINGS
                /* static bool equilibratedMu = true; 
                
                
                if ( (!options.updatebls) && (updateCount >= 2500) && (!equilibratedMu) ) { //!options.updatebls && 
                    // Update mu and the rate matrix until mu stabilizes - only when branchlenghts are NOT being sampled
                    bool notDone = true;
                    
                    int count = 0;
                    double lastMu = 0.0;
                    double currMu; 
                    int goodtimes = 0;
                    cout << "           [ ** Equilbrating the rate matrix and uniformization constant ** ]" << endl;


                    // Testing code below:
                    
//                    // Init mu to something small
//                    int last = options.capUniformizationRate;
//                    options.capUniformizationRate = 0;
//                    mModel.forceUpdateUniformization();
//                    
//                    mModel.setUniformization(0, ((*mModel.returnUniformizationConstants())[0] / 20.0) );
//                    currMu = (*mModel.returnUniformizationConstants())[0];
//                    cout << "           INIT MU to " << currMu << endl;
//                    
//                    options.capUniformizationRate = last;
//                    for (int ii=0; ii<50; ii++)   MCMC_updateRateMatrix();
// 
//                      last = options.capUniformizationRate;
//                    options.capUniformizationRate = 0;
//                    mModel.forceUpdateUniformization();
//                    options.capUniformizationRate = last;
//                   
                    
                    while (notDone ) {
                        MCMC_updateRateMatrix();
                        
                        // if (options.__currentGen < 10000) {
                            int last = options.capUniformizationRate;
                            options.capUniformizationRate = 0;
                            mModel.forceUpdateUniformization();
                            
                            currMu = (*mModel.returnUniformizationConstants())[0];
                            cout << "           UPDATED MU to " << currMu << endl;
                            
                            options.capUniformizationRate = last;
                        // }
                        
                        
                        
                      //  MCMC_updateRateMatrix();
                      //  mModel.forceUpdateUniformization();
                        
                        
                      //  tree_countChanges( mModel, 1 );
                        
                        if ( num->Random() < 0.1 ) {
                            // Added resegmentation here
                            tree_countChanges( mModel, 1 );
                            if ( options.allowASRV) Gibbs_updateSiteAssignments_RateAssignments(mModel );	

                            double maxBL = options.maxBL / ( mModel.getRateSum( 0, mTree ) - options.maxBL * (*mModel.returnUniformizationConstants())[0] );
                            
                            mTree.subdivideLongBranches( mTree.root(), maxBL, mModel.returnUniformizationConstants(), 1 ); // recursive version
                            mTree.root()->NormalizeBranchSegmentLengths(theData, mTree.root() );
                            
                             Gibbs_updateAllAncestralAndTransientStates_full(  mModel );
                        } else {
                            tree_countChanges( mModel, 1 );
                        }
                        
                        // Changed this from 1e-03
                        if ( abs( currMu - lastMu ) < 1e-04 ) {
                            goodtimes++;
                        } else {
                            goodtimes = 0;
                        }
                        
                        if (goodtimes > 5) notDone = false;
                        lastMu = currMu;
                        
                        if (count > 1000) notDone = false;
                        count++;
                    }
                    
                    cout << "           [ ** Done.  Rate matrix and uniformization constant have converged ** ]" << endl;
                    
                    equilibratedMu = true;
                }
                */

                if ( (updateCount >= 6500) && (!options.capNumTreeNodes) ) {
					std::cout << "		" << updateCount++ << " [ Starting tree-length: " << setprecision(4) << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << "; inflation factor: " << setprecision(3) << (double)((double)mTree.root()->CountBranches()/(double)options.__initialNumBranches) << "X (max. segment-size: " << options.maxBL <<") ]" << endl;
                    
                    // Now resegment
                    double maxBL = options.maxBL / ( mModel.getRateSum( 0, mTree ) - options.maxBL * (*mModel.returnUniformizationConstants())[0] );
                    mTree.subdivideLongBranches( mTree.root(), maxBL, mModel.returnUniformizationConstants(), 1 ); // recursive version
                    mTree.root()->NormalizeBranchSegmentLengths(theData, mTree.root() );
                    cout << "                [ SUBDIVISION THRESHOLD (in bl parameter-space) = " << maxBL << "; max expected subs. per site = " << options.maxBL << " ]" << endl;
                    
                    tree_countChanges( mModel );

                    numUpdated += Gibbs_updateAllAncestralAndTransientStates_full(  mModel );
                    
                    std::cout <<  "                [ CAPPING THE NUMBER OF TREE NODES ]" << endl;
                    
                    options.capUniformizationRate = 1;
                    options.capNumTreeNodes = true;
                }
			}
        
        double whichUpdate = num->Random();
			if ( (whichUpdate < ancestralStateProb)  ) {
				// Update the rate model etc.
                
                double rannum2 = num->Random();
				
				// UPDATE MODEL...
				if ( (rannum2 < mixtureAssignmentProb) || (numModels <=1)) {
					// Update the rate model and sometimes site assignments
                    success = true;
					
					//----------------------------
					// About to update Model rates
					//----------------------------
					int whichModel = 0;
					
					if ( ( (options.doRAR == 0) && (options.fixedRates == -1)  && (options.m_nMethod == 6) )  || ( ( updateCount < 5000 ) && (options.m_nMethod == 6) &&  (options.fixedRates == -1) ) ) {
                        
                        // Combined update of rate matrix and site assingments
                        
                        MCMC_updateRateMatrix();
                        
                       // options.reversibleModel = 0;
                       // mModel.setReversible(0);
                        
                        // Update the site assignments too, every once in a while
                        if (options.allowASRV || numModels > 1) {
                            if ( (num->Random() <= 0.01) ) {
                                MCMC_updateSiteAssignments();
                                //cout << "MCMC_updateSiteAssignments() #1" << endl;
                            }					
                        }
                        
                    } else {
                        // Not sure why to do this stuff, other than to initialize RAR from an unconstrained model...
                        
						if (mModel.getRARdistributed() && (updateCount > 5500) ) {
                            // Switching from unconstrained model to RAR...
                            _util_initializeRARmodelFromChain();
                            
                            // This selects from among assignment updates and parameter updates
                            MCMC_updateRARmodel();
                        } else {
                            // if NOT doing getRARdistributed():
                            cout << "MC_ratematrixUpdate 1" << endl;
                            if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " Resampling model parameters (non-RAR) /     ()" << endl;
                            numUpdated += MC_updateRateMatrix_nonRAR( mModel);
                            
                            mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(), 1);
                        }
                        
                        if (! options.capNumTreeNodes && (updateCount >6500) ) {

                            options.capNumTreeNodes = true;
                            std::cout << "		[ --- FIXING NUMBER OF TREE NODES --- ]" << endl;
                            
                            numUpdated += Gibbs_updateAllAncestralAndTransientStates_full( mModel );
                        } 
                    }
                    
                    // Now updates rates, assignments, etc.
                    if (mModel.getRARdistributed() ) { //&& (updateCount > 5500) ) {
                        
                        if ((updateCount==1) || ( (updateCount%30)==0)) {
                            // Do a RAR parameter assignment sweep...
                            
                            if (mModel.getDependentRates() && (mModel.GetNumModels() > 1)) {	
                                if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount++ << " [Resampling rate parameter values with re-assignment sweep / RunMCMCFullParameterSweep_dependentAssignment 100%]" << endl;
                                numUpdated += Gibbs_updateRateParameterAssignments_RAR_DRA_B1u(mModel, 1.0, -1);
                            } else {
                                whichModel = (num->Random() * numModels);
                                
                                if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount++ << " [Resampling rate parameter assignments / (a) RunMCMCFullParameterSweep 100%]" << endl;
                                numUpdated += Gibbs_updateRateParameterAssignments_RAR_B1u(mModel, 1.0, -1);
                            }
                            
                        } else {
                            if (mModel.getDependentRates()) {	
                                whichModel = -1;
                            } else {
                                whichModel = int(num->Random() * numModels); //[jdk] Numerics::getGlobalNumerics().Random();
                            }	
                            
                            if ( num->Random() <= 0.3) { // 0.3 ) { // was last 0.3 // BLOCKED July 2009
                                // RAR parameter sweep...
                                
                                if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [Resampling rate parameter assignments / (b) RunMCMCFullParameterSweep / FULL]" << endl;
                                
                                if (mModel.getDependentRates()) {	
                                    numUpdated += Gibbs_updateRateParameterAssignments_RAR_DRA_B1u(mModel, 0.99, whichModel);
                                } else {
                                    numUpdated += Gibbs_updateRateParameterAssignments_RAR_B1u(mModel, 1.0, whichModel);
                                }
                                
                            } else {
                                // Update the RAR rate parameter distribution...
                                
                                if ( (num->Random() <= 0.5) || (mModel.getFixedFreqs() == 1) ) {
                                    // Update the RAR parameter distribution...
                                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [Resampling rate parameter values with re-assignment sweep / RunMCMCDistributionParametersWithSweep <= 100%]" << endl;
                                    numUpdated += MC_updateRARDistributionParameters_B1u( mModel, 1.0, whichModel);
                                } else {
                                    cout << "MC_ratematrixUpdate 2" << endl;
                                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " Resampling other model parameters (non-RAR) / MC_updateRateMatrix_nonRAR()" << endl;
                                    numUpdated += MC_updateRateMatrix_nonRAR( mModel);
                                }
                                
                            }
                        }
                    } 
                    
//                    else {
//                        
//                        MCMC_updateRateMatrix();
//                        
//                        /*
//                        // if NOT doing getRARdistributed():
//                        
//                        cout << "MC_ratematrixUpdate 3" << endl;
//                        if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " Resampling model parameters (non-RAR) / MC_updateRateMatrix_nonRAR()" << endl;
//                        numUpdated += MC_updateRateMatrix_nonRAR( mModel);
//                        
//                        mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels(), 1);
//                         */
//                         
//                    }
                    
                    
                    //[debug]
                    
                    lastUpdateType = 0;
                    
                    // Call the gamma shape updater for ASRV if appropriate...
                    // 1% of 95% of parameter updates
                    if ( options.allowASRV && (options.numCatASRV > 1) && options.updateASRVparms && (num->Random() < 0.05) ) { // was 0.05
                        //	shapeCount++;
                        double var;
                        //				if (options.m_nMethod == 6) var = 0.001;
                        //		else var = 0.2;
                        
                        // pumping this up for a test
                        if (updateCount <= 5000) {
                            var = 2.0;
                        } else {
                            var = 0.02; //0.02
                        }
                        
                        if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Resampling ASRV distribution parameters ]" << endl; 
                        
                        tree_countChanges( mModel, 1 );
                        
                        MC_updateASRV_shape( mModel, var );
                        
                        tree_countChanges( mModel, 1 );
                    }
                    
                } else {
                    // Just update the site assignments
                    if (  (lastUpdateType == 1))  { success = false; continue; }
                    
                    success = true;
                    
                    MCMC_updateSiteAssignments();
                    //cout << "MCMC_updateSiteAssignments() #2" << endl;
                    
                    lastUpdateType = 1;
                }
            } else {
                // UPDATE ANCESTRAL STATES / TRANSIENT STATES...
                if ( (updateCount > 200) && (lastUpdateType == 2) ) { success = false; continue; }
                lastUpdateType = 2;
                success = true;
                //---------------------------------
                // About to update internal states 
                //---------------------------------
                int rannode;
                //			double fullUpdateFreq = 0.3;
                
                int which;
 
                                 
                which = 1;
                if ( options.forceFullMarginalUpdate == 1 ) { which = 2; }
                // [ do we really need this? ] if ( updateCount < 200 ) { which = 2; }
 
                if ( which == 1 ) {
                    //if ( (updateCount % options.outputFrequency) == 0) 
                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		[ Resampling ancestral/transient states with local updater... ]" << endl;
                    rannode = (int)( num->Random() * mTree.numRegularNodes); // options.m_nNumBranches); //mTree.leaves() ); // options.m_nNumBranches);
                     
                     // Differentiate between root frequencies: parameters for predefined & reversible model.  O/w use stationary freqs.
                    if (  options.scaleByFrequencies || mModel.getReversible() ) {
                        numUpdated += Gibbs_updateAllAncestralAndTransientStates_local( rannode, mModel, mModel.returnStateFrequencies() );
                    } else {
                        numUpdated += Gibbs_updateAllAncestralAndTransientStates_local( rannode, mModel, mModel.returnStationaryFrequencies() );
                    }
                    
                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [Gibbs_updateAllAncestralAndTransientStates_local from node " << rannode << "]" << endl;			
                    
                } else  {
                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		[ Resampling ancestral/transient states with full marginal updater...]" << endl;
                    
                    numUpdated += Gibbs_updateAllAncestralAndTransientStates_full(  mModel );
                    
                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [Complete ancestral state assignment (marginal sample)]" << endl;			
                    
                }
                
                // BLOCK
                /*	// PP
                 static int hasBeenDone = 0;
                 if (hasBeenDone == 0) {
                 hasBeenDone = 1;
                 //		std::cout << "Calculating the observed T(X)..." << endl;
                 posteriorPredictiveOut << "Observed:" << "\t" << mTree.calculateObservedSequenceTestStatistic(mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates) << endl;
                 //		std::cout << endl;
                 }
                 
                 if (updateCount > 1000) {
                 //	if ( num->Random()	<= 0.02 ) {
                 mModel.updateStationaryFreqs();
                 
                 std::cout << endl << "	[Calculating the multinomial likelihood, T(X), of a posterior predictive sample...]" << endl << endl;
                 posteriorPredictiveOut << "Predicted:" << "\t" << mTree.simulatePosteriorPredictiveTestStatistic_multinomialLikelihood(mModel.returnStateFrequencies(), m_vecSiteAssignment, m_vecStateIndex, m_vecState, m_nMinstate, m_nNumStates) << endl;
                 //std::cout << endl;
                 //	}
                 }
                 */
            }	
        		
        
        // Override sub resampling >3k gens
        //  if ( (updateCount >2990) && (updateCount < 3020) ) { cout << " TURNING OFF RESAMPLING OF SUB" << endl; }
        //  if (updateCount > 3000) { options.__overrideSubResampling = 1; }
        
         realNumGens++;
    } // while !success

    updateCount++;
    options.__currentGen = updateCount;    
}

// ****************************************************************************************
// MARK: Likelihood calculations
// ****************************************************************************************
double MCMC::CalculateCompleteDataLogLikelihood( MixtureModel& mModel2) {	
	
 	return CalculateCompleteDataLogLikelihood( mModel2, 0);
}

double MCMC::CalculateCompleteDataLogLikelihood( MixtureModel& mModel2,
                                                int version) {		
	
	double logLikelihood;
    mTree.setprobmatrixalongbranch( mModel2, mModel2.GetFreqModels() );
	mTree.calculateCompleteDataLogL_overBranches( mModel2 , m_vecSiteAssignment, logLikelihood, version );	
	
	return(logLikelihood);
}

double MCMC::CalculateCompleteDataLogLR_overTree(
                                                 MixtureModel& mModel1,
                                                 MixtureModel& mModel2
                                                 ) {
	
    // Entirely general - uses substitution probability matrices along branches
    
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(1);
	mTree.setprobmatrixalongbranch( mModel1, mModel1.GetFreqModels());
	double logLikelihood1 = CalculateCompleteDataLogLikelihood(mModel1);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(1);
	mTree.setprobmatrixalongbranch(mModel2, mModel2.GetFreqModels());
	double logLikelihood2 = CalculateCompleteDataLogLikelihood(mModel2);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	return( (logLikelihood2 - logLikelihood1) );
}

// ****************************************************************************************
// MARK: Partial likelihood evaluations
// ****************************************************************************************
double MCMC::CalculateCompleteDataLogLikelihood_Delta_B1u(
														  MixtureModel& mModel,
														  
														  int theModel, int x, int y) {		
	
	// Calculate the contribution of row x, col y to the complete-data log-likelihood under B1u
	double logLikelihood = mTree.calculateCompleteDataLogLdeltab1uni( mModel, m_vecSiteAssignment, logLikelihood, theModel, x, y);	
    
	// ********** WARNING **********
	// If capUniformizationRate is not being used, then a call to this function can be allowed to increase the uni constant
	if (options.capUniformizationRate == 0) {
		std::cout << "	** WARNING **  You've called a function that can alter the uniformization constant!  CalculateCompleteDataLogLikelihood_Delta_B1u()" << endl;
	}
	
	return(logLikelihood);
}

// ****************************************************************************************
// MARK: Samplers: slice updates
// ****************************************************************************************

/*
 if ( updateCount > 500 ) {
 
 int x = (int)( num->Random() * mModel.GetNumStates());
 int y = x;
 while (y==x) {
 y = (int)( num->Random() * mModel.GetNumStates());
 }
 
 DoubleMatrix3D    *CXY = mTree.root()->getChangeMatrix3D( mTree.root() ); //&GetData().countModelchange3D;
 DoubleMatrix      *C1  = mTree.root()->getAugmentedCountMatrix( mTree.root() );
 
 cout << "Outputting slice samples for (x,y)=" << x << ", " << y << "; counts are " << CXY->mRows.mDataArray[0].mRows.mDataArray[x].mDataArray[y] << ", " << C1->mRows.mDataArray[0].mDataArray[x] << endl;
 
 ofstream out;
 out.open("test.out");
 cout << mModel.returnRateMatrix3D()->mRows.mDataArray[0].mRows.mDataArray[x].mDataArray[y] << endl;
 for (int iii=0; iii<50000; iii++) {
 num->slice_sampleIndependentRateParameter(x, y, mModel.getUniformization(0), CXY->mRows.mDataArray[0], C1->mRows.mDataArray[0], mModel.returnRateMatrix3D()->mRows.mDataArray[0] );
 out << mModel.returnRateMatrix3D()->mRows.mDataArray[0].mRows.mDataArray[x].mDataArray[y] << endl;
 }
 out.close();
 exit(1);
 */

int MCMC::slice_updateRateMatrix_genNonRev_B1u( PriorDistribution &prior ) {
    
    // Requires an already initialized prior distribution
	int numStates  = mModel.GetNumStates();
	int numModels  = mModel.GetNumModels();
    
	int whichModel = -1; // Default to update all mixture components
    int whichRateFrom, whichRateTo, modelStart, modelEnd;
    
	DoubleVector    *mu = mModel.returnUniformizationConstants();
	// DoubleMatrix    *stateF = &mModel.returnStateFrequencies();
	
    // Counts matrices
	DoubleMatrix3D  *CXY = mTree.root()->getChangeMatrix3D( mTree.root() );
	DoubleMatrix    *C1  = mTree.root()->getAugmentedCountMatrix( mTree.root() );
	
    static int callCount=0;
 	int count=0;
    
    // For timing
    // clock_t start = clock();
    
    int theNumUpdates = (numStates * (numStates-1));

    if (options.reversibleModel) {
        theNumUpdates /= 2.0;
    } 
    
    // Let the initial width (slice.w) adapt to average
    static DoubleMatrix moveWidthTracker;
    static DoubleMatrix averageMoveWidths;
    
    // Initialize
    if ( averageMoveWidths.empty() ) {
        averageMoveWidths.resize( numModels, numStates*numStates, 0.0 );
    }
    
    if ( moveWidthTracker.empty() ) {
        moveWidthTracker.resize( numModels, numStates*numStates, 0.0 );
    }
    
    if ( (callCount % 100) == 0 ) {
        // Update the average widths for slice sampler
        for (int ii=0; ii < numModels; ii++) {
            for (int jj=0; jj < numStates*numStates; jj++) {
                averageMoveWidths.mRows.mDataArray[ii].mDataArray[jj] = moveWidthTracker.mRows.mDataArray[ii].mDataArray[jj] / 100.0;
                moveWidthTracker.mRows.mDataArray[ii].mDataArray[jj] = 0.0;
            }
        }
    }
    
    // Perform enough updates to do the entire matrix on average
    for (int numUpdates = 0; numUpdates < theNumUpdates; numUpdates++) {
        
        // Randomly choose a state transition to update
        whichRateFrom  = (int)(num->Random() * (numStates));
		whichRateTo    = (int)(num->Random() * (numStates));
		
        // Make sure we're not on the diagonal
		while (whichRateTo == whichRateFrom) {
			whichRateTo = (int)(num->Random() * (numStates));
		}

        // Set bounds for mixture model components
        if (whichModel == -1) {	modelStart = 0; modelEnd = numModels; } else { modelStart = whichModel; modelEnd = modelStart+1; }
        
        // Update the same state transition's rate parameter for each specific mixture component
		for (int theModel=modelStart; theModel < modelEnd; theModel++) {
			count++;
            
            static ModelPropertiesForUpdate modelProperties;

            double moveSize = 0.0;
            
            // Set links to the appropriate data structures
            if (options.scaleByFrequencies) {
                modelProperties.attachData(CXY->mRows.mDataArray[theModel], *C1, mModel.returnRateMatrix3D()->mRows.mDataArray[theModel], mModel.returnAlphaMatrix3D()->mRows.mDataArray[theModel], mModel.returnStateFrequencies().mRows.mDataArray[theModel], mTree.root()->GetData().siteCats->siteRateDistribution );
            } else {
                modelProperties.attachData(CXY->mRows.mDataArray[theModel], *C1, mModel.returnRateMatrix3D()->mRows.mDataArray[theModel], mModel.returnAlphaMatrix3D()->mRows.mDataArray[theModel], mModel.returnStationaryFrequencies().mRows.mDataArray[theModel], mTree.root()->GetData().siteCats->siteRateDistribution );
            }
            
            // Choose which parameters to update
            modelProperties.setUpdateProperties( mu->mDataArray[ theModel ], whichRateFrom, whichRateTo, theModel );
            
            // Perform slice sampling
            if ( options.reversibleModel != 1 ) {
                moveSize = num->slice_sampleIndependentRateParameter( prior, modelProperties, averageMoveWidths.mRows.mDataArray[ theModel ] );
            } else {
                moveSize = num->slice_sampleReversibleRateParameter( prior, modelProperties, averageMoveWidths.mRows.mDataArray[ theModel ] );
            }
            
            // Add in the current move distance
            moveWidthTracker.mRows.mDataArray[theModel].mDataArray[ (whichRateFrom * numStates) + whichRateTo ] += moveSize;
        }
 	}
    
    // cout << "Time: " << ((double)clock() - start ) / CLOCKS_PER_SEC << endl;
    
    // Keep track of how many times called
    callCount++;
    
    // Return number of rate parameters updated
	return count;
}

// ****************************************************************************************
// MARK: Samplers: conjugate Gibbs
// ****************************************************************************************
int MCMC::cGibbs_updateRateMatrix_genNonRev_B1u(
                                                ProteinModel& mModel
                                                ) {
    
// [override]	if (options.allowASRV != 0) { std::cout << "	ERROR! CONJUGATE GIBBS SAMPLING DOESN'T WORK WITH AMONG SITE RATE VARIATION YET..." << endl; exit(-1); }
	
	int numStates  = mModel.GetNumStates();
	int numModels  = mModel.GetNumModels();
    
	int whichModel = -1; // Default to update all mixture components
    int whichRateFrom, whichRateTo, modelStart, modelEnd;
    
	DoubleVector *mu = mModel.returnUniformizationConstants();
	DoubleMatrix* stateF = &mModel.returnStateFrequencies();
	
    // Counts matrices
	DoubleMatrix3D    *CXY = mTree.root()->getChangeMatrix3D( mTree.root() );
	DoubleMatrix      *C1  = mTree.root()->getAugmentedCountMatrix( mTree.root() );
	
    // Prior hyperparameters: fix these to focus the prior on small values
    double betaPrior_beta = 5.0;
    double betaPrior_alpha = 0.01; 
    
   // betaPrior_beta = 3.0;
   // betaPrior_alpha = 0.1;
   
    betaPrior_beta = betaPrior_alpha = 1.0;
    
	int count=0;
    
    // For timing
    // clock_t start = clock();
    
    // New version: update each row in random order
    // Perform enough updates to do the entire matrix on average
    static vector<int> permutation;
    
    if (permutation.size() != numStates) permutation.resize( numStates );
    
    // For each row...
    for ( int ii=0; ii < numStates; ii++ ) {
        // Generate a permutation of the columns and update in random order
        
        // Start with identity permutation
        for ( int jj=0; jj<numStates; jj++) {
            permutation[jj] = jj;
        }
        
        // Do a Knuth shuffle to get a new ordering
        for (int l = numStates - 1; l > 0; l--) {
            int j = (int)( num->Random() * l );
            int temp = permutation[j];
            permutation[j] = permutation[l];
            permutation[l] = temp;
        }
        
        // Randomly choose a state transition to update
        whichRateFrom  = ii;
		
        for (int jj=0; jj < numStates; jj++) {
            whichRateTo    = permutation[jj];

            // Skip the diagonal
            if ( whichRateTo == whichRateFrom ) continue;
            
            // Set bounds for mixture model components
            if (whichModel == -1) {	modelStart = 0; modelEnd = numModels; } else { modelStart = whichModel; modelEnd = modelStart+1; }
            
            // Update the same state transition's rate parameter for each specific mixture component
            for (int theModel=modelStart; theModel < modelEnd; theModel++) {
                count++;
                
                // Set the starting rate
                double oldRate = mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo];
                
                // Set the counts for the current state transition
                double cxy_1 = (*CXY).mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo];
                double c1_1  = (*C1).mRows.mDataArray[ theModel ].mDataArray[whichRateFrom];
                
                // Sanity check on the counts
                if (c1_1 < 0) { std::cout << "      ERROR: c1_1 is < 0!! " << c1_1 << endl; }
                if ( cxy_1 > 100000000000) { cout << "  ERROR?  We probably shouldn't have " << cxy_1 << " x->y substitutions!" << endl; }
                
                // Set Delta (maximum allowed value given mu and LambdaXp)
                double LambdaX = -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom];
                double Delta = (*mu).mDataArray[theModel] - LambdaX + oldRate;
                
                // Sample from the conditional posterior distribution given beta prior (beta posterior)
                double newValue = num->betaVariate( (cxy_1 + betaPrior_alpha ), (c1_1 + betaPrior_beta )) * ( Delta );
                
                // In case of numerical error:
                //            if (newValue < 0) newValue = 0.0;
                
                // Putting a floor on rates seems important for avoiding underflow that kills the stationary frequency calculations...
                // if (newValue <= 1e-7) newValue = 1e-7;
                
                // Update the alpha matrix (ignore the diagonal since this is auto. set in another function - ?)
                if (options.scaleByFrequencies) {
                    double stateFreq = stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
                    mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo]   = newValue / stateFreq;
                    mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += (oldRate - newValue) / stateFreq;
                } else {
                    mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = newValue ;
                    mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate - newValue;
                }
                
                // Update the rate matrix and its diagonal
                mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = newValue;
                mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate - newValue;
            }
        }
    }

    
    // Random updates:
    /*
	// Perform enough updates to do the entire matrix on average
    for (int numUpdates = 0; numUpdates < (numStates * (numStates-1)); numUpdates++) {
        
        // Randomly choose a state transition to update
        whichRateFrom  = (int)(num->Random() * (numStates));
		whichRateTo    = (int)(num->Random() * (numStates));
		
        // Make sure we're not on the diagonal
		while (whichRateTo == whichRateFrom) {
			whichRateTo = (int)(num->Random() * (numStates));
		}
		
        // Set bounds for mixture model components
        if (whichModel == -1) {	modelStart = 0; modelEnd = numModels; } else { modelStart = whichModel; modelEnd = modelStart+1; }
		
        // Update the same state transition's rate parameter for each specific mixture component
		for (int theModel=modelStart; theModel < modelEnd; theModel++) {
			count++;
            
            // Set the starting rate
            double oldRate = mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo];
            
            // Set the counts for the current state transition
			double cxy_1 = (*CXY).mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo];
            double c1_1  = (*C1).mRows.mDataArray[ theModel ].mDataArray[whichRateFrom];
            
            // Sanity check on the counts
            if (c1_1 < 0) { std::cout << "      ERROR: c1_1 is < 0!! " << c1_1 << endl; }
			if ( cxy_1 > 100000000000) { cout << "  ERROR?  We probably shouldn't have " << cxy_1 << " x->y substitutions!" << endl; }
            
            // Set Delta (maximum allowed value given mu and LambdaXp)
			double LambdaX = -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom];
			double Delta = (*mu).mDataArray[theModel] - LambdaX + oldRate;
            
            // Sample from the conditional posterior distribution given beta prior (beta posterior)
            double newValue = num->betaVariate( (cxy_1 + betaPrior_alpha ), (c1_1 + betaPrior_beta )) * ( Delta );
            
            // In case of numerical error:
//            if (newValue < 0) newValue = 0.0;
            
            // Putting a floor on rates seems important for avoiding underflow that kills the stationary frequency calculations...
            // if (newValue <= 1e-7) newValue = 1e-7;
            
            // Update the alpha matrix (ignore the diagonal since this is auto. set in another function - ?)
            if (options.scaleByFrequencies) {
                double stateFreq = stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
                mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo]   = newValue / stateFreq;
                mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += (oldRate - newValue) / stateFreq;
            } else {
                mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = newValue ;
                mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate - newValue;
            }
            
            // Update the rate matrix and its diagonal
            mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = newValue;
            mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate - newValue;
		}
	}
    */
    
    // cout << "Time: " << ((double)clock() - start ) / CLOCKS_PER_SEC << endl;
    
    // Return number of rate parameters updated
	return count;
}

// ****************************************************************************************
// MARK: Samplers: Gibbs update mechanisms
// ****************************************************************************************
int MCMC::Gibbs_updateAllAncestralAndTransientStates_local(
                                                           int& rannode,
                                                           MixtureModel& theModel,
                                                           
                                                           DoubleMatrix& stateFreqMatrix) {
    
        // NEW: combined branchlength update with ancestral/transient state update following
        
        // Adding stateFreqMatrix here because sometimes we want to call with different frequencies at the root of the tree...
        //	return mTree.ResampleAncestralAndTransientStates_local(rannode, theModel, theData, m_vecSiteAssignment, stateFreqMatrix );
        
        int numUpdated_local = 0;
        
        // Stationary frequencies are also updated every generation by call from main loop (PLEX.cpp)
        // mModel.forceUpdateStationaryFreqs();
        
        
        // When branchlengths are free, we need to keep the rate matrix and uniformization constant scaled to avoid non-identifiability
        // NOTE: we are overriding the capUniformizationRate flag here.
		double theRateFactor = mModel.getRateSum( mTree ) / mModel.GetNumModels();
		mModel.scaleDiscreteRates( theRateFactor ) ;
		mModel.scaleRateMatrices( theRateFactor );

	 //   if (options.__currentGen < 10000) {
			int last = options.capUniformizationRate;
			options.capUniformizationRate = 0;
			mModel.forceUpdateUniformization();
			
			options.capUniformizationRate = last;
	 //  }
		
		// [diagnostic output] cout << "           UPDATED MU to " << (*mModel.returnUniformizationConstants())[0] << endl;

        
        int beforeOrAfter = 0;
        if ( num->Random() < 0.5 ) {
            beforeOrAfter = 1;
        }
        
        if (beforeOrAfter == 0) {
            //  for (int ii=0; ii<5; ii++) {
            // **************************************************************************************************
            // Branch-length updater
            // **************************************************************************************************
            if (options.updatebls && updateCount > 5) {  
                int maxIter = 1;
                // Don't update every time...
                if ( options.__currentGen >= 2500 ) {
                    maxIter  = 3; //0;
                }
                
                for (int ii=0; ii<maxIter; ii++) {
                    // Start node shouldn't make a difference (no dependence between bl updates conditional on augmented data)
                    int rannode2 = 0; //( num->Random() * options.m_nNumBranches);
                    
					tree_countChanges(theModel, 1);
                    
                    // mode 2 is conjugate gibbs; mode 1 is CG
                    tree_updateBranchLengths( rannode2, theModel, 1, 0.1 );
                    
                    numUpdated_local += mTree.branches();
                    
                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Branch-lengths updated. Tree-length: " << setprecision(4) << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << "; inflation factor: " << setprecision(3) << (double)((double)mTree.root()->CountBranches()/(double)options.__initialNumBranches) << "X (max. segment-size: " << options.maxBL <<") ]" << endl;
                    
                }
				
				tree_countChanges(theModel, 1);
				
            }
        }
        
        // These are currently required by updater, but need not be
        setAllProbMatricesOnTree( mModel );
        
        if (  options.scaleByFrequencies ) {
            numUpdated_local += mTree.ResampleAncestralAndTransientStates_local( rannode, theModel, theData, m_vecSiteAssignment, stateFreqMatrix );
        } else {
            numUpdated_local += mTree.ResampleAncestralAndTransientStates_local( rannode, theModel, theData, m_vecSiteAssignment, stateFreqMatrix );
        } 
        
        tree_countChanges(theModel);
        
        if (beforeOrAfter == 1) {
            //  for (int ii=0; ii<5; ii++) {
            // **************************************************************************************************
            // Branch-length updater
            // **************************************************************************************************
            if (options.updatebls && updateCount > 5) {  
                int maxIter = 1;
                // Don't update every time...
                if ( options.__currentGen >= 2500 ) {
                    maxIter  = 3; //0;
                }
                                
                for (int ii=0; ii<maxIter; ii++) {
                    tree_countChanges(theModel, 1);
                    
                    // Start node shouldn't make a difference (no dependence between bl updates conditional on augmented data)
                    int rannode2 = 0; //( num->Random() * options.m_nNumBranches);
                    
                    // mode 2 is conjugate gibbs; mode 1 is CG
                    tree_updateBranchLengths( rannode2, theModel, 1, 0.1 );
                    
                    numUpdated_local += mTree.branches();
                    
                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Branch-lengths updated. Tree-length: " << setprecision(4) << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << "; inflation factor: " << setprecision(3) << (double)((double)mTree.root()->CountBranches()/(double)options.__initialNumBranches) << "X (max. segment-size: " << options.maxBL <<") ]" << endl;
                    
                }
				
				tree_countChanges(theModel, 1);
				
            }
        }
        
        return numUpdated_local;
}

int MCMC::Gibbs_updateAllAncestralAndTransientStates_full( MixtureModel& theModel ) {
    // NEW: combined branchlength update with ancestral/transient state update following
    
    // Adding stateFreqMatrix here because sometimes we want to call with different frequencies at the root of the tree...
    //	return mTree.ResampleAncestralAndTransientStates_local(rannode, theModel, theData, m_vecSiteAssignment, stateFreqMatrix );
    
    int numUpdated_local = 0;
    
    // Stationary frequencies are also updated every generation by call from main loop (PLEX.cpp)
    // mModel.forceUpdateStationaryFreqs();
    

        // When branchlengths are free, we need to keep the rate matrix and uniformization constant scaled to avoid non-identifiability
        // NOTE: we are overriding the capUniformizationRate flag here.
        if ( ( options.updatebls != 1 )  ) {
            double theRateFactor = mModel.getRateSum( mTree ) / mModel.GetNumModels();
            mModel.scaleDiscreteRates( theRateFactor ) ;
            mModel.scaleRateMatrices( theRateFactor );
            
           //  if (options.__currentGen < 10000) {
                int last = options.capUniformizationRate;
                options.capUniformizationRate = 0;
                mModel.forceUpdateUniformization();
                
                options.capUniformizationRate = last;
           // }
            
            // [diagnostic output] cout << "           UPDATED MU to " << (*mModel.returnUniformizationConstants())[0] << endl;
        } else if (options.updatebls) {
      
            
            double theRateFactor = mModel.getRateSum( mTree ) / mModel.GetNumModels();
            mModel.scaleDiscreteRates( theRateFactor ) ;
            mModel.scaleRateMatrices( theRateFactor );
            
          //  if (options.__currentGen < 10000) {
                int last = options.capUniformizationRate;
                options.capUniformizationRate = 0;
                mModel.forceUpdateUniformization();
                
                options.capUniformizationRate = last;
          //  }

        }

    int beforeOrAfter = 0;
    if ( num->Random() < 0.5 ) {
        beforeOrAfter = 1;
    }
    
    if (beforeOrAfter == 0) {
      //  for (int ii=0; ii<5; ii++) {
            // **************************************************************************************************
            // Branch-length updater
            // **************************************************************************************************
            if (options.updatebls && updateCount > 5) {  
                int maxIter = 1;
                // Don't update every time...
                if ( options.__currentGen >= 2500 ) {
                    maxIter  = 3; //0;
                }
                
                for (int ii=0; ii<maxIter; ii++) {
                    tree_countChanges(theModel, 1);
                    
                    // Start node shouldn't make a difference (no dependence between bl updates conditional on augmented data)
                    int rannode2 = 0; //( num->Random() * options.m_nNumBranches);

                    // mode 2 is conjugate gibbs; mode 1 is CG
                    tree_updateBranchLengths( rannode2, theModel, 1, 0.1 );
                    
                    numUpdated_local += mTree.branches();
                    
                    if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Branch-lengths updated. Tree-length: " << setprecision(4) << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << "; inflation factor: " << setprecision(3) << (double)((double)mTree.root()->CountBranches()/(double)options.__initialNumBranches) << "X (max. segment-size: " << options.maxBL <<") ]" << endl;
                    
                }

				tree_countChanges(theModel, 1);
				
            }
    }
    
    // These are currently required by updater, but need not be
    setAllProbMatricesOnTree( mModel );
    
    if (  options.scaleByFrequencies || mModel.getReversible() ) {
        numUpdated_local += mTree.AssignAncestralStatesByMarginalProbability( theModel, theData, m_vecSiteAssignment, theModel.returnStateFrequencies(), 1);
    } else {
        numUpdated_local += mTree.AssignAncestralStatesByMarginalProbability( theModel, theData, m_vecSiteAssignment, theModel.returnStationaryFrequencies(), 1);
    } 
    
    tree_countChanges(theModel);
    
    if (beforeOrAfter == 1) {
        //  for (int ii=0; ii<5; ii++) {
        // **************************************************************************************************
        // Branch-length updater
        // **************************************************************************************************
        if (options.updatebls && updateCount > 5) {  
            int maxIter = 1;
            // Don't update every time...
            if ( options.__currentGen >= 2500 ) {
                maxIter  = 3; //0;
            }
            
            for (int ii=0; ii<maxIter; ii++) {
                tree_countChanges(theModel, 1);
                
                // Start node shouldn't make a difference (no dependence between bl updates conditional on augmented data)
                int rannode2 = 0; //( num->Random() * options.m_nNumBranches);
                
                // mode 2 is conjugate gibbs; mode 1 is CG
                tree_updateBranchLengths( rannode2, theModel, 1, 0.1 );
                
                numUpdated_local += mTree.branches();
                
                if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Branch-lengths updated. Tree-length: " << setprecision(4) << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << "; inflation factor: " << setprecision(3) << (double)((double)mTree.root()->CountBranches()/(double)options.__initialNumBranches) << "X (max. segment-size: " << options.maxBL <<") ]" << endl;
                
				tree_countChanges(theModel, 1);
            }
        }
    }
    
    return numUpdated_local;
}


int MCMC::Gibbs_updateSiteAssignments_MixtureComponents(
                                                        MixtureModel& mModel
                                                        ) {
	return mTree.Gibbs_updateSiteAssignments_MixtureComponents( mModel, theData, m_vecSiteAssignment );	
}

void MCMC::Gibbs_updateSiteAssignments_RateAssignments(
                                                       MixtureModel& mModel
                                                       
                                                       )

{
    
    // Skip as check...
    // if ( options.__currentGen > 10000 ) return;
    
    // Allow blocking of this function when global override flag is set...
	if (options.__overrideSiteResamp == 0) {
        mTree.KEEP_Gibbs_updateSiteAssignments_RateAssignments( mModel, theData, m_vecSiteAssignment );	
    }
}

int MCMC::Gibbs_updateRateParameterAssignments_RAR_B1u(
                                                       ProteinModel& mModel, 
                                                       
                                                       double updateFraction, int whichModel) {
	// ********************************************************************************************************
	// Resample rate parameter assignments to state exchanges under RAR models and B1u partial sampling
	// ********************************************************************************************************
	if ( (options.m_nMethod != 6)) { std::cout << "ERROR! ParameterSweep only implemented for B1-type UNIFORMIZED methods at present" << endl; exit(-1); }
	
	// This function is for discrete-distributed rates only (for now)!
	// [updating this to use the updateFraction as an upper limit]
	
	// Apr 2010: re-writing to never exceed the current uniformization constant
	
	std::vector<double> *theRates;
	std::vector<int> *rateAssignments;
	
	int theModel, i,j,k, startSite;
	int startingState;
	int newState;
	double theFraction;
	double logLikelihood2;
	
	int numStates = mModel.GetRateMatrix3D().cols();
	int numModels = mModel.GetRateMatrix3D().rows();
	int numRates = mModel.returnNumRates();
	int allStatesInvalid = 0;
	double marginalConditional[(int)numRates];
	double rannum;
	
	theRates = mModel.returnDiscreteRates();
	
	if (updateFraction < 1.0) { theFraction = num->Random() * updateFraction; } else { theFraction = updateFraction; }
	
	int modelStart, modelEnd;
	if (whichModel == -1) {
		modelStart = 0; modelEnd = numModels;
	} else {
		modelStart = whichModel; modelEnd = modelStart+1;
	}
	
	int jj;
	int numUpdated_local = 0;
	
	int accept, reject;
	accept = reject = 0;
	
	for (theModel=modelStart; theModel < modelEnd; theModel++) {
		for (int ii=0; ii<numStates*1.1; ii++) {
			if (mModel.getReversible()) { startSite = i; } else { startSite = 0; }
			
			// Randomly select row, i
			i = int(num->Random() * numStates + startSite);
			
			// Update more than every state transition
			for (jj=startSite; jj<(int)(numStates*1.4); jj++) {
				// Randomly select a j
				j = int(num->Random() * (numStates-startSite)) + startSite;
				
				if ( (j != i) ) {
					if (theFraction != 1.0) {
						rannum = num->Random(); // [JdK] Numerics::getGlobalNumerics().Random();
						if (rannum > theFraction) continue;
					}
					numUpdated_local++;
					
					// WE COULD ADD SOMETHING USEFUL HERE -- IF THERE ARE ZERO COUNTS THEN RANDOMLY SELECT A RATE ASSIGNMENT!  DONT HAVE TO DO THE MARGINALS
					rateAssignments = mModel.returnRateAssignments();
					
					startingState = (*rateAssignments)[ ((theModel * numStates * numStates) + (i * numStates) + j) ];
					DoubleVector *mu = mModel.returnUniformizationConstants();
					
					if (  -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[i] > (*mu).mDataArray[theModel]) { 
						std::cout << "ERROR!  Before resampling, LambdaX for state " << i << " is " << (-mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[i]) << ", while mu is " << (*mu).mDataArray[theModel] << endl;
					}
					
					double remainderOfLambdaX = -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[i].mDataArray[i] - (*theRates)[ (theModel * mModel.returnNumRates() ) + startingState ];
					
					allStatesInvalid = 1;
					// Evaluate change in likelihood by reassigning state transition (i,j) to all discrete rate parameters, k
					for (k=0;k<numRates; k++) {
						if ( ( remainderOfLambdaX + (*theRates)[ (theModel * mModel.returnNumRates() ) + k ] ) > (*mu).mDataArray[theModel] ) {
							// Skip this state
							marginalConditional[k] = -std::numeric_limits<double>::max( );
							reject++;
                            //	cout << "	Rejecting state " << k << "; lambda remainder is " << remainderOfLambdaX << " while rate is " << (*theRates)[k] << endl;
						} else {
							allStatesInvalid = 0;
							
							mModel.UpdateRateAssignmentB1(theModel,i,j,k);
							logLikelihood2 = CalculateCompleteDataLogLikelihood_Delta_B1u(mModel,  theModel, i, j);
							
							marginalConditional[k] = logLikelihood2;
							//cout << "Exchange " << i << ", " << j << "; assingment to " << k << " gives " << marginalConditional[k] << endl;
							accept++;
						}
					}
					//exit(-1);
					
					// Sometimes the choices at one row state assignment can invalidate all choices at the current state...
					if ( allStatesInvalid == 1) {
						jj -= 1;
						std::cout << "	[WARNING: all states for exchange " << i << ", " << j << " are now invalid.  Trying elsewhere in the row.]" << endl;
						mModel.UpdateRateAssignmentB1(theModel,i,j,startingState);
						continue;
					}
					
					newState = _util_sampleFromCumulative(&marginalConditional[0], numRates);
					while ( marginalConditional[ newState ] == -std::numeric_limits<double>::max( ) ) {
						std::cout << "	[ Rare error: we chose a restricted state!  Lazy programmer! ]" << endl;
						newState = _util_sampleFromCumulative(&marginalConditional[0], numRates);
					}						
					mModel.UpdateRateAssignmentB1(theModel,i,j,newState);
				}
			}
		}
	}
    
    //	if ( ( accept/(accept+reject) ) <= 0.5 ) {
    //		std::cout << "		[ Possible problem in RAR resampler; ratio of in-bounds and out-bounds proposals (due to uniformization constraint): " << ( accept/(accept+reject) ) << " ]" << endl;
    //	}
	return numUpdated_local;
}

int MCMC::Gibbs_updateRateParameterAssignments_RAR_DRA_B1u( ProteinModel& mModel, 
                                                           
                                                           double updateFraction, int whichModel) {
    
    // * Don't forget about likelihood heating *
    // This version is obsolete.  Plug-in new version.
    
    std::cout << "Gibbs_updateRateParameterAssignments_RAR_DRA_B1u() is obsolete.  Plug-in new version!" << endl; return 0;
    
    /*
     static ProteinModel tempModel1;
     static ProteinModel tempModel2;
     
     tempModel1 = mModel;
     tempModel2 = mModel;
     int numUpdated = 0;
     
     //	double logLR;
     if (whichModel != -1) { std::cout << "ERROR!  _dependentassignment only works for whichModel == -1 for now" << endl; exit(-1); }	
     if ( (options.m_nMethod != 1) && (options.m_nMethod != 3)) { std::cout << "ERROR! ParameterSweep only implemented for B1 at present" << endl; exit(-1); }
     
     int theModel, i,j,k, startSite;
     int startingState;
     std::vector<double> *theRates;
     std::vector<int> *rateAssignments;
     //	double windowsize = mModel.GetUpdateWindowsize();
     double logLikelihood1, logLikelihood2;
     
     int newState;
     int numStates = mModel.GetRateMatrix3D().cols();
     int numModels = mModel.GetRateMatrix3D().rows();
     
     int numRates = mModel.returnNumRates();
     
     theRates = mModel.returnDiscreteRates();
     
     double theFraction;
     
     if (updateFraction < 1.0) { theFraction = num->Random() * updateFraction; } else { theFraction = updateFraction; }
     
     double LRsum;
     int theDim = (int)numRates;
     
     double rannum;
     
     int modelStart, modelEnd;
     if (whichModel == -1) {
     modelStart = 0; modelEnd = numModels; 
     } else {
     modelStart = whichModel; modelEnd = modelStart+1;
     }
     
     double marginalConditional[(int)numRates*numModels];
     double marginalConditional2[(int)numRates];
     double marginalConditional3[(int)numRates];
     double marginalConditional4[(int)numRates];
     
     if (whichModel == -1) { theDim = numRates; } else { theDim = 1; }
     
     //int numUpdates = 0;
     // Iterate over all relevant i->j's (taking care to not be redundant on symmetric models)
     for (i=0; i<numStates; i++) {
     if (mModel.getReversible()) { startSite = i; } else { startSite = 0; }
     for (j=startSite; j<numStates; j++) {
     // Don't do all of them, if theFraction is < 1
     if (theFraction != 1.0) {
     rannum = num->Random(); 
     if (rannum > theFraction) continue;
     }
     //numUpdates++;
     
     // Now iterate over all models and calculate the deltaLnL...
     for (theModel=modelStart; theModel < modelEnd; theModel++) {
     //if ( (j==0) ) {
     numUpdated++;
     
     logLikelihood1  = CalculateCompleteDataLogLikelihood_Delta_B1(tempModel1, m_vecSiteAssignment, theModel, i);
     
     if ( j != i ) {
     // Current state (just evaluated and stored as logLikelihood1)
     rateAssignments = tempModel2.returnRateAssignments();
     startingState = (*rateAssignments)[ ((theModel * numStates * numStates) + (i * numStates) + j) ];
     
     // Now do the rest of them...
     LRsum = 0.0;
     for (k=0;k<numRates; k++) {
     if (k != startingState) {
     tempModel2 = tempModel1;
     
     tempModel2.UpdateRateAssignmentB1(theModel,i,j,k);
     // [full eval] logLikelihood2a = CalculateCompleteDataLogLikelihood(tempModel2, m_vecSiteAssignment,);
     logLikelihood2 = CalculateCompleteDataLogLikelihood_Delta_B1(tempModel2, m_vecSiteAssignment, theModel, i);
     marginalConditional[(theModel-modelStart)*theDim + k] = logLikelihood2;
     } else {
     marginalConditional[(theModel-modelStart)*theDim + k] = logLikelihood1;
     }
     }
     } // j!=i
     } // for models..
     
     // These are fixed things we need whether we have dependencies or not.  Can reuse!
     for (theModel=0; theModel < numModels; theModel++) {
     // Now for each model, go through and collect the dependent rates stuff for #0, then: 1) assign; 2) sample...
     // [DO #0, and shared assignments LAST]
     if ( j != i ) {
     if ( ( theModel == 0 ) && ( mModel.getDependentRates() ) ) {
     // If we have dependent rates, go through all the non-zero models to collect the contribution of 'non-zero' mixture components to this state transitions marginal rate distribution...
     for (k=0; k<numRates; k++) {
     marginalConditional2[k] = marginalConditional[k];
     }
     
     for (int j1=1; j1 < numModels; j1++) {
     if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
     for (k=0; k<numRates; k++) {
     marginalConditional2[k] += marginalConditional[(j1-modelStart)*theDim + k];
     }
     } // if dependency
     } // for models
     } // if model==0 and dependent rates
     
     // NOW, set marginalConditional3 for the current theModel...
     double maxL = -std::numeric_limits<double>::max( );
     for (k=0;k<numRates; k++) {
     marginalConditional3[k] = marginalConditional[(theModel-modelStart)*theDim + k];
     if (marginalConditional3[k] > maxL) { maxL = marginalConditional3[k]; }
     }
     
     // Get LRs...
     LRsum=0.0;
     double tempVal;
     for (k=0;k<numRates; k++) {
     tempVal = exp((marginalConditional3[k] - maxL));
     marginalConditional3[k] = tempVal;
     LRsum += marginalConditional3[k];
     }
     
     // Finish transformation...
     for (k=0;k<numRates; k++) {
     marginalConditional3[k] /= LRsum;
     }
     
     // MarginalConditional3 now contains the posterior prob's for rate assignments at model 'theModel' (assuming no dependency)
     
     //  Now set marginalConditional4 based on all the OTHER dependent rates (excluding me, if i'm dependent)...
     if ( (theModel > 0) && (tempModel1.dependencyIndicators[theModel][i][j] == 1) ) {
     for (k=0; k<numRates; k++) {
     marginalConditional4[k] = marginalConditional2[k] - marginalConditional[(theModel-modelStart)*theDim + k];
     }
     } // if dependency
     else if (theModel >0)  {
     for (k=0; k<numRates; k++) {
     marginalConditional4[k] = marginalConditional2[k]; // - marginalConditional[(theModel-modelStart)*theDim + k];
     }
     }
     
     if (theModel > 0) {
     // Transform marginalconditional4 to probabilities...
     tempVal = 0.0;
     maxL = -std::numeric_limits<double>::max( );
     for (k=0;k<numRates; k++) {
     if (marginalConditional4[k] > maxL) { maxL = marginalConditional4[k]; }
     }
     
     // Get LRs...
     LRsum = 0.0;
     for (k=0;k<numRates; k++) {
     tempVal = exp((marginalConditional4[k] - maxL));
     marginalConditional4[k] = tempVal;
     LRsum += marginalConditional4[k];
     }
     
     // Finish transformation...
     for (k=0;k<numRates; k++) {
     marginalConditional4[k] /= LRsum;
     }
     
     // MarginalConditional4 now contains the posterior prob's for rate assignments across all the other dependent categories
     }
     
     double pSame, pDiff, probDiff;
     pSame = pDiff = 0.0;
     probDiff = 0.0;
     
     double differencePrior = mModel.getIndependentFractionPrior();
     
     // Calculate the probability of having the same rate assignment as model #0...
     if (theModel > 0) {
     for (int i1=0; i1<numRates; i1++) {
     for (int j1=0; j1<numRates; j1++) {
     if (j1==i1) { pSame += ( marginalConditional4[i1] ) * (marginalConditional3[j1]); }
     else { pDiff += ( marginalConditional4[i1] * marginalConditional3[j1] ) ;  }
     }
     }
     
     probDiff = ( pDiff * differencePrior ) / ( (pDiff * differencePrior) + (pSame * (1.0-differencePrior) ) );
     //		std::cout << "model " << theModel << ", for " << i << " -> " << j << ", the probability of being different is " << probDiff << "; pdiff is " << pDiff << ", pSame is " << pSame << endl;
     
     double u = num->Random();
     if (u <= probDiff) {
     tempModel1.dependencyIndicators[theModel][i][j] = 0;
     
     double rannum = num->Random();; // [JdK] Numerics::getGlobalNumerics().Random();
     
     newState = 0;
     //double thebest = 0.0; int best = 0;
     
     for (int ii=0; ii<numRates;ii++) {
     if (ii>0) { marginalConditional3[ii] += marginalConditional[ii-1]; }
     }
     
     for (int ii=0; ii<numRates;ii++) {
     if (rannum <= marginalConditional3[ii]) {newState=ii; break;}
     }
     
     // Update stuff...
     tempModel1.UpdateRateAssignmentB1(theModel,i,j,newState);
     //[Don't need to do this?] logLikelihood1 = CalculateCompleteDataLogLikelihood_Delta_B1(tempModel1, m_vecSiteAssignment, theModel, i);
     
     } else {
     tempModel1.dependencyIndicators[theModel][i][j] = 1;
     }
     }
     } // j!=i
     } // for themodel
     
     // NOW sample the dependent guys...
     int numDependent = 0;
     if ( j != i ) {
     for (int j1=1; j1 < numModels; j1++) {
     if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
     numDependent++;
     }
     }
     
     // special case: model==0 and NO dependent rates...
     if (numDependent == 0) {
     newState = _util_sampleFromCumulative(&marginalConditional[0],numRates);
     tempModel1.UpdateRateAssignmentB1(0,i,j,newState);
     //tempModel1.ProteinModelParameterstoRates();
     } else {
     for (k=0; k<numRates; k++) {
     marginalConditional2[k] = marginalConditional[k];
     }
     
     for (int j1=1; j1 < numModels; j1++) {
     if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
     for (k=0; k<numRates; k++) {
     marginalConditional2[k] += marginalConditional[(j1-modelStart)*theDim + k];
     }
     } // if dependency
     } // for models
     
     // NOTE: _util_sampleFromCumulative now assumes LOG liks
     newState = _util_sampleFromCumulative(&marginalConditional2[0],numRates);
     tempModel1.UpdateRateAssignmentB1(0,i,j,newState);
     
     for (int j1=1; j1 < numModels; j1++) {
     if (tempModel1.dependencyIndicators[j1][i][j] == 1) {
     tempModel1.UpdateRateAssignmentB1(j1,i,j,newState);
     } // if dependency
     } // for models
     //					tempModel1.ProteinModelParameterstoRates();
     
     
     }
     } // j!=i
     
     // STEP 2.  IF (THEMODEL > 0) NOW NEED TO DRAW u TO DETERMINE IF THIS MODEL COMPONENT WILL BE SHARED OR NOT.  SET INDICATOR APPROPRIATELY
     
     // STEP 2A.  IF SHARED, RECALCULATE THE POSTERIOR DISTRIBUTION BY SUMMING THE APPROPRIATE LOG-L VALUES
     
     // STEP 3.  IF (INDICATOR SAYS DEPENDENT || MODELCOMPONENT == 0) SAMPLE FROM POSTERIOR AND SET NEW STATE
     } // j
     } // i
     
     //	std::cout << "FINISHED FULL PARAMETER SWEEP..." << endl;
     
     mModel = tempModel1;
     m_bUpdate = true;
     
     return numUpdated;
     */
}

// ****************************************************************************************
// MARK: Samplers: Metropolis updates
// ***************************************************************************************
int MCMC::MC_updateASRV_shape(
							  ProteinModel& mModel, 
							  
							  double var) {
	
    // WARNING: SYMMETRIC PROPOSAL?  CHECK BOUNDARY CONDITIONS
    
	const int RATE    = 1;
	
    // Undiagnosed problem: if shape parameter goes low, distribution becomes uniform (expected). These moves are often accepted, and produce wild drops in incomplete-data logL (unexpected).
    double shapeParameterFloor = 0.008; 
    
	double logLikelihood1a, logLikelihood2;
	double logLR;
	
	static int    unibackupAlloc = 0;
	static double *backupUni;
	if (unibackupAlloc == 0) {
		backupUni = new double[mModel.GetNumModels()];
		unibackupAlloc = 1;
	}
	
	// ***************************************************************************************************
	// Get the likelihoods for the starting and ending states
	// ***************************************************************************************************
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(0);
	
	double proposal;
	double startShape = mTree.root()->GetData().siteCats->getShapeParm();
	
    proposal = -startShape - 5;
	while ( (startShape + proposal) <= 0.0 ) {
		proposal   = num->gaussianVariate(0, var );
	}
	
	int uniChanged = 0;
	
	if (options.m_nMethod == 6) {
		uniChanged = 0;
		// Backup uniformization constants...
		for (int ii=0; ii<mModel.GetNumModels(); ii++) {
			backupUni[ii] = mModel.getUniformization(ii);
		}
		
        // [block - new approach]  if ( (startShape + proposal) < shapeParameterFloor) proposal = shapeParameterFloor - startShape;
        
		if (mTree.root()->GetData().siteCats->setGammaDistribution( (startShape ) ) > 0 ) {
            cout << "ERROR: starting value of gamma shape parameter " << mTree.root()->GetData().siteCats->getShapeParm() << " is not allowed!" << endl;
            exit(-1);
        }
 	}
	
	// Model 1 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood1a = CalculateCompleteDataLogLikelihood(mModel, RATE);
	
    int rejectcount=0;
	int resampleAssignments = 0;
    
    int notDone = 1;
	while ( notDone ) {
        proposal = -startShape;
        while ( (startShape + proposal) <= 0.0 ) {
            proposal   = num->gaussianVariate(0, var );
            rejectcount++;
            if (rejectcount > 100) { cout << "      * WAITING FOR VALID GAMMA SHAPE PROPOSAL *" << endl; }
        }
        
        if ( mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) ) == 0 ) {
            notDone = mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) );
            //cout << "   OKAY, we've got one: " << mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) ) << endl;
        } else {
            notDone = 1;
        }
    }
    
	// Sometimes resample the site assignments given the proposal distribution
    if (num->Random() < 0.1 ) {
        resampleAssignments = 1;
        mTree.root()->GetData().siteCats->backupSiteAssignments();
        Gibbs_updateSiteAssignments_RateAssignments(mModel );	
    }
    
    // Model 2 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood2 = CalculateCompleteDataLogLikelihood(mModel, RATE);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	// Got the log-likelihood ratio: corrected from previous
    //logLR = (logLikelihood1a - logLikelihood2);
    logLR = (logLikelihood2 - logLikelihood1a);
    
	// std::cout << "		[Gamma shape proposal: LR is " << logLR << "; from " << startShape << " -> " << (startShape + proposal) << " ]" << endl;
    //std::cout << "       lnL1 = " << logLikelihood1a << ", lnL2 = " << logLikelihood2 << endl;
    
	// ***************************************************************************************************
	
	// Accept / reject?
	bool m_bUpdate = false;
	if (logLR >= 0.0 )
	{
        //        cout << "       ACCEPT" << endl;
		m_bUpdate = true;
	} else {
		
		if (num->evaluateProposal_logL( logLR ) ) {
            //            cout << "       ACCEPT" << endl;
 			m_bUpdate = true;
		} else {
			m_bUpdate = false;
		}
	}
    
	if (m_bUpdate == false) {
        //        cout << "       REJECT" << endl;
		mTree.root()->GetData().siteCats->setGammaDistribution( startShape );
        
		if (resampleAssignments == 1) {
			if (options.m_nMethod == 6) {
				for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
					mTree.root()->reassignSiteCountsBetweenRateCategories_Uni(ii, mTree.root()->GetData().siteCats->getAssignment(ii), mTree.root()->GetData().siteCats->getBackupAssignment(ii), theData, m_vecSiteAssignment );
				}
			}
			mTree.root()->GetData().siteCats->restoreSiteAssignments();
			
		}
		return 0;
	} else {
		//	std::cout << "	[Updated gamma shape from " << startShape << " to " << (startShape + proposal) << "]" << endl;
		return 1;
	}
	
}

int MCMC::MC_updateASRV_shape(
							  GTRModel& mModel, 
							  
							  double var) {
	
    // WARNING: SYMMETRIC PROPOSAL?  CHECK BOUNDARY CONDITIONS
    
    // Code removed here that allowed for changes to the uniformization constant.
    // Now keeping mu fixed.  (Safer, easier.)
    
	// const int BOTH    = 0;
	const int RATE    = 1;
	// const int BRANCH  = 2;
	
    // Undiagnosed problem: if shape parameter goes low, distribution becomes uniform (expected). These moves are often accepted, and produce wild drops in incomplete-data logL (unexpected).
    double shapeParameterFloor = 0.008; 
    
	double logLikelihood1a, logLikelihood2;
	double logLR;
	
	static int    unibackupAlloc = 0;
	static double *backupUni;
	if (unibackupAlloc == 0) {
		backupUni = new double[mModel.GetNumModels()];
		unibackupAlloc = 1;
	}
	
	// ***************************************************************************************************
	// Get the likelihoods for the starting and ending states
	// ***************************************************************************************************
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(0);
	
	double proposal;
	double startShape = mTree.root()->GetData().siteCats->getShapeParm();
	
	proposal = -startShape - 5;
	while ( (startShape + proposal) <= 0.0 ) {
		proposal   = num->gaussianVariate(0, (var * 2.0 ) );
	}
	
	int uniChanged = 0;
	
	if (options.m_nMethod == 6) {
		uniChanged = 0;
		// Backup uniformization constants...
		for (int ii=0; ii<mModel.GetNumModels(); ii++) {
			backupUni[ii] = mModel.getUniformization(ii);
		}
		
        // [block - new approach]  if ( (startShape + proposal) < shapeParameterFloor) proposal = shapeParameterFloor - startShape;
        
		if (mTree.root()->GetData().siteCats->setGammaDistribution( (startShape ) ) > 0 ) {
            cout << "ERROR: starting value of gamma shape parameter " << mTree.root()->GetData().siteCats->getShapeParm() << " is not allowed!" << endl;
            exit(-1);
        }
 	}
	
	// Model 1 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood1a = CalculateCompleteDataLogLikelihood(mModel, RATE);
	
    int rejectcount=0;
	int resampleAssignments = 0;
    
    int notDone = 1;
	while ( notDone ) {
        proposal = -startShape;
        while ( (startShape + proposal) <= 0.0 ) {
            proposal   = num->gaussianVariate(0, (var * 2.0 ) );
            rejectcount++;
            if (rejectcount > 100) { cout << "      * WAITING FOR VALID GAMMA SHAPE PROPOSAL *" << endl; }
        }
        
        if ( mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) ) == 0 ) {
            notDone = mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) );
            //cout << "   OKAY, we've got one: " << mTree.root()->GetData().siteCats->setGammaDistribution( (startShape + proposal) ) << endl;
        } else {
            notDone = 1;
        }
    }
    
	// Always resample the site assignments given the proposal distribution
    // NOTE: this seems very important for performance
    if ( num->Random() <= 1 ) {
		resampleAssignments = 1;
		
		mTree.root()->GetData().siteCats->backupSiteAssignments();
		Gibbs_updateSiteAssignments_RateAssignments(mModel );	
	} 
    
	// Model 2 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
	logLikelihood2 = CalculateCompleteDataLogLikelihood(mModel, RATE);
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	// Got the log-likelihood ratio
	logLR = (logLikelihood2 - logLikelihood1a);
	
    //	std::cout << "		[Gamma shape proposal: LR is " << logLR << "; from " << startShape << " -> " << (startShape + proposal) << " ]" << endl;
    //    std::cout << "       lnL1 = " << logLikelihood1a << ", lnL2 = " << logLikelihood2 << endl;
    
	// ***************************************************************************************************
	
	// Accept / reject?
	bool m_bUpdate = false;
	if (logLR >= 0.0 )
	{
        //        cout << "       ACCEPT" << endl;
		m_bUpdate = true;
	} else {
		
		if (num->evaluateProposal_logL( logLR ) ) {
            //            cout << "       ACCEPT" << endl;
 			m_bUpdate = true;
		} else {
			m_bUpdate = false;
		}
	}
    
	if (m_bUpdate == false) {
        //        cout << "       REJECT" << endl;
		mTree.root()->GetData().siteCats->setGammaDistribution( startShape );
        
		if (resampleAssignments == 1) {
			if (options.m_nMethod == 6) {
				for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
					mTree.root()->reassignSiteCountsBetweenRateCategories_Uni(ii, mTree.root()->GetData().siteCats->getAssignment(ii), mTree.root()->GetData().siteCats->getBackupAssignment(ii), theData, m_vecSiteAssignment );
				}
			}
			mTree.root()->GetData().siteCats->restoreSiteAssignments();
			
		}
		return 0;
	} else {
		//	std::cout << "	[Updated gamma shape from " << startShape << " to " << (startShape + proposal) << "]" << endl;
		return 1;
	}
	
}

int MCMC::MC_updateRateMatrix_nonRAR(
                                     ProteinModel& mModel
                                     
                                     ) {
	
    // Metropolis-Hasting updater of the rate matrix
    // Requires a backup/restore for all rate matrices
    bool m_bUpdate;
    
    // Special case handling:
    if (options.m_nMethod != 6) { cerr << "Running PLEX in legacy mode without B1u? Unsupported." << endl; exit(-1); }
    
    // Do state frequencies, if free
    if ( options.freeFreqs ) {
        MC_updateStateFreqs_B1u(mModel ); 
    }
    
    // For predefined rate matrices, update the scaling if needed
    if ( mModel.getPredefinedMatrix() > -1 ) { return MC_updateRateMatrixScaling_nonRAR_B1u( mModel ); }
    
    static DoubleMatrix3D rateMatrix;
    const int             RATE    = 1;
    
    if (rateMatrix.empty()) {
        rateMatrix.resize( mModel.GetNumModels(), mModel.GetNumStates(), mModel.GetNumStates(), 0.0 );
    }
    
	double  logLR;
	int     numModels = mModel.GetRateMatrix3D().rows();
	
	for (int whichModel = 0; whichModel < numModels; whichModel++) {	
		double logLikelihood1a, logLikelihood2;
		int numUpdated_local = 0;
		int tempUpdate = 0;
		
		Vector<double> *theFreqs;
		theFreqs = &mModel.returnStateFrequencies()[whichModel];
		
        // Get starting likelihood wrt rate parameters
        int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(0);
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		logLikelihood1a = CalculateCompleteDataLogLikelihood(mModel, RATE);
        
        // Backup the rate matrix in case of rejection
        mModel.backup3Dmatrix( mModel.GetNumModels(), mModel.GetNumStates(), *mModel.returnAlphaMatrix3D(), rateMatrix );
        
        // Update the rate matrix
        mModel.UpdateAlphaMatrix3D( options.m_nMethod, whichModel ); // block re-assignment of rates update
        mModel.ProteinModelParameterstoRates();
        
        // Don't know what this is.  For free scaling i guess...
        /*
         double scaleFactor = mModel.getRateSum(whichModel);
         
         double u, proposal;
         proposal = 0.0;
         
         // Make proposal
         do {
         u = num->gaussianVariate(0.0, windowsize);
         
         proposal = scaleFactor + u;
         
         } while (proposal <= 0.000001);
         
         mModel.scaleRateMatrices( whichModel, proposal );
         mModel.ProteinModelParameterstoRates();        
         */
        
        // Get proposal likelihood wrt rate parameters
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		logLikelihood2 = CalculateCompleteDataLogLikelihood(mModel, RATE);
		mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		
		// Got the likelihood ratio (assuming the larger of the two uniformization constants)...
		logLR = (logLikelihood2 - logLikelihood1a);		
        
		m_bUpdate = false;
        
		if (logLR >= 0.0 ) {
            // Accept
			numUpdated_local += tempUpdate;
			m_bUpdate = true;
            
		} else {
            // Check
			if (num->evaluateProposal_logL( logLR ) ) {
                // Accept
                numUpdated_local += tempUpdate;
				m_bUpdate = true;
			} else {
				// Reject
                m_bUpdate = false;
				numUpdated_local = 0;
                
                // Restore rate matrix
                mModel.backup3Dmatrix( mModel.GetNumModels(), mModel.GetNumStates(), rateMatrix, *mModel.returnAlphaMatrix3D() );
                mModel.ProteinModelParameterstoRates();
            }
		}
 	}	
	
	if (m_bUpdate) return 1;
	else return 0;
}

int MCMC::MC_updateRateMatrixScaling_nonRAR_B1u(
                                                ProteinModel& mModel
                                                
                                                ) {
    
	// Simple Metropolis-Hastings update of the rate-matrix scaling (if bl's are fixed)
    // Uses the B1u likelihood function
    
    if (options.m_nMethod != 6) { cerr << "Running PLEX in legacy mode without B1u? Unsupported." << endl; exit(-1); }
    if (options.updatebls) return 0;
    
    const int             RATE    = 1;
    static DoubleMatrix3D rateMatrix;
    
    if (rateMatrix.empty()) {
        rateMatrix.resize( mModel.GetNumModels(), mModel.GetNumStates(), mModel.GetNumStates(), 0.0 );
    }
    
	double  logLR;
	int     numModels = mModel.GetRateMatrix3D().rows();
	bool m_bUpdate;
    
	for (int whichModel = 0; whichModel < numModels; whichModel++) {	
		double logLikelihood1a, logLikelihood2;
		int numUpdated_local = 0;
		int tempUpdate = 0;
		
		Vector<double> *theFreqs;
		theFreqs = &mModel.returnStateFrequencies()[whichModel];
		
		double windowsize = mModel.GetUpdateWindowsize();
		
        // Get starting likelihood
        int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(0);
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		logLikelihood1a = CalculateCompleteDataLogLikelihood(mModel, RATE);
        
        // Backup rate and alpha matrix
        mModel.backup3Dmatrix( mModel.GetNumModels(), mModel.GetNumStates(), *mModel.returnRateMatrix3D(), rateMatrix );
        
        double scaleFactor = mModel.getRateSum(whichModel, mTree);
        
        double u, proposal;
        proposal = 0.0;
        
        // Make proposal
        do {
            u = num->gaussianVariate(0.0, windowsize);
            
            proposal = scaleFactor + u;
            
        } while (proposal <= 0.000001);
        
        // Also propose changes to the frequencies, if required
        if (!mModel.getFixedFreqs()) mModel.UpdateFrequencyMatrix();	
        
        mModel.scaleRateMatrices( whichModel, proposal );
        mModel.ProteinModelParameterstoRates();        
		
		// Model 2 likelihood... (only the parts that depend on RATE parameters)
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		logLikelihood2 = CalculateCompleteDataLogLikelihood(mModel, RATE);
        
		// Clean-up
        mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		
        // Get the logLR and evaluate
		logLR = (logLikelihood2 - logLikelihood1a);		
        
        m_bUpdate = false;
		if (logLR >= 0.0 ) {
            // Accept
			numUpdated_local += tempUpdate;
			m_bUpdate = true;
            
		} else {
            // Check
			if (num->evaluateProposal_logL( logLR ) ) {
                // Accept
                numUpdated_local += tempUpdate;
				m_bUpdate = true;
			} else {
				// Reject
                m_bUpdate = false;
				numUpdated_local = 0;
                
                // Restore rate and alpha matrix
                mModel.backup3Dmatrix( mModel.GetNumModels(), mModel.GetNumStates(), rateMatrix, *mModel.returnRateMatrix3D() );
                mModel.ProteinModelParameterstoRates();
            }
		}
        
        // Don't recalculate all the matrices; let functions that need updated matrices do it themselves
		/* 
         if (m_bUpdate == true) {
         mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
         } 
         */
	}	
	
	if (m_bUpdate) return 1;
	else return 0;
}

int MCMC::MC_updateRARDistributionParameters_B1u(
												 ProteinModel& mModel, 
												 double updateFraction, 
                                                 int whichModel) {
    
    // This function needs total replacement
    
	// ****************************************************************************************************************************
	// Update the RAR parameters using a Metropolis-Hastings step (for RAR models with B1u partial sampling)
	// ****************************************************************************************************************************
	
	// THIS VERSION UPDATES THE RATES PER CATEGORY WITH AN MC STEP...
	// std::cout << "	[Starting RAR distribution update...]" << endl;
	//return;
	
	//const int BOTH    = 0;
	const int RATE    = 1;
	//const int BRANCH  = 2;
	
	// This does a proposal affecting all models in the mixture.  Maybe not a good idea...
	//	return 0;
	
	double logLR;
	double logLikelihood1a, logLikelihood2;
	int numUpdated_local = 0;
	int tempUpdate = 0;
	if (!mModel.getRARdistributed()) { std::cout << "GammaShapeWithSweep update called when the model doesn't have gamma-stuff set!" << endl; exit(-1); }
	
    // Should make these static, unless they aren't cleared out properly on assignment
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	
	std::vector<double> *theRates1, *theRates2;
	theRates1 = tempModel1.returnDiscreteRates();
	theRates2 = tempModel2.returnDiscreteRates();
	
	//std::cout << "MODEL 2: " << endl;
	if (whichModel == -1) { std::cout << "ERROR UPDATE CALLED FOR WHICHMODEL -1" << endl; exit(-1); }
	
	double startMu, newMu;
	tempModel2.UpdateRateDistributionRAR_orderedDiscrete_MC(whichModel, 0.1); // last was 0.005 
	tempModel2.rewriteAlphaMatrix3D(); // write the updated rate matrix
	tempModel2.ProteinModelParameterstoRates(); // write the updated rate matrix scaled by freqs
	
    // If the above proposal exceeds mu, retry (this is quite wasteful)
	if (options.capUniformizationRate) {
        startMu =  mModel.getUniformization(whichModel);
        newMu = tempModel2.returnOptimalUniformization(whichModel);
		
		int rejectCount = 0;
		while (newMu > startMu) {
			for (int ii=0; ii<theRates2->size(); ii++) {
				theRates2->at(ii) = theRates1->at(ii);
			}
			
			tempModel2.UpdateRateDistributionRAR_orderedDiscrete_MC(whichModel, 1.0); //0.001); // last was 0.005 			
			tempModel2.rewriteAlphaMatrix3D(); // write the updated rate matrix
			tempModel2.ProteinModelParameterstoRates(); // write the updated rate matrix scaled by freqs
			
			newMu = tempModel2.returnOptimalUniformization(whichModel);
			rejectCount++;
			if (rejectCount > 10) { std::cout << "*" << endl; return 0; } // std::cout << "*" << endl;
		}		
	}
	
	if (whichModel != -1) {
		numUpdated_local++;
	} else {
		numUpdated_local += mModel.GetNumModels();
	}
	
	
	// Don't do this all the time!
	if (num->Random() <= 0.15) {
		if (tempModel2.getDependentRates() && (mModel.GetNumModels() > 1) ) {
			tempUpdate += Gibbs_updateRateParameterAssignments_RAR_DRA_B1u(tempModel2, 1.0, -1);
		} else {
			for (int i=0; i<mModel.GetNumModels(); i++) {
				tempUpdate += Gibbs_updateRateParameterAssignments_RAR_B1u( tempModel2, 1.0, i );
			}
		}
		
		// Rebuild rate matricies based on parameter assignments; reset uniformization rate...
		tempModel2.rewriteAlphaMatrix3D();
		tempModel2.ProteinModelParameterstoRates();
	}
	
	int uniDirtyFlag = 0;
	if (!options.capUniformizationRate) {
		// Get the larger uniformization constant over tempModel1 and tempModel2...
		tempModel2.forceUpdateUniformization(whichModel);
		
		double theMax = 0.0;
		DoubleVector *mu1 = tempModel1.returnUniformizationConstants();
		DoubleVector *mu2 = tempModel2.returnUniformizationConstants();
		if ( (*mu1)[whichModel] > theMax) { theMax = (*mu1)[whichModel]; }
		if ( (*mu2)[whichModel] > theMax) { theMax = (*mu2)[whichModel]; }
		
		uniDirtyFlag = 0;
		if ( theMax > (*mu1)[whichModel] ) {
			// Set both models' uniformization constant to be the same...
			tempModel2.setUniformization(whichModel, theMax * 1.01);
			tempModel1.setUniformization(whichModel, theMax * 1.01);
			uniDirtyFlag = 1;
		}
	}
	
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	mTree.root()->setCalculateFullProbabilityDistribution(0);
	
	// Model 1 likelihood... (only the parts that depend on RATE parameters)
	mTree.setprobmatrixalongbranch(tempModel1, tempModel1.GetFreqModels());
	if (uniDirtyFlag == 1) tree_countChanges( tempModel1 );	
	
	logLikelihood1a = CalculateCompleteDataLogLikelihood(tempModel1, RATE);
	
	// Model 2 likelihood... (only the parts that depend on RATE parameters)
	mTree.setprobmatrixalongbranch(tempModel2, tempModel2.GetFreqModels());
	if (uniDirtyFlag == 1) tree_countChanges( tempModel2 );	
	logLikelihood2 = CalculateCompleteDataLogLikelihood(tempModel2, RATE);
	
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
	
	// Got the likelihood ratio (assuming the larger of the two uniformization constants)...
	logLR = (logLikelihood2 - logLikelihood1a);
	
	bool m_bUpdate = false;
	if (logLR >= 0.0 )
	{
		mModel = tempModel2;
		numUpdated_local += tempUpdate;
		m_bUpdate = true;
	} else {
		
		if (num->evaluateProposal_logL( logLR ) ) {
			mModel = tempModel2;
			numUpdated_local += tempUpdate;
			m_bUpdate = true;
		} else {
			m_bUpdate = false;
			numUpdated_local = 0;
		}
	}
	
	if (m_bUpdate == true) {
		if (!options.capUniformizationRate) mModel.forceUpdateUniformization(whichModel);
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
	} else {
		// Need to reset the counts
		// [test] 	if (uniDirtyFlag == 1) tree_countChanges( mModel );	
	}
	
	return numUpdated_local;
}


int MCMC::MC_updateStateFreqs_B1u(
                                  
                                  ProteinModel& mModel
                                  
                                  ) {
    
	// Simple Metropolis update of the freq parameters
    // Uses the B1u likelihood function
    
    if (options.m_nMethod != 6) { cerr << "Running PLEX in legacy mode without B1u? Unsupported." << endl; exit(-1); }
    
    const int             RATE    = 1;
    static DoubleMatrix freqMatrix;
    
    if (freqMatrix.empty()) {
        freqMatrix.resize( mModel.GetNumModels(), mModel.GetNumStates(), 0.0 );
    }
    
	double  logLR;
	int     numModels = mModel.GetRateMatrix3D().rows();
	bool m_bUpdate;
    
	for (int whichModel = 0; whichModel < numModels; whichModel++) {	
		double logLikelihood1a, logLikelihood2;
		int numUpdated_local = 0;
		int tempUpdate = 0;
		
        // Get starting likelihood
        int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(0);
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		logLikelihood1a = CalculateCompleteDataLogLikelihood(mModel, RATE);
        
        // Backup rate and alpha matrix
        mModel.backupMatrix( mModel.GetNumModels(), mModel.GetNumStates(), mModel.returnStateFrequencies(), freqMatrix );
        
        mModel.UpdateFrequencyMatrix();
        mModel.ProteinModelParameterstoRates();
        
        // Model 2 likelihood... (only the parts that depend on RATE parameters)
		mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		logLikelihood2 = CalculateCompleteDataLogLikelihood(mModel, RATE);
        
		// Clean-up
        mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		
        // Get the logLR and evaluate
		logLR = (logLikelihood2 - logLikelihood1a);		
        
        m_bUpdate = false;
		if (logLR >= 0.0 ) {
            // Accept
			numUpdated_local += tempUpdate;
			m_bUpdate = true;
            
		} else {
            // Check
			if (num->evaluateProposal_logL( logLR ) ) {
                // Accept
                numUpdated_local += tempUpdate;
				m_bUpdate = true;
			} else {
				// Reject
                m_bUpdate = false;
				numUpdated_local = 0;
                
                // Restore freq and alpha matrix
                mModel.backupMatrix( mModel.GetNumModels(), mModel.GetNumStates(), freqMatrix, mModel.returnStateFrequencies() );
                mModel.ProteinModelParameterstoRates();
            }
		}
	}	
	
	if (m_bUpdate) return 1;
	else return 0;
}

// MARK: Samplers: Metropolis (multi-dim) all-parm updaters
void MCMC::proposal_rateMatrixAndFreqParameters(
                                                MixtureModel& m_Model,
                                                const double& m_fWindowsize) {
	
	m_Model.SetUpdateWindowsize(m_fWindowsize );
	m_Model.UpdateParameter(options.m_nMethod);
}

int MCMC::MC_updateRateMatrix(
                              ProteinModel& mModel
                              ) {
    
	ProteinModel tempModel1(mModel);
	ProteinModel tempModel2(mModel);
	double logLR;
	
	double windowsize = mModel.GetUpdateWindowsize();
	
	proposal_rateMatrixAndFreqParameters(tempModel2, windowsize);
	
	logLR = CalculateCompleteDataLogLR_overTree(tempModel1, tempModel2);
    
    bool m_bUpdate = false;
    
    if ( num->evaluateProposal_logL( 1.0, (1.0 - logLR), 1.0, 1.0 ) == 1 ) {
		mModel = tempModel2;
		m_bUpdate = true;
	} else {
		m_bUpdate = false;
	}
	
	if (m_bUpdate) return 1;
	else return 0;
}


void MCMC::MC_updateRateMatrix(
                               
                               GTRModel& mModel
                               
                               )
{
	// For uniformized B1, we'd like three versions: BRANCH, RATE, ALL [these will return either the full (ALL) likleihood, or just the parts that depend on rates or branches]
    //	const int BOTH    = 0;
	const int RATE    = 1;
    //	const int BRANCH  = 2;
	
	static int firstCall = 1;
	static GTRModel tempModel1(mModel);
	static GTRModel tempModel2(mModel);
	if (firstCall == 1) {
		firstCall = 0;
	} else {
		tempModel1 = mModel;
		tempModel2 = mModel;
		firstCall = 0;
	}
	
	double logLikelihood1a, logLikelihood2;
	double logLR;
	
	int numModels = mModel.GetRateMatrix3D().rows();
	
	for (int whichModel = 0; whichModel < numModels; whichModel++) {
		double startMu; //, newMu;
		
		Vector<double> *theFreqs1, *theFreqs2;
		theFreqs1 = &tempModel1.returnStateFrequencies()[whichModel];
		theFreqs2 = &tempModel2.returnStateFrequencies()[whichModel];
		double windowsize = mModel.GetUpdateWindowsize();
		
		if (options.capUniformizationRate) startMu =  mModel.getUniformization(whichModel);
		
		// THIS IS INEFFICIENT: SHOULD CHECK IF THE PROPOSAL EXCEEDS MU IN THE
		// proposal_rateMatrixAndFreqParameters() FXN AND REFLECT BACK IF CAPUNI IS ON...
		
		// Multidimensional update...
		proposal_rateMatrixAndFreqParameters(tempModel2, windowsize);
        
		// ***************************************************************************************************
		// This version for enforcing a capped uniformization constant...
		// ***************************************************************************************************
		/*		if (options.capUniformizationRate) {
		 tempModel2.forceUpdateUniformization(whichModel);
		 newMu = tempModel2.getUniformization(whichModel); /// / 1.04;
		 
		 int rejectCount = 0;
		 while (newMu > startMu) {
		 tempModel2.setUniformization(whichModel, startMu);
		 for (int ii=0; ii<theFreqs2->size(); ii++) {
		 theFreqs2->mDataArray[ii] = theFreqs1->mDataArray[ii];
		 }
		 
		 proposal_rateMatrixAndFreqParameters(tempModel2, windowsize, options.m_nMethod);
		 tempModel2.forceUpdateUniformization(whichModel);
		 newMu = tempModel2.getUniformization(whichModel); /// / 1.04;
		 rejectCount++;
		 //		std::cout << "	startmu was " << startMu << "; newmu is " << newMu << endl;
		 if (rejectCount > 10) { 
		 std::cout << "*" << endl; 
		 if (windowReduce < 5) { 
		 mModel.SetUpdateWindowsize( (windowsize * 0.9) ); 
		 std::cout << "	[reduced proposal window size to " << mModel.GetUpdateWindowsize() << "]" << endl; 
		 windowReduce++;
		 continue;
		 }
		 return;
		 }
		 }
		 //	std::cout << " *** OKAY *** "<< endl;
		 
		 tempModel2.setUniformization(whichModel, (startMu) );
		 } */
		// ***************************************************************************************************
		
		int uniDirtyFlag = 0;
		// ***************************************************************************************************
		// This version has to check if the implied uniformization constant from the proposal needs to be
		// used for evaluating the delta log-L
		// ***************************************************************************************************
		if (!options.capUniformizationRate) {
			tempModel1.updateUniformization(whichModel);
			tempModel2.forceUpdateUniformization(whichModel);
			
			double theMax = 0.0;
			DoubleVector *mu1 = tempModel1.returnUniformizationConstants();
			DoubleVector *mu2 = tempModel2.returnUniformizationConstants();
			if ( (*mu1)[whichModel] > theMax) { theMax = (*mu1)[whichModel]; }
			if ((*mu2)[whichModel] > theMax) { theMax = (*mu2)[whichModel]; }
			
			uniDirtyFlag = 0;
			if ( theMax >= (*mu1)[whichModel] ) {
				// Set both models' uniformization constant to be the same...
				tempModel2.setUniformization(whichModel, theMax ); // * 1.01
				tempModel1.setUniformization(whichModel, theMax );
				uniDirtyFlag = 1;
			} else {
				tempModel2.setUniformization(whichModel, (*mu1)[whichModel] );
			}
		}	
		
		//		tempModel2.forceUpdateUniformization();
		
		// ***************************************************************************************************
		
		// ***************************************************************************************************
		// Get the likelihoods of the start and end
		// ***************************************************************************************************
		int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
		mTree.root()->setCalculateFullProbabilityDistribution(0);
		
		// NOTE: unless the uniformization constant was changed for evaluation, we shouldn't need to recalculate the likelihood here for mModel (tempModel1)
		
		// Model 1 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
		logLikelihood1a = CalculateCompleteDataLogLikelihood(tempModel1, RATE);
		
		// Model 2 likelihood... (only the parts that depend on RATE parameters - when m_nMethod allows)
		logLikelihood2 = CalculateCompleteDataLogLikelihood(tempModel2, RATE);
		
		mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
		
		// Got the log-likelihood ratio (assuming the larger of the two uniformization constants)...
		logLR = (logLikelihood2 - logLikelihood1a);
		
		//		std::cout << "LR is " << logLR << endl;
		// ***************************************************************************************************
		
		
		// ***************************************************************************************************
		// Accept / reject?
		// ***************************************************************************************************
		bool m_bUpdate = false;
		if (logLR >= 0.0 )
		{
			mModel = tempModel2;
			m_bUpdate = true;
		} else {
			if (num->evaluateProposal_logL( logLR ) ) {
				mModel = tempModel2;
				m_bUpdate = true;
			} else {
				m_bUpdate = false;
			}
		}
		// ***************************************************************************************************
		
		
		
		if (m_bUpdate == true) {
			if (!options.capUniformizationRate) mModel.forceUpdateUniformization(whichModel);
			if (options.m_nMethod == 6) mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
		} else {
			// Update the prob-matrix at the root...
			if (options.m_nMethod == 6) mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
            
		}
	}	
}

// ****************************************************************************************
// MARK: Samplers: wrappers
// ****************************************************************************************
void MCMC::MCMC_updateRateMatrix( void ) {
    
    // Set a prior distribution for rate parameters using slice sampler
    PriorDistribution* thePrior;
    static UniformPrior _uniformPrior;
    static BetaPrior    _betaPrior( 0.01, 5.0 );
    static ExpPrior    _expPrior( 10.0 );
    
    // Set the desired prior distribution here
    thePrior = &_expPrior;
    
    // Choose the type of updater to use
    if ( options.allowASRV || options.reversibleModel && ( updateCount > 200) ) {
        
        static int symStep = 0;
        
        // Symmetrize the rate matrix
        if ( options.reversibleModel ) {
            if (symStep == 0) {
                cout << "       [ Symmetrizing the rate matrix ]" << endl;
                for (int i=0; i < mModel.GetNumModels(); i++) {
                    mModel.symmetrize(i);
                }
                mModel.ProteinModelParameterstoRates();
                
                // Force an update of the uniformization constants
                int temp = options.capUniformizationRate;
                options.capUniformizationRate = 0;
                mModel.forceUpdateUniformization();
                options.capUniformizationRate = temp;
                
                 symStep = 1;
            }
        }
        
        // Slice sampler (general non-reversible)
        if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Resampling rate matrix by slice sampler / slice_updateRateMatrix_genNonRev_B1u() ]" << endl;
        numUpdated += slice_updateRateMatrix_genNonRev_B1u( *thePrior );
    } else {
     
        // Conjugate Gibbs with ASRV not finished
        /*
        if (options.allowASRV) {
            // Set branchlength scaling
            options.__ASRVmode = 0;
            tree_countChanges( mModel, 0 );
        } */
        
        // Conjugate Gibbs (general non-reversible)
        if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Resampling rate matrix (general non-rev) model by conjugate Gibbs / cGibbs_updateRateMatrix_genNonRev_B1u() ]" << endl;
        numUpdated += cGibbs_updateRateMatrix_genNonRev_B1u(mModel);

        // Conjugate Gibbs with ASRV not finished
        /*
        if (options.allowASRV) {
            // Set branchlength scaling
            options.__ASRVmode = 1;
            tree_countChanges( mModel, 0 );
        }
         */
    }    
    // Metropolis update
    // if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Resampling full non-rev model by Metropolis / MC_updateRateMatrix_nonRAR() 100% ]" << endl;
    // numUpdated += MC_updateRateMatrix_nonRAR( mModel);

    // If state frequencies are free, flip a coin and update
    if ( options.freeFreqs ) {
        if ( num->Random() <= 0.05 ) {
            if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Updating state frequencies by Metropolis step / MC_updateStateFreqs_B1u() ]" << endl;

            numUpdated += MC_updateStateFreqs_B1u(mModel ); 
        }
    }
}


void MCMC::MCMC_updateSiteAssignments( void ) {

    if ( options.allowASRV && (options.numCatASRV > 1) ) {
        if (options.m_nMethod != 6) {
            setAllProbMatricesOnTree( mModel );
        }			
        
        // Resample the virtual substitutions - [TESTING Nov 9]
        // if (num->Random() <= options.freqAuxVirtualSubResampling) tree_countChanges( mModel, 1 );
        
        if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [ Resampling site rate assignments / Gibbs_updateSiteAssignments_RateAssignments() ]" << endl; 
        Gibbs_updateSiteAssignments_RateAssignments(mModel);
    }
    
    if ( ( mModel.GetNumModels() > 1) ) { 
        if ( (updateCount % options.outputFrequency) == 0) std::cout << "		[ Resampling site assignments for mixture model / Gibbs_updateSiteAssignments_MixtureComponents() ]" << endl; 
        numUpdated += Gibbs_updateSiteAssignments_MixtureComponents(mModel);	
    }
}

void MCMC::MCMC_updateRARmodel(void) {
    int numModels = mModel.GetNumModels();
    int whichModel;
    
    if ((updateCount==1) || ( (updateCount%30)==0)) {
        // Do a RAR parameter assignment sweep...
        if (mModel.getDependentRates() && (mModel.GetNumModels() > 1)) {	
            if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount++ << " [Resampling rate parameter values with re-assignment sweep / RunMCMCFullParameterSweep_dependentAssignment 100%]" << endl;
            numUpdated += Gibbs_updateRateParameterAssignments_RAR_DRA_B1u(mModel, 1.0, -1);
        } else {
            whichModel = (num->Random() * numModels);
            
            if ( (updateCount % options.outputFrequency) == 0)  std::cout << "		" << updateCount++ << " [Resampling rate parameter assignments / (a) RunMCMCFullParameterSweep 100%]" << endl;
            numUpdated += Gibbs_updateRateParameterAssignments_RAR_B1u(mModel, 1.0, -1);
        }
    } else {
        if (mModel.getDependentRates()) {	
            whichModel = -1;
        } else {
            whichModel = int(num->Random() * numModels); //[jdk] Numerics::getGlobalNumerics().Random();
        }	
        
        if ( num->Random() <= 0.3) { // 0.3 ) { // was last 0.3 // BLOCKED July 2009
            // RAR parameter sweep...
            if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [Resampling rate parameter assignments / (b) RunMCMCFullParameterSweep / FULL]" << endl;
            
            if (mModel.getDependentRates()) {	
                numUpdated += Gibbs_updateRateParameterAssignments_RAR_DRA_B1u(mModel, 0.99, whichModel);
            } else {
                numUpdated += Gibbs_updateRateParameterAssignments_RAR_B1u(mModel, 1.0, whichModel);
            }
            
        } else {
            if ( (num->Random() <= 0.5) || (mModel.getFixedFreqs() == 1) ) {
                // Update the RAR parameter distribution...
                if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " [Resampling rate parameter values with re-assignment sweep / RunMCMCDistributionParametersWithSweep <= 100%]" << endl;
                numUpdated += MC_updateRARDistributionParameters_B1u(mModel, 1.0, whichModel);
            } else {
                if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount++ << " Resampling other model parameters (non-RAR) / MC_updateRateMatrix_nonRAR()" << endl;
                
                cout << "MC_ratematrixUpdate 5" << endl;
                numUpdated += MC_updateRateMatrix_nonRAR( mModel );
            }
        }
    }
}

// ****************************************************************************************
// MARK: Tree method wrappers
// ****************************************************************************************
void MCMC::tree_countChanges( MixtureModel& mModel ) {
	mTree.CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment );
}

void MCMC::tree_countChanges( MixtureModel& mModel, int version) {
	mTree.CountFullChangesOverTree( mModel, theData, m_vecSiteAssignment, version );
}

/*
 // deprecated
void MCMC::tree_updateBranchLengths(
                                    
                                    int startNode,
                                    GTRModel& mModel, 
                                    
                                    int version, double variance) {
	
    // [Feb 1 block]
    // if (updateCount > 5000) return;
    
    if ( version == 0 ) {
        setAllProbMatricesOnTree( mModel );
    }

    startNode=0; // force update from root
	
    // Recount the virtual and branch-specific subs...
    tree_countChanges( mModel, 1 );

    // Versions: 0 = standard branch-wise deltaL, 1 = conjugate gibbs (b1u), 2 = simple proposal evaluation (b1u)
	int dirty = mTree.KEEP_updateBranchLength( startNode, variance, mModel, theData, m_vecSiteAssignment, version );
    
    // Recount the virtual subs...
    tree_countChanges( mModel, 1 );
}
 */

void MCMC::tree_updateBranchLengths(
                                    
                                    int startNode,
                                    MixtureModel& mModel, 
                                    
                                    int version, double variance) {
	
    // [Feb 1 block]
    // if (updateCount > 5000) return;

    if (options.updatebls != 1) {
        cout << "       **** ERROR: updateBLs called while branchlengths fixed! *******" << endl;
        exit(-1);
    }
    
    if ( version == 0 ) {
        setAllProbMatricesOnTree( mModel );
    }

    // This was forcing start at the root...
    // startNode=0; // force update from root
    
    // Recount the virtual and branch-specific subs...
    // [ let calling function do this! ] tree_countChanges( mModel, 1 );

    double maxBL = options.maxBL / ( mModel.getRateSum(0, mTree) - options.maxBL * (*mModel.returnUniformizationConstants())[0] );
    
    if ( options.__currentGen > 1000 ) {
        // Don't subdivide branches until the rate matrix and mu are equilibrated a little bit
        // Bad starting points can yield a non-sensical maximum BL (e.g., < 0)
        
        // cout << "       Subdividing using " << maxBL << "; rateSum is " << mModel.getRateSum(0, mTree) << ", mu is " << (*mModel.returnUniformizationConstants())[0] << endl;
        
        if (maxBL > 0) {
            // Set this so branchlength updater can stay within bounds
            options.maxExpectedSubs = maxBL;
        }
    }
    
    // versions: 0 = standard branch-wise deltaL, 1 = conjugate gibbs (b1u), 2 = simple proposal evaluation (b1u)
    int dirty = mTree.KEEP_updateBranchLength( startNode, variance, mModel, theData, m_vecSiteAssignment, version );
	  
    // [diagnostic] cout << "       **** Updated tree depth is: " << mTree.root()->GetData().treeLength << endl;

    if ( ( options.__currentGen > 1000 ) && !options.capNumTreeNodes && (dirty > 0) ) {
        mTree.subdivideLongBranches( mTree.root(), options.maxExpectedSubs, mModel.returnUniformizationConstants(), 1 );
    }
    
    if (!options.capNumTreeNodes) {
        mTree.root()->NormalizeBranchSegmentLengths( theData, mTree.root() );
    }
    
    // Recount the virtual subs...
    // [ let calling function do this! ] tree_countChanges( mModel, 1 );
    
}


// ****************************************************************************************
// MARK: Samplers: mixed methods
// ****************************************************************************************
int MCMC::UpdateGTRrates_unidimensional_b1u(
											
											GTRModel& mModel, 
											
											double updateFraction, int whichModel) {
    
	// ********************************************************************************************************************************
	// This version was used for the B1u paper; de Koning et al (2009) MBE
	// ********************************************************************************************************************************
	// NOTE!  THIS ONLY WORKS WHEN USING UNNORMALIZED RATE MATRICES
	
    if ( (options.m_nMethod != 6)) { std::cout << "ERROR! ParameterSweep only implemented for B1-type UNIFORMIZED methods at present" << endl; exit(-1); }
	DoubleVector *mu = mModel.returnUniformizationConstants();
	
	double logger1, logger2, deltaL;
	DoubleMatrix3D    *realChangeCount3D = mTree.root()->getChangeMatrix3D( mTree.root() ); //&GetData().countModelchange3D;
    //	DoubleMatrix3D *ChangeCount3D     = mTree.root()->getChangeMatrix3Duni( mTree.root() ); //&GetData().countModelchange3D;
	DoubleMatrix    *augmentedCount      = mTree.root()->getAugmentedCountMatrix( mTree.root() );
	
	double oldRate, oldRate_1, proposal, proposal_1, oldRate_2, proposal_2, changeamount, var;
	//int numUpdated = 0;
	
	// This function is for GTR models only...
	int numStates  = mModel.GetRateMatrix3D().cols();
	int numModels  = mModel.GetRateMatrix3D().rows();
	DoubleMatrix* stateF = &mModel.returnStateFrequencies();
	
	// Pick a single rate to update...
	int whichRateFrom;
	int whichRateTo;
	
	updateFraction = 1.5;
	
	int updateLimit = (updateFraction * (numStates * (numStates-1) ) );
	int rejectFlag = 0;
	int rejectCount = 0;
	
	for (int numUpdates = 0; numUpdates < updateLimit; numUpdates++) {
		
		// Put a limit on number of attempts we make to sample when hitting the mu boundary under capuniformization...
		if (rejectCount > 10) {
			rejectFlag = 0; rejectCount = 0;
			updateLimit++;
			continue;
		}
		
		if (rejectFlag == 0) {
			// Don't resample the states if we just rejected a move due to capuniformization...
			whichRateFrom  = (int)(num->Random() * (numStates));
			whichRateTo = (int)(num->Random() * (numStates));
			
			while (whichRateTo == whichRateFrom) {
				whichRateTo = (int)(num->Random() * (numStates));
			}
		}
		
		rejectFlag = 0;
		// Got a rate to do...
		int modelStart, modelEnd;
		if (whichModel == -1) {
			modelStart = 0; modelEnd = numModels;
		} else {
			modelStart = whichModel; modelEnd = modelStart+1;
		}
		
		for (int theModel=modelStart; theModel < modelEnd; theModel++) {
			double startMu = mu->mDataArray[theModel];
			oldRate = mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] / stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo]; //mModel.
			
			if ( Numerics::getGlobalNumerics().Random() < 0.05 ) {
				changeamount = ( Numerics::getGlobalNumerics().Random() - 0.5) * 0.001 * oldRate; //trying to scale change amount by parm size to equalize mixing...
				proposal = oldRate + changeamount;
				
				if (proposal <= 0.0 )
					proposal = oldRate - changeamount;
			} else {
				var = 0.15;
				
				if (oldRate <= 0.68) {
					var = 0.028;
				} else if (oldRate <= 0.35) { 
					var = 0.009;
				} else if (oldRate <= 0.11) {
					var = 0.012; 
				} else if (oldRate <= 0.06) {
					var = 0.003;  
				}
				var *= 0.5;
				
				changeamount = num->gaussianVariate( 0.0, var ); // (rannum - 0.5) * mWindowsize;
				proposal = oldRate + changeamount;
				
				while ( proposal <= 0.0 ) {
					changeamount *= -1;
					proposal = oldRate + changeamount;
				}
			}
			proposal_1 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			proposal_2 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			oldRate_1  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			oldRate_2  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			
			// ***************************************************************************************************
			// Need to see if mu requires updating...
			// ***************************************************************************************************
			double newLambda;
			// Row 1...
			newLambda= -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] - oldRate_1 + proposal_1;
			newLambda /= 1.04;
			if (newLambda > mu->mDataArray[theModel]) {
				if (!options.capUniformizationRate) {
					mu->mDataArray[theModel] = newLambda;
				} else {
					// proposal causes newLambda to exceed our fixed mu...
					rejectFlag = 1; rejectCount++; updateLimit++;
					std::cout << "* ";
					continue;
				}
			}
			
			// Row 2...
			newLambda = -mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo] - oldRate_2 + proposal_2;
			newLambda /= 1.04;
			if (newLambda > mu->mDataArray[theModel]) {
				if (!options.capUniformizationRate) {
					mu->mDataArray[theModel] = newLambda;
				} else {
					// proposal causes newLambda to exceed our fixed mu...
					rejectFlag = 1; rejectCount++; updateLimit++;
					std::cout << "* ";
					continue;
				}
			}
			
			// Now do change in likelihood...
			deltaL = (*realChangeCount3D).mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] * num->logarithm( (oldRate_1 / proposal_1) );
			deltaL += (*realChangeCount3D).mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] * num->logarithm( (oldRate_2 / proposal_2) );
			
			// Now do the virtual substitutions (may be different across rates classes)...
			int rateClassLimit = options.numCatASRV;
			if ( options.allowASRV != 1 ) rateClassLimit = 1;
			for (int rateclass= 0; rateclass < rateClassLimit; rateclass++ ) {
				logger1 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] * mTree.root()->GetData().siteCats->getRate( rateclass ) );
				logger2 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] + oldRate_1 - proposal_1)* mTree.root()->GetData().siteCats->getRate( rateclass );
				if (logger1 <= 0) logger1  = 1e-70;
				if (logger2 <= 0) logger2  = 1e-70;
				
				deltaL += (*augmentedCount).mRows.mDataArray[ num->returnModelIndex( theModel, rateclass, 0) ].mDataArray[whichRateFrom] * ( num->logarithm( ( logger1 )) - num->logarithm( logger2 ) );
			}
			
			// Now do the virtual substitutions (may be different across rates classes)...
			rateClassLimit = options.numCatASRV;
			if ( options.allowASRV != 1 ) rateClassLimit = 1;
			for (int rateclass= 0; rateclass < rateClassLimit; rateclass++ ) {
				logger1 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo] * mTree.root()->GetData().siteCats->getRate( rateclass) );
				logger2 = mu->mDataArray[theModel] + (mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo] + oldRate_2 - proposal_2) * mTree.root()->GetData().siteCats->getRate( rateclass);
				if (logger1 <= 0) logger1 = 1e-70;
				if (logger2 <= 0) logger2 = 1e-70;
				
				deltaL += (*augmentedCount).mRows.mDataArray[ num->returnModelIndex( theModel, rateclass, 0) ].mDataArray[whichRateTo] * ( num->logarithm( ( logger1 )) - num->logarithm( logger2 ) );
			}
            
			deltaL *= -1.0;
			
			if ( num->evaluateProposal_logL( 1.0, (1.0 - deltaL) ) == 1 ) {
				// ACCEPT
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal;
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal;
				
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal_2;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate_1 - proposal_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo]     += oldRate_2 - proposal_2;
				
				// To be consistent with the forceUpdateUniformization() in the multidimensional proposals...
				if ( mu->mDataArray[theModel] != startMu ) mu->mDataArray[theModel] *= 1.01;
				
			} else {
				mu->mDataArray[theModel] = startMu;
			}
		}
	}
	return 1;
}

// GTR version:
int MCMC::UpdateGTRrates_unidimensional(
										
										GTRModel& mModel, 
										
										double updateFraction, int whichModel) {
    
	// ********************************************************************************************************************************
	// This version was used for the B1u paper; de Koning et al (2009) MBE
	// ********************************************************************************************************************************
    
    //	double likelihood1, likelihood2;
	
	double oldRate, oldRate_1, proposal, proposal_1, oldRate_2, proposal_2, changeamount, var;
	
	// This function is for GTR models only...
	
	//	double logLikelihood1, logLikelihood2;
	int numStates  = mModel.GetRateMatrix3D().cols();
	int numModels  = mModel.GetRateMatrix3D().rows();
	DoubleMatrix* stateF = &mModel.returnStateFrequencies();
	
	// Pick a single rate to update...
	int whichRateFrom;
	int whichRateTo;
    
	updateFraction = 1.5; 
	
	int updateLimit = (updateFraction * (numStates * (numStates-1) ) );
    //	int rejectFlag = 0;
    //	int rejectCount = 0;
	
	for (int numUpdates = 0; numUpdates < updateLimit; numUpdates++) {
		whichRateFrom  = (int)(num->Random() * (numStates));
		whichRateTo = (int)(num->Random() * (numStates));
		
		while (whichRateTo == whichRateFrom) {
			whichRateTo = (int)(num->Random() * (numStates));
		}
		
		// Got a rate to do...
		int modelStart, modelEnd;
		if (whichModel == -1) {
			modelStart = 0; modelEnd = numModels;
		} else {
			modelStart = whichModel; modelEnd = modelStart+1;
		}
		
		for (int theModel=modelStart; theModel < modelEnd; theModel++) {
			//likelihood1 = CalculateCompleteDataLogLikelihood_Delta_B1(mModel,m_vecSiteAssignment, theModel, whichRateFrom);
			double full1 = CalculateCompleteDataLogLikelihood(mModel, theModel);
			
			oldRate = mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] / stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo]; //mModel.
			
			if ( Numerics::getGlobalNumerics().Random() < 0.05 ) {
				changeamount = ( Numerics::getGlobalNumerics().Random() - 0.5) * 0.001 * oldRate; //trying to scale change amount by parm size to equalize mixing...
				proposal = oldRate + changeamount;
				
				if (proposal <= 0.0 )
					proposal = oldRate - changeamount;
			} else {
				var = 0.15;
				
				// Approximate posterior variances determined from output
				// (for analysis of 224 taxon cyt-b data set in de Koning et al, 2009)
				if (oldRate <= 0.68) { 
					var = 0.028;  
				} else if (oldRate <= 0.35) { 
					var = 0.009;
				} else if (oldRate <= 0.11) { 
					var = 0.012; 
				} else if (oldRate <= 0.06) {
					var = 0.003; 
				}
				var *= 0.5;
				
				changeamount = num->gaussianVariate( 0.0, var ); // (rannum - 0.5) * mWindowsize;
				proposal = oldRate + changeamount;
				
				while ( proposal <= 0.0 ) {
					changeamount *= -1;
					proposal = oldRate + changeamount;
				}
			}
			proposal_1 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			proposal_2 = proposal * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			oldRate_1  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateTo];
			oldRate_2  = oldRate * stateF->mRows.mDataArray[theModel].mDataArray[whichRateFrom];
			
			// Set all the parms for likelihood evaluation:
			mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal;
			mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal;
			
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = proposal_1;
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = proposal_2;
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += oldRate_1 - proposal_1;
			mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo]     += oldRate_2 - proposal_2;
			
			// **** Standard B1 likelihood evaluation *****
			double full2 = CalculateCompleteDataLogLikelihood(mModel, theModel);
			
			if ( num->evaluateProposal_logL( full2, full1 ) == 1 ) {
				// ACCEPT
			} else {
				// REJECT
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = oldRate;
				mModel.mAlphaMatrix3D.mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = oldRate;
				
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateTo] = oldRate_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateFrom] = oldRate_2;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateFrom].mDataArray[whichRateFrom] += -oldRate_1 + proposal_1;
				mModel.GetRateMatrix3D().mRows.mDataArray[theModel].mRows.mDataArray[whichRateTo].mDataArray[whichRateTo]     += -oldRate_2 + proposal_2;
			}
		}
	}
	return 1;
}

// ****************************************************************************************
// MARK: Tree: operations
// ****************************************************************************************
void MCMC::setAllProbMatricesOnTree( MixtureModel& mModel ) {
	// Force a recalculation of all branches' probability matrices using whatever method is set
	// (Note: this will force their calculation even for B1u, which usually will not require it)
	
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	
	mTree.root()->setCalculateFullProbabilityDistribution(1);
	mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
}

void MCMC::setAllProbMatricesOnTree( MixtureModel& mModel, int version ) {
	// Force a recalculation of all branches' probability matrices using whatever method is set
	// (Note: this will force their calculation even for B1u, which usually will not require it)
	
	int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
	
    // Do this for model 0
    options.__currentRateSum = mModel.getRateSum(0, mTree);
    
	mTree.root()->setCalculateFullProbabilityDistribution(version);
	mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
	mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
}

// ************************************************************************//
// MARK: Coevolution functions
// ************************************************************************//

void MCMC::coevo_DataReplacementSimulation(void) {
    static int runYet = 0;
    
    // Execute at generation 10,000 (if requested)
    if ( !(  (options.coevol_DataReplacementSimulation) && (updateCount > 10000) && (runYet == 0) ) ) { 
        return;
    }

    // Resize site-specific vectors etc. if requested to do so...
    if ( (runYet == 0) && options.coevol_DataReplacement_differentNumSites ) {
        theData.resetAlignmentLength( options.coevol_DataReplacement_newNumSites );
        m_vecSiteAssignment.resize( options.coevol_DataReplacement_newNumSites );
        mTree.root()->GetData().siteCats->siteAssignments.resize( options.coevol_DataReplacement_newNumSites );
        mTree.root()->GetData().siteCats->backupAssignments.resize( options.coevol_DataReplacement_newNumSites );
        
        mTree.root()->GetData().countModelchange_perSite.resize( options.coevol_DataReplacement_newNumSites );
        mTree.root()->GetData().timeSpentSame_perSite.resize( options.coevol_DataReplacement_newNumSites, mModel.GetNumStates(), 0.0);
    }

//[block]    
	runYet = 1;
    
    options.__overrideSiteResamp = 1;
    
    // *******************************************************************************************************
    // RESCALE EVERYTHING FIRST SO THAT THE RATE MATRIX SCALING IS CONSISTENT
    // *******************************************************************************************************
    
    int last = options.capUniformizationRate;
    options.capUniformizationRate = 0;
    mModel.forceUpdateUniformization();
    
    double theRateFactor = mModel.getRateSum( mTree ) / mModel.GetNumModels();
    mModel.scaleDiscreteRates( theRateFactor ) ;
    double mu = mModel.returnUniformizationConstants()->mDataArray[0] / theRateFactor;
    mModel.setUniformization(0, mu ); //(mu / theRateFactor) );
    mModel.scaleRateMatrices( theRateFactor );
    
    // [diagnostic output] cout << "           UPDATED MU to " << (*mModel.returnUniformizationConstants())[0] << endl;
    options.capUniformizationRate = last;
    //mTree.root()->TransformBranchLength_B1u( mTree.root(), ( 1.0 / theRateFactor), mu );
    
    // _util_rescaleForBranchLengths();
    
    // Even out the segments, resample subs, recount
    // if (options.updatebls) mTree.root()->NormalizeBranchSegmentLengths(theData, mTree.root() );
    
    //mModel.forceUpdateStationaryFreqs();
    //numUpdated += Gibbs_updateAllAncestralAndTransientStates_full( mModel );
    tree_countChanges( mModel );

    // *******************************************************************************************************
        
    cout << "           [ ** REPLACING ALL DATA WITH POSTERIOR-PREDICTIVE SAMPLE; option coevolveDataReplacementSimulation ** ]" << endl;
    
    // Second, recalculate full substitution probability matrices across the tree (for all rate classes)
/*
    if (options.updatebls) {
        setAllProbMatricesOnTree( mModel );
    } else {
        setAllProbMatricesOnTree( mModel, 2 ); // experimental version using rescaled bl's
    }
*/
    // [no longer needed - doing full sampling ] setAllProbMatricesOnTree( mModel );
    
     int numActual = 0;
    // Third, resimulate the data all the way to the tips (use existing site rates, number of sites)
    for (int ss=0; ss < theData.getAlignmentLength(); ss++) {
        
        if (options.reversibleModel ==1 ) {
			numActual = mTree.root()->posteriorPredictive_replaceData( ss, mTree.root(), mModel, theData, m_vecSiteAssignment, mTree.GetModelFrequencyVector(), mModel.returnStateFrequencies() );
        } else {
            numActual = mTree.root()->posteriorPredictive_replaceData( ss, mTree.root(), mModel, theData, m_vecSiteAssignment, mTree.GetModelFrequencyVector(), mModel.returnStationaryFrequencies() );
        }
   
       //cout << "Site " << ss << ", rate " << mTree.root()->GetData().siteCats->getRate( mTree.root()->GetData().siteCats->getAssignment( ss ) ) << ", num actual subs: " << numActual << endl;
    }
    
    // Now recount subs
    tree_countChanges( mModel );
    
    // Now, output MI and pMI: pMI for this generation is based on the 'true' (simulated) model and sub. history
    forceCoevolution = 1;
    
    // Very important!  If did gapsAsMissing, must turn this off!
    options.gapsAsMissing = 0;

}

void MCMC::coevo_calculateStatistics(void) {
    // This needs a lot of cleaning up
    
    // Only run after generation 6000 (if requested) 
    if ( ! ( (options.coevol) && (updateCount >= 10001) ) ) { return; }
    
    // [Feb 1 block]  if ( (updateCount > 1) && ((updateCount % options.outputIncompleteLogL) == 0) && (options.outputIncompleteLogL > 0) || (forceCoevolution == 1) ) {
    if ( (updateCount > 1) || (forceCoevolution == 1) ) {

        int numModels = mModel.GetNumModels();
        
        cout << "               [ CALCULATING POSTERIOR STATISTICS OF SUBSTITUTION HISTORIES ]" << endl;
        static int firstCoevolRound =1;
        
        // If requested, restart output of coevo results
        if (forceCoevolution == 1) { 
            firstCoevolRound = 1;
        }
        
        if (numModels > 1) {
            std::cout << "	ERROR, NO SUPPORT FOR MIXTURES WITH SELETED OPTIONS YET." << endl;
            exit(-1);
        }
        
        static ofstream posteriorCoevolution;
        static int posteriorCoevolutionOpen = 0;
        if (posteriorCoevolutionOpen == 0) {
            posteriorCoevolution.open( "posteriorCoevolution", fstream::out);
            posteriorCoevolutionOpen = 1;
        }
        
        static ofstream miCoevolution;
        static int miCoevolutionOpen = 0;
        if (miCoevolutionOpen == 0) {
            miCoevolution.open( "miCoevolution", fstream::out);
            miCoevolutionOpen = 1;
        }
        
        static ofstream posteriorSubs;
        static int posteriorSubsOpen = 0;
        if (posteriorSubsOpen == 0) {
            posteriorSubs.open( "posteriorSubsPerSite", fstream::out);
            posteriorSubsOpen = 1;
        }
        
        static ofstream posteriorDwell;
        if (options.coevol_outputDwell) {
            static int posteriorDwellOpen = 0;
            if (posteriorDwellOpen == 0) {
                posteriorDwell.open( "posteriorDwell", fstream::out);
                posteriorDwellOpen = 1;
            }
        }
        
        static ofstream posteriorJointDwell;
        if ( options.coevol_outputJointDwell ) {
            static int posteriorJointDwellOpen = 0;
            if (posteriorJointDwellOpen == 0) {
                posteriorJointDwell.open( "posteriorJointDwell", fstream::out);
                posteriorJointDwellOpen = 1;
            }
        }
        
        posteriorSubs << "Gen " << updateCount << "\t";
        
        // Dwell-time based MI for coevolution:
        DoubleVector* treeDepth;
        
        static ofstream ppCoevolution;
        if (options.coevol_doPosteriorPredictiveNull) {
            static int ppCoevolutionOpen = 0;
            if (ppCoevolutionOpen == 0) {
                ppCoevolution.open( "ppCoevolution", fstream::out);
                ppCoevolutionOpen = 1;
            }
        }
        
        
        static ofstream ppCoevolutionStat;
        if (options.coevol_doPosteriorPredictiveNull) {
            static int ppCoevolutionStatOpen = 0;
            if (ppCoevolutionStatOpen == 0) {
                ppCoevolutionStat.open( "ppCoevolutionStat", fstream::out);
                ppCoevolutionStatOpen = 1;
            }
        }
        
        // Allocate for observed freqs and joint dwell times
        // Just output differences in dwell versus observed state times:
        if (mTree.root()->GetData().observedFreqs.empty()) {
            mTree.root()->GetData().observedFreqs.resize( m_vecSiteAssignment.size(), mModel.GetNumStates(), 0.0); 
        } else {
            //	if (ii==0) {
            if (firstCoevolRound) mTree.root()->GetData().observedFreqs.fill(0.0);
            //	}
        }
        
        // Allocate a vector to keep track of the number of subs per site
        static IntVector subsPerSite;
        
        if (subsPerSite.empty()) {
            subsPerSite.resize( m_vecSiteAssignment.size(), 0 );
        }
        subsPerSite.fill( 0 );
        
        // Precompute the marginal dwell times at every position...
        for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
            double siteRate1 = 1.0;

            if (options.allowASRV) {
                // set the site rates here
                siteRate1 = mTree.root()->GetData().siteCats->getRate( mTree.root()->GetData().siteCats->getAssignment( ii ) );
            }

            subsPerSite.mDataArray[ ii ] = mTree.calculateDwellTimeOverBranches( ii, siteRate1, mModel, theData, m_vecSiteAssignment );	
            
            // Normalize observed freqs
            for (int i=0; i < mModel.GetNumStates(); i++) { mTree.root()->GetData().observedFreqs.mRows.mDataArray[ii].mDataArray[i] /= theData.getNumSequences(); }
        }
        
        
        // Temporary output for testing
        static ofstream pHist, ppHist;
        
        if ( ! pHist.is_open() ) {
            pHist.open("pHist.txt", ios::out );
        }
        
        if ( ! ppHist.is_open() ) {
            ppHist.open("ppHist.txt", ios::out );
        }
        
        // Doing site ii versus jj...
        for (int ii=0; ii<m_vecSiteAssignment.size(); ii++) {
            posteriorCoevolution << updateCount << "\t" << ii << "\t";
            // if (firstCoevolRound) miCoevolution << updateCount << "\t" << ii << "\t";
            
             if (mTree.root()->GetData().observedJFreqs.empty()) {
                mTree.root()->GetData().observedJFreqs.resize( 1, mModel.GetNumStates()*mModel.GetNumStates(), 0.0); 
            } else {
                //	if (ii==0) {
                if (firstCoevolRound) mTree.root()->GetData().observedJFreqs.fill(0.0);
                //	}
            }
            
            double siteRate1 = 1.0;
            double siteRate2 = 1.0;
            if (options.allowASRV) {
                // set the site rates here
                siteRate1 = mTree.root()->GetData().siteCats->getRate( mTree.root()->GetData().siteCats->getAssignment( ii ) );
            }
            
            
            
            // *******************************************************************************************************
            // RESCALE EVERYTHING FIRST SO THAT THE RATE MATRIX SCALING IS CONSISTENT
            // *******************************************************************************************************
            
			// Added this in for testing...
//			if (options.coevol_doPosteriorPredictiveNull) {
//				setAllProbMatricesOnTree( mModel );
//			}
			/* mModel.forceUpdateStationaryFreqs();
             
             double theRateFactor = mModel.getRateSum() / mModel.GetNumModels();
             mModel.scaleDiscreteRates( theRateFactor ) ;
             double mu = mModel.returnUniformizationConstants()->mDataArray[0] / theRateFactor;
             
             // Change uniformization and ratescale for Qs
             mModel.setUniformization(0, mu ); //(mu / theRateFactor) );
             mModel.scaleRateMatrices( theRateFactor );
             mTree.root()->TransformBranchLength_B1u( mTree.root(), ( 1.0 / theRateFactor), mu );
             */
            // *******************************************************************************************************
            
            //}
            
            // Display the site's dwell time stuff (if comparing to frequencies)...
            /*
             DoubleMatrix* marginalDwell = mTree.getRoot()->getMarginalDwell(mTree.getRoot());
             mTree.calculateDwellTimeOverBranches(ii, siteRate1, mModel, theData, m_vecSiteAssignment);	
             DoubleVector* treeDepth = &mTree.getRoot()->GetData().tempTreeDepth;
             
             // Force stationary freqs update...
             mModel.updateStationaryFreqs(1);
             
             double tempSum = 0.0;
             for (int i=0; i<mModel.GetNumStates(); i++) {
             std::cout << "Site\t" << ii << "\tState\t" << i << "\t" << ( marginalDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] ) << "\t" << (mTree.root()->GetData().observedFreqs.mRows.mDataArray[0].mDataArray[i] / mTree.leaves()) << "\t" << ( mModel.returnStationaryFrequencies().mRows.mDataArray[0].mDataArray[i] ) << "\t" << endl;
             //		std::cout << "site " << ii << ", state " << i << ", margDwell: " << tempMarginal.mRows.mDataArray[0].mDataArray[i] << endl;
             tempSum += (marginalDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] );
             }
             //std::cout << "	sum over dwell times at site is " << tempSum << endl;
             */
            
            // comment out below if just getting the dwell versus freq times...			
            DoubleMatrix* marginalDwell = mTree.getRoot()->getMarginalDwell(mTree.getRoot());
            DoubleMatrix* jointDwell    = mTree.getRoot()->getJointDwell( mTree.getRoot() );
            
            // *******************************************************************************************************
            // Site ii -> 'tempMarginal'
            // *******************************************************************************************************
            double numSubs1 = subsPerSite.mDataArray[ ii ]; //mTree.calculateDwellTimeOverBranches( ii, siteRate1, mModel, theData, m_vecSiteAssignment );	
            
            posteriorSubs << numSubs1;
            if ( ii < (m_vecSiteAssignment.size() - 1) ) posteriorSubs << "\t";
            else posteriorSubs << endl;
            
            // BLOCK THIS HERE
            /*
            if (numSubs1 < options.coevol_minNumSubstitutions) {
                // Just output a row of 0.0's for the MI statistic and continue...
                for (int jj=0; jj < m_vecSiteAssignment.size(); jj++) {
                    if (jj >= ii) {
                        posteriorCoevolution << "0.0" << "\t";
                        if (firstCoevolRound) miCoevolution << "0.0" << "\t";
                    } else { 
                        posteriorCoevolution << "NA\t";
                        if (firstCoevolRound) miCoevolution << "NA\t";
                    }
                } posteriorCoevolution << endl;
                if (firstCoevolRound) miCoevolution << endl;
                continue;
            }
            */
            
            // Dwell time output header
            static int posteriorDwellHeader = 0;
            if ( (options.coevol_outputDwell) && (ii==0) && (posteriorDwellHeader == 0) ) {
                posteriorDwell << "Gen\tSite\t";
                
                // For dwell times
                for (int i=0; i < mModel.GetNumStates(); i++) {
                    posteriorDwell << theData.getState(i);
                    posteriorDwell << "\t";
                }
                
                // And observed freqs
                for (int i=0; i < mModel.GetNumStates(); i++) {
                    posteriorDwell << "f" << theData.getState(i);
                    if (i < (mModel.GetNumStates()-1) ) {
                        posteriorDwell << "\t";
                    } else {
                        posteriorDwell << endl;
                    }
                }
                
                posteriorDwellHeader = 1;
            }
            
            if (options.coevol_outputDwell) {
                posteriorDwell << updateCount << "\t" << ii << "\t";
            }
            
            // Dwell times to file
            for (int i=0; i<mModel.GetNumStates(); i++) {
                // tempMarginal is for the current site
                // [this is redundant] tempMarginal.mRows.mDataArray[0].mDataArray[i] = marginalDwell->mRows.mDataArray[ 0 ].mDataArray[ i ];
                
                if (options.coevol_outputDwell) {
                    posteriorDwell << marginalDwell->mRows.mDataArray[ ii ].mDataArray[ i ];
                    posteriorDwell << "\t";
                }
            }
            
            // Output observed freqs to file
            if (options.coevol_outputDwell) {
                for (int i=0; i<mModel.GetNumStates(); i++) {
                    posteriorDwell << mTree.root()->GetData().observedFreqs.mRows.mDataArray[ii].mDataArray[i];
                    if (i < mModel.GetNumStates()-1) {
                        posteriorDwell << "\t";
                    } else {
                        posteriorDwell << endl;
                    }
                }
            }
            
            if (options.coevol_doPosteriorPredictiveNull) {
                ppCoevolutionStat << updateCount << "\t" << ii << "\t";
                ppCoevolution << updateCount << "\t" << ii << "\t";
            }
            
            for (int jj=0; jj<m_vecSiteAssignment.size(); jj++) {
                
                if (jj > ii) {
                    // *******************************************************************************************************
                    // Site jj -> marginalDwell
                    // *******************************************************************************************************
                    // Joint dwell time output header
                    static int posteriorJDwellHeader = 0;
                    if ( (options.coevol_outputJointDwell) && (ii==0) && (jj==1) && (posteriorJDwellHeader == 0) ) {
                        posteriorJointDwell << "Gen\tSite1\tSite2\t";
                        
                        // For dwell times
                        for (int i=0; i < mModel.GetNumStates(); i++) {
                            for (int j=0; j < mModel.GetNumStates(); j++) {
                                
                                posteriorJointDwell << theData.getState(i) << "x" << theData.getState(j);
                                posteriorJointDwell << "\t";
                            }
                        }
                        
                        // And observed freqs
                        for (int i=0; i < mModel.GetNumStates(); i++) {
                            for (int j=0; j < mModel.GetNumStates(); j++) {
                                posteriorJointDwell << "f" << theData.getState(i) << "x" << theData.getState(j);
                                if (i*mModel.GetNumStates()+j < (mModel.GetNumStates()*mModel.GetNumStates()-1) ) {
                                    posteriorJointDwell << "\t";
                                }
                            }
                        }    
                        
                        posteriorJDwellHeader = 1;
                    }
                    
                    if (options.allowASRV) {
                        // set the site rates here
                        siteRate2 = mTree.root()->GetData().siteCats->getRate( mTree.root()->GetData().siteCats->getAssignment( jj ) );
                    }
                    
                    /*
                    // [ Added Feb 5 ]
                    // Skip if numsubs1 is below threshold!
                    if (numSubs1 < options.coevol_minNumSubstitutions) {
                        // cout << "           ** SKIPPING SITE " << jj << " IN COEVOL CALCS; NumSubs < threshold **" << endl;
                        
                        if (jj >= ii) {
                            posteriorCoevolution << "0.0" << "\t";
                            if (firstCoevolRound) miCoevolution << "0.0\t";
                        } else { 
                            posteriorCoevolution << "NA\t";
                            if (firstCoevolRound) miCoevolution << "NA\t";
                        }
                        continue;
                    }	
                    */
                    
                    double sum = 0.0;
                    double numSubs2 = subsPerSite.mDataArray[ jj ]; // mTree.calculateDwellTimeOverBranches( jj, siteRate2, mModel, theData, m_vecSiteAssignment );	
                    //for (int k=0; k<numModels; k++) {
                    
                    for (int i=0; i<mModel.GetNumStates(); i++) {
                        sum += marginalDwell->mRows.mDataArray[jj].mDataArray[i];
                    }
                    
                    //cout << "   Marginal dwell at site " << jj << " totals " << sum << endl;
                    
                    // BLOCKING
                    /*
                    if (numSubs2 < options.coevol_minNumSubstitutions) {
                       // cout << "           ** SKIPPING SITE " << jj << " IN COEVOL CALCS; NumSubs < threshold **" << endl;
                        
                        if (jj >= ii) {
                            posteriorCoevolution << "0.0" << "\t";
                            if (firstCoevolRound) miCoevolution << "0.0\t";
                        } else { 
                            posteriorCoevolution << "NA\t";
                            if (firstCoevolRound) miCoevolution << "NA\t";
                        }
                        continue;
                    }
                     */
                    
                    // *******************************************************************************************************
                    // Now joint dwell times for ii x jj...
                    // *******************************************************************************************************
                    treeDepth = &mTree.getRoot()->GetData().tempTreeDepth;
                    //	std::cout << "DIAG OUTPUT: TREE DEPTH IS " << (*treeDepth)[0] << endl;
                    
                    mTree.calculateJointDwellTimeOverBranches( ii,jj, siteRate1, siteRate2, mModel, theData, m_vecSiteAssignment );	
                    for (int i=0; i < mModel.GetNumStates()*mModel.GetNumStates(); i++) { mTree.root()->GetData().observedJFreqs.mRows.mDataArray[0].mDataArray[i] /= theData.getNumSequences(); }
                    
                    // ***** OUTPUT *******
                    if ( options.coevol_outputJointDwell ) {
                        // Normalize the joint counts
                        //    double ssum=0.0;
                        
                        posteriorJointDwell << updateCount << "\t" << ii << "\t" << jj << "\t";
                    }
                    
                    // Set Dwell times to file
                    for (int i=0; i<mModel.GetNumStates()*mModel.GetNumStates(); i++) {
                        if (options.coevol_outputJointDwell) {
                            posteriorJointDwell << jointDwell->mRows.mDataArray[ 0 ].mDataArray[ i ] / (*treeDepth)[0];
                            posteriorJointDwell << "\t";
                        }
                    }
                    
                    // Output observed freqs to file
                    if (options.coevol_outputJointDwell) {
                        for (int i=0; i<mModel.GetNumStates()*mModel.GetNumStates(); i++) {
                            posteriorJointDwell << mTree.root()->GetData().observedJFreqs.mRows.mDataArray[0].mDataArray[i];
                            if (i < mModel.GetNumStates()*mModel.GetNumStates()-1) {
                                posteriorJointDwell << "\t";
                            } else {
                                posteriorJointDwell << endl;
                            }
                        }
                    }
                    
                    
                    // *******************************************************************************************************
                    // Calculate phylogenetic MI statistic...
                    // *******************************************************************************************************
                    double sitePairMI = 0.0;
                    
                    // We can test the effectiveness of scaling by the joint entropy when everything is done properly...
                    // [make this an option, maybe] double jointEntropy = 0.0;
                    
                    sitePairMI = _calculateMI( mModel.GetNumStates(), marginalDwell->mRows.mDataArray[ ii ], marginalDwell->mRows.mDataArray[ jj ], jointDwell->mRows.mDataArray[0], (*treeDepth)[0] );
					
					//	sitePairMI /= jointEntropy;
                    
                    // Get the standard MI too...
                    /*cout << "OBSERVED.  dwell times " << endl;
					for (int iii=0; iii<mModel.GetNumStates(); iii++) {
						for (int jjj=0; jjj<mModel.GetNumStates(); jjj++) {
							cout << "marginal " << iii << ": " << marginalDwell->mRows.mDataArray[ ii ].mDataArray[iii] << "; marginal " << jjj << ": " << marginalDwell->mRows.mDataArray[ jj ].mDataArray[jjj] << "; joint: " << (jointDwell->mRows.mDataArray[0].mDataArray[ (iii * mModel.GetNumStates() + jjj) ] / (*treeDepth)[0]) << endl;
						}
					} */
					
                    double sitePairMI_freq = 0.0;
                    
                    
                    // We can test the effectiveness of scaling by the joint entropy when everything is done properly...
                    // [make this an option, maybe] double jointEntropy = 0.0;
                    if (firstCoevolRound) {
                        sitePairMI_freq = _calculateMI( mModel.GetNumStates(), mTree.root()->GetData().observedFreqs.mRows.mDataArray[ii], mTree.root()->GetData().observedFreqs.mRows.mDataArray[jj], mTree.root()->GetData().observedJFreqs.mRows.mDataArray[0], 1.0 );
                        
                        // if (numSubs1 < options.coevol_minNumSubstitutions) cout << "Site " << ii << ", numsubs " << numSubs1 << "; MI is " << sitePairMI_freq << endl;
                        
                        // if (sitePairMI_freq <= 1e-08) sitePairMI_freq = 0.0;
                    }
                    
                    // if (sitePairMI <= 1e-08) sitePairMI = 0.0;
                    
                    // Output MI to pHist file
                    pHist << ii << "\t" << jj << "\t" << numSubs1 << "\t" << numSubs2 << "\t" << sitePairMI << endl;
                    
                    posteriorCoevolution << setprecision(5) << setw(10) << sitePairMI << "\t";
                    if (firstCoevolRound) miCoevolution << setprecision(5) << setw(10) << sitePairMI_freq << "\t";
                    
                    // *******************************************************************************************************
                    // Posterior predictive null
                    // *******************************************************************************************************
                    
                    if (options.coevol_doPosteriorPredictiveNull == 1) {
                        double fractionExceeding = 0.0;
                        double mean_pp_pMI = 0.0;
  
                        if ( sitePairMI >= options.coevol_thresholdForPP ) {
                            
                            // *******************************************************************************************************
                            // Calculate posterior predictive dist
                            // *******************************************************************************************************
                            
                            double pp_pMI = 0.0;
                            mean_pp_pMI = 0.0;
                            
                            // THIS IS  SLOW; SKIPPING SITES W VERY LOW MI
                            // This could also be parallelized very simply.
   
                            static DoubleMatrix rateMatrixSite1, rateMatrixSite2;
                            static DoubleMatrix scaledMatrixSite1, scaledMatrixSite2;
                        
                            // Site comparison for ii vs jj is done...
                            static DoubleVector mDwell1, mDwell2, jDwell;
                            
                            // Allocating matrices for all marginal and joint dwell times; this could really get big and isn't very smart
                            
                            if (mDwell1.empty()) {
                                mDwell1.resize( mModel.GetNumStates(), 0.0 );
                            } else {
                                mDwell1.fill( 0.0 );
                            }
                            
                            if (mDwell2.empty()) {
                                mDwell2.resize( mModel.GetNumStates(), 0.0 );
                            } else {
                                mDwell2.fill( 0.0 );
                            }
                            
                            if (jDwell.empty()) {
                                jDwell.resize( mModel.GetNumStates() * mModel.GetNumStates(), 0.0 );
                            } else {
                                jDwell.fill( 0.0 );
                            }
                            

                            if (options.reversibleModel == 1) {
                                // For reversible model, use the marginal-dwell time distributions to yield frequencies and rescale rate matrix
                                   double originalScaling, newScalingSite1, newScalingSite2;
                                
                                if (rateMatrixSite1.empty()) {
                                    rateMatrixSite1.resize( mModel.GetNumStates(), mModel.GetNumStates(), 0.0);
                                    rateMatrixSite2.resize( mModel.GetNumStates(), mModel.GetNumStates(), 0.0);
                                }
                                
                                if (scaledMatrixSite1.empty()) {
                                    scaledMatrixSite1.resize( mModel.GetNumStates(), mModel.GetNumStates(), 0.0);
                                    scaledMatrixSite2.resize( mModel.GetNumStates(), mModel.GetNumStates(), 0.0);
                                }
                                
                                // Get the starting scaling factor
                                originalScaling = mModel.getRateSum(0, mTree);
                                
                                double rowsum1, rowsum2;
                                
                                /*
                                // Scale the marginal distribution carefully
                                double freqSum = 0.0;
                                for (int iii=0; iii< mModel.GetNumStates(); iii++) {
                                    if ( marginalDwell->mRows.mDataArray[ jj ].mDataArray[iii] <= 1e-04 ) marginalDwell->mRows.mDataArray[ jj ].mDataArray[iii] = 0.0;
                                    freqSum += marginalDwell->mRows.mDataArray[ jj ].mDataArray[iii];
                                }
                                
                                for (int iii=0; iii< mModel.GetNumStates(); iii++) {
                                    marginalDwell->mRows.mDataArray[ jj ].mDataArray[iii] /= freqSum;
                                }
                                */
                                
                                // Construct a new matrix for sampling from, scaled by empirical frequencies at site (from marginal dwell times)
                                for (int iii=0; iii < mModel.GetNumStates(); iii++) {
                                    rowsum1 = rowsum2 = 0.0;
                                    for (int jjj=0; jjj < mModel.GetNumStates(); jjj++) {
                                        if (jjj > iii) {
                                            // We cannot do this with mixtures yet.  Default to the first rate matrix...
                                            if ( mModel.GetNumModels() > 1 ) { cout << "  *** ERROR - coevolution does not yet work with mixture models!" << endl; exit(-1); }
                                            
                                            // Sets the rate matrix from the upper triangle (NOTE: if not reversible model, this makes it reversible in an arbitrary way!)
                                            
                                            // Site 1:
                                            rateMatrixSite1.mRows.mDataArray[iii].mDataArray[jjj] = mModel.returnAlphaMatrix3D()->mRows.mDataArray[ 0 ].mRows.mDataArray[iii].mDataArray[jjj] *  marginalDwell->mRows.mDataArray[ ii ].mDataArray[jjj]; // * siteRate1;
                                            rateMatrixSite1.mRows.mDataArray[jjj].mDataArray[iii] = mModel.returnAlphaMatrix3D()->mRows.mDataArray[ 0 ].mRows.mDataArray[iii].mDataArray[jjj] *  marginalDwell->mRows.mDataArray[ ii ].mDataArray[iii]; // * siteRate1;
                                        
                                            rateMatrixSite2.mRows.mDataArray[iii].mDataArray[jjj] = mModel.returnAlphaMatrix3D()->mRows.mDataArray[ 0 ].mRows.mDataArray[iii].mDataArray[jjj] * marginalDwell->mRows.mDataArray[ jj ].mDataArray[jjj]; // * siteRate2;
                                            rateMatrixSite2.mRows.mDataArray[jjj].mDataArray[iii] = mModel.returnAlphaMatrix3D()->mRows.mDataArray[ 0 ].mRows.mDataArray[iii].mDataArray[jjj] * marginalDwell->mRows.mDataArray[ jj ].mDataArray[iii]; // * siteRate2;
                                                
                                        }
                                    }
                                }
                                
                                for (int iii=0; iii < mModel.GetNumStates(); iii++) {
                                    rowsum1 = rowsum2 = 0.0;
                                    for (int jjj=0; jjj < mModel.GetNumStates(); jjj++) {
                                        if (jjj != iii) {
                                            // We cannot do this with mixtures yet.  Default to the first rate matrix...
                                            
                                            // Scaled by marginalDwell for site 1 (ii), which has already been normalized
                                            rowsum1 += rateMatrixSite1.mRows.mDataArray[iii].mDataArray[jjj];
                                            
                                            // Scaled by marginalDwell for site 1 (ii), which has already been normalized
                                            rowsum2 += rateMatrixSite2.mRows.mDataArray[iii].mDataArray[jjj];
                                        }
                                    }
                                    
                                    rateMatrixSite1.mRows.mDataArray[iii].mDataArray[iii] = -rowsum1 ;
                                    rateMatrixSite2.mRows.mDataArray[iii].mDataArray[iii] = -rowsum2 ;
                                }
                                
                                // [ check - symmetrize the unscaled rate matrix ] 
                                //mModel.symmetrize( rateMatrixSite1 );
                                //mModel.symmetrize( rateMatrixSite2 );
                                
                                newScalingSite1 = mModel.getRateSum( rateMatrixSite1, marginalDwell->mRows.mDataArray[ ii ], siteRate1 ); // * originalScaling;
                                newScalingSite2 = mModel.getRateSum( rateMatrixSite2, marginalDwell->mRows.mDataArray[ jj ], siteRate2 ); // * originalScaling;
                        //            newScalingSite1 = mModel.getRateSum( rateMatrixSite1, mModel.returnStateFrequencies().mRows.mDataArray[0], siteRate1 ); // * originalScaling;
                        //            newScalingSite2 = mModel.getRateSum( rateMatrixSite2, mModel.returnStateFrequencies().mRows.mDataArray[0], siteRate2 ); // * originalScaling;
                                
                             //   cout << "Numsubs1 is " << numSubs1 << ", and 2 " << numSubs2 << "; expectation is " << (*treeDepth)[0] << endl;
                                
                                // Rescale to produce the expected number of substitutions per site observed
                                newScalingSite1 /= ( numSubs1 * 1.0 ) / (*treeDepth)[0];
                                newScalingSite2 /= ( numSubs2 * 1.0 ) / (*treeDepth)[0];
                                
                                // Ad hoc treatment of sites with zero substitutions: 
                                if ( newScalingSite1 == 0 ) {
                                    // When a site has zero subs, adjust the rates away to stay on the observed state
                                    int theState = -1;
                                    for (int xi=0; xi<mModel.GetNumStates(); xi++) {
                                        //cout << "State " << xi << ", " << marginalDwell->mRows.mDataArray[ ii ].mDataArray[ xi ] << endl;
                                        if ( marginalDwell->mRows.mDataArray[ ii ].mDataArray[ xi ] >= 0.9 ) {
                                            theState = xi;
                                        }
                                    }
                                    if (theState == -1) {
                                        cout << "ERROR - site 1 has no substitutions but dwell times aren't 100% on one state" << endl;
                                        exit(-1);
                                    }
                                    
                                    for (int xi=0; xi<mModel.GetNumStates(); xi++) {
                                        for (int xj=0; xj<mModel.GetNumStates(); xj++) {
                                            if ( xj == xi ) {
                                                rateMatrixSite1.mRows.mDataArray[xi].mDataArray[xj] = -10000;
                                                if (xj == theState) rateMatrixSite1.mRows.mDataArray[xi].mDataArray[xj] = -0.00000001;
                                            } else {
                                                rateMatrixSite1.mRows.mDataArray[xi].mDataArray[xj] = 0.0;
                                            }
                                        }
                                    }
                                } else {
                                    double dummy = mModel.scaleRateMatrix( rateMatrixSite1, newScalingSite1 );
                                }

                                if ( newScalingSite2 == 0 ) {
                                    int theState = -1;
                                    for (int xi=0; xi<mModel.GetNumStates(); xi++) {
                                        if ( marginalDwell->mRows.mDataArray[ jj ].mDataArray[ xi ] >= 0.9 ) {
                                            theState = xi;
                                        }
                                    }
                                    if (theState == -1) {
                                        cout << "ERROR - site 2 has no substitutions but dwell times aren't 100% on one state" << endl;
                                        exit(-1);
                                    }
                                    
                                    for (int xi=0; xi<mModel.GetNumStates(); xi++) {
                                        for (int xj=0; xj<mModel.GetNumStates(); xj++) {
                                            if ( xj == xi ) {
                                                rateMatrixSite2.mRows.mDataArray[xi].mDataArray[xj] = -10000;
                                                if (xj == theState) rateMatrixSite2.mRows.mDataArray[xi].mDataArray[xj] = -0.00000001;
                                            } else {
                                                rateMatrixSite2.mRows.mDataArray[xi].mDataArray[xj] = 0.0;
                                            }
                                        }
                                    }
                                    
                                } else {
                                    double dummy        = mModel.scaleRateMatrix( rateMatrixSite2, newScalingSite2 );
                                }
                                
                                // Now check -> should sum to 1
                                // newScalingSite1 = mModel.getRateSum( rateMatrixSite1, marginalDwell->mRows.mDataArray[ ii ], siteRate1 );
                                // newScalingSite2 = mModel.getRateSum( rateMatrixSite2, marginalDwell->mRows.mDataArray[ jj ], siteRate2 );
                            }

                            int nSubs1, nSubs2;
                            for (int kk=0; kk<options.coevol_numPPsitePairs; kk++) {
                                // Use the dwell times at each site divided by tree length as the root distribution...
 
                                if ( options.reversibleModel == 1 ) {
                                    
                                    nSubs1 = nSubs2 = 0;
									// Use the actual simulated event timings (fully sampled substitution histories)
									 mTree.root()->posteriorPredictive_coevolStatistics_actualTimings(  mTree.root()->GetData().siteCats->getAssignment( ii ),  mTree.root()->GetData().siteCats->getAssignment( jj ), mTree.root(), marginalDwell->mRows.mDataArray[ ii ], marginalDwell->mRows.mDataArray[ jj ], rateMatrixSite1, rateMatrixSite2, m_vecSiteAssignment, mModel, theData, mDwell1, mDwell2, jDwell, nSubs1, nSubs2 );
                                    
                                    // Using global state freqs, but rate rescaled matrix
                                    // mTree.root()->posteriorPredictive_coevolStatistics_actualTimings(  mTree.root()->GetData().siteCats->getAssignment( ii ),  mTree.root()->GetData().siteCats->getAssignment( jj ), mTree.root(), mModel.returnStateFrequencies().mRows.mDataArray[0],mModel.returnStateFrequencies().mRows.mDataArray[0], rateMatrixSite1, rateMatrixSite2, m_vecSiteAssignment, mModel, theData, mDwell1, mDwell2, jDwell, nSubs1, nSubs2 );

                               //     cout << "   PP nsubs1 is " << nSubs1 << ", and " << nSubs2 << endl;
									
									// mTree.root()->posteriorPredictive_coevolStatistics(  mTree.root()->GetData().siteCats->getAssignment( ii ),  mTree.root()->GetData().siteCats->getAssignment( jj ), mTree.root(), marginalDwell->mRows.mDataArray[ ii ], marginalDwell->mRows.mDataArray[ jj ], rateMatrixSite1, rateMatrixSite2, m_vecSiteAssignment, mModel, theData, mDwell1, mDwell2, jDwell );

									// Use actual timings from simulation, with total site homogeneity...
									// mTree.root()->posteriorPredictive_coevolStatistics_actualTimings(  mTree.root()->GetData().siteCats->getAssignment( ii ),  mTree.root()->GetData().siteCats->getAssignment( jj ), mTree.root(), mModel.returnStateFrequencies().mRows.mDataArray[0], mModel.returnStateFrequencies().mRows.mDataArray[0], mModel.returnRateMatrix3D()->mRows.mDataArray[0], mModel.returnRateMatrix3D()->mRows.mDataArray[0], m_vecSiteAssignment, mModel, theData, mDwell1, mDwell2, jDwell, nSubs1, nSubs2 );

									/*
									if ( (ii==2) && (jj==9) ) {
										
										double sumM = 0.0;
										cout << "PP sim, comparison of estimated marginal at site 1 and input:" << endl;
										for (int ll=0; ll < mModel.GetNumStates() * mModel.GetNumStates(); ll++) {
											sumM += jDwell.mDataArray[ ll ];
											cout << ll << ":	" <<  setprecision(10) << jDwell.mDataArray[ ll ] << endl;
										}
									
										cout << "SUM OF JOINTS = " << setprecision(10) << sumM << ", treeL is " << (*treeDepth)[0] << endl;
										exit(-1);
									}
									*/
									//									mTree.root()->posteriorPredictive_coevolStatistics(  mTree.root()->GetData().siteCats->getAssignment( ii ),  mTree.root()->GetData().siteCats->getAssignment( jj ), mTree.root(), marginalDwell->mRows.mDataArray[ ii ], marginalDwell->mRows.mDataArray[ jj ], rateMatrixSite1, rateMatrixSite2, m_vecSiteAssignment, mModel, theData, mDwell1, mDwell2, jDwell );
									
									
									
                                    // Use state frequency parameters...
                                    // mTree.root()->posteriorPredictive_coevolStatistics(  mTree.root()->GetData().siteCats->getAssignment( ii ),  mTree.root()->GetData().siteCats->getAssignment( jj ), mTree.root(), mModel.returnStateFrequencies().mRows.mDataArray[ 0 ], mModel.returnStateFrequencies().mRows.mDataArray[ 0 ], rateMatrixSite1, rateMatrixSite2, m_vecSiteAssignment, mModel, theData, mDwell1, mDwell2, jDwell );
                                    
                                } else {
                                    // For non-reversible model, use the default model and stationary frequencies...
                                    
                                    mTree.root()->posteriorPredictive_coevolStatistics(  mTree.root()->GetData().siteCats->getAssignment( ii ),  mTree.root()->GetData().siteCats->getAssignment( jj ), mTree.root(), m_vecSiteAssignment, mModel, theData, mDwell1, mDwell2, jDwell );
                                }
                                
                                // Normalize the marginals here
								 for (int ll=0; ll < mModel.GetNumStates(); ll++) {
								   mDwell1.mDataArray[ll] /= (*treeDepth)[0];
                                    mDwell2.mDataArray[ll] /= (*treeDepth)[0];
                                }

								// Just do calcs for one site pair...
								pp_pMI = _calculateMI( mModel.GetNumStates(), mDwell1, mDwell2, jDwell, (*treeDepth)[0] );
								
                                // Does using the real dwell times (used to sim with) help?
                                //pp_pMI = _calculateMI( mModel.GetNumStates(), marginalDwell->mRows.mDataArray[ ii ], marginalDwell->mRows.mDataArray[ jj ], jDwell, (*treeDepth)[0] );

								// Makes it extremely conservative; is this what we want?
                                // ECDF is very different.
                                
                                ppHist << ii << "x" << jj << "\t" << nSubs1 << "\t" << nSubs2 << "\t" << pp_pMI << endl;
 
                                mean_pp_pMI += pp_pMI;
                                if (sitePairMI > pp_pMI) fractionExceeding += 1.0;

                            }
							
                            mean_pp_pMI /= (double)options.coevol_numPPsitePairs;
                            fractionExceeding /= (double)options.coevol_numPPsitePairs;
                         }
                        
                 
                        ppCoevolution << fractionExceeding << "\t";
                        ppCoevolutionStat << mean_pp_pMI << "\t";
                    }				
                    
                } else { 
                    posteriorCoevolution << "NA\t";  
                    if (firstCoevolRound) miCoevolution << "NA\t"; 
                    //cout << "           OUTPUTTING " << ii << " x " << jj << " IS NULL" << endl;
                    
                    if (options.coevol_doPosteriorPredictiveNull) {
                        ppCoevolution << "NA\t";
                        ppCoevolutionStat << "NA\t";
                    }
                }
                
            } 
            
            posteriorCoevolution << endl;
            if (firstCoevolRound) miCoevolution << endl;
            if (options.coevol_doPosteriorPredictiveNull) {
                ppCoevolution << endl;
                ppCoevolutionStat << endl;
            }
        } 
        posteriorCoevolution << endl;
        if (options.coevol_doPosteriorPredictiveNull) {
            ppCoevolution << endl;
            ppCoevolutionStat << endl;
        }
        
        //pHist.close();
        //ppHist.close();
        
        
        if (firstCoevolRound == 1) {
            forceCoevolution = 0; 
            firstCoevolRound = 0; // this wasn't set before
            
          // cout << "******* UPDATING A FEW THINGS AFTER SIMULATION ********" << endl;
            MCMC_updateRateMatrix();
           
            Gibbs_updateSiteAssignments_RateAssignments(mModel);
            
            tree_countChanges(mModel);
            
            MCMC_updateRateMatrix();
            
            int last = options.capUniformizationRate;
            options.capUniformizationRate = 0;
            mModel.forceUpdateUniformization();
 
            options.capUniformizationRate = last;
            
            tree_countChanges( mModel, 1 );
            
            Gibbs_updateAllAncestralAndTransientStates_full(  mModel );
            
            tree_countChanges( mModel, 0 );
            
            // Allow site rate resampling now...
            if (options.coevol_DataReplacement_fixSiteAssignments == 0) options.__overrideSiteResamp = 0;
            
            // Removed mu equilibration here...
        }
        
    }
}


double MCMC::_calculateMI( int numStates, DoubleVector& marginal1, DoubleVector& marginal2, DoubleVector &jointDwell, double treeLength) {
    double sitePairMI = 0.0;
    double a, b, c;
    
//	cout << "MI calculation output:" << endl;
	
	int validCount = 0;
    for (int i=0; i<( numStates ); i++) {
        for (int j=0; j<( numStates ); j++) {
            a = jointDwell.mDataArray[ (i*numStates) + j] / treeLength;
            
            // Marginal dwell times are already scaled by the tree depth in the Tree::calculateDwellTimeOverBranches() function...
            b = marginal1.mDataArray[i];
            c = marginal2.mDataArray[j];
            
			double temp = 0.0;
			
			// Instead of skipping terms with bc=0, set a floor and use..
			double product = b * c;
			
			// if (b*c <= 1e-30) product = 1e-30;
			
            if ( (a>0) && ( (b * c) > 0 ) ) {
				double ratio = a / product;
				// cout << "   i,j = " << i << ", " << j << ": a b c " << a << " " << b << " " << c << endl;
                
                //	jointEntropy -= a * log2( a );
                
                // Modified this to use base 20
//                temp = ( a ) * ( ( log( a ) / log( 20 ) ) - ( ( log( product ) / log ( 20 ) ) ) );
                temp = ( a ) * ( log( ratio ) / log( 20 ) );
               
			//	if (temp < 0.0) temp = 0.0;
				
				if ( temp == temp ) {
					validCount++;
					// cout << a << "\t" << b << "\t" << c << "\t" << product << "\t" << temp << endl;
					 
					 sitePairMI += temp;
                }
                //	std::cout << "state " << i << " x " << j << "; " << temp << ": " << a << ", " << b << ", " << c << endl;
            }
        }
    }
    
//	cout << "	VALIDCOUNT IS " << validCount << endl;
	
    // Floor, in case of numerical error
    if (sitePairMI < 0.0) sitePairMI = 0.0;
    
    return sitePairMI;
}


// ************************************************************************//
// MARK: Utility functions
// ************************************************************************//
void MCMC::_util_initializeChain(void) {
    int numModels = mModel.GetNumModels();
    
    // Force a full ancestral state update
    std::cout << "		Performing initial ancestral/transient state sample..." << endl;
    
    for (int ii=0; ii<numModels; ii++) {
        mModel.setUniformization( ii, 1.0 );
    } 
    
    if (mModel.getPredefinedMatrix() != -1) {
        if (options.m_nMethod == 6) mModel.forceUpdateUniformization();
    }
    
    // Need stationary freqs
    if (!options.useEmpiricalStationary) mModel.forceUpdateStationaryFreqs();
    
    // Fully sample ancestral and transient states
    numUpdated += Gibbs_updateAllAncestralAndTransientStates_full( mModel );
    tree_countChanges( mModel );
    
    // Update site-rate assignments if doing ASRV
    if ( options.allowASRV && (options.numCatASRV > 1) ) {
       if ( (updateCount % options.outputFrequency) == 0) std::cout << "		" << updateCount << " [ Resampling site-rate assignments from ASRV distribution ]" << endl; 
       Gibbs_updateSiteAssignments_RateAssignments(mModel);	
    }
    
    // _util_rescaleForBranchLengths();

    // Fully sample ancestral and transient states
    numUpdated += Gibbs_updateAllAncestralAndTransientStates_full( mModel );
    tree_countChanges( mModel );

    // Calculate the initial complete-data logL
    likelihood = CalculateCompleteDataLogLikelihood(mModel);
    mTree.completeDataLnL = likelihood;

    cout << endl;
}


void MCMC::_util_rescaleForBranchLengths(void) {
    // THIS VERSION IS DEPRECATED.
    
    // **************************************************************************************************
    // Rescale rates, mu, and branch-lengths to give branch-lengths their expected meaning
    // **************************************************************************************************
    // NOTE: this function changes the complete-data logL with virtual subs sampled, but leaves the virtual-sub
    //       integrated logL the same.
    
    // if (updateCount < 3000) return;
    
    // Check:
    // cout << "Vsub integrated likleihood before transform: " << setprecision(10) << mTree.root()->_b1u_completeData_recursiveLikelihood( 0, mModel, m_vecSiteAssignment, theData) << endl;

    // Check2: 
    // cout << "Likelihood before: " << setprecision(10) << CalculateCompleteDataLogLikelihood(mModel) << endl;
    
    cout << "   ******************** CALLING _util_rescaleForBranchLengths() ********************" << endl;
   
//    if (options.numGenerations <4000) {
        cout << "   ******************** OVERRIDING _util_rescaleForBranchLengths() ********************" << endl;
        return;
//    }
    
    
    double theRateFactor = mModel.getRateSum( mTree ); // / mModel.GetNumModels();
    
    if (options.allowASRV ) { //&& updateCount > 1000) {
        double scalar = 0.0;
        
        static double *tempCounts;
        static int countAlloc = 0;
        if (countAlloc == 0) {
            tempCounts = new double[options.numCatASRV];
            countAlloc = 1;
        }
        for (int ii=0; ii<options.numCatASRV; ii++) {
            tempCounts[ii] = 0;
        }
        
        for (int jj=0; jj< mTree.root()->GetData().siteCats->siteAssignments.size(); jj++) {
            tempCounts[  mTree.root()->GetData().siteCats->getAssignment(jj) ]++;
        } 

        for (int jj=0; jj< options.numCatASRV; jj++) {
            scalar += mTree.root()->GetData().siteCats->getRate(jj) * ( tempCounts[ jj ] / mTree.root()->GetData().siteCats->siteAssignments.size() ) ;
        }
    
        theRateFactor *= scalar;
        
        // TheRateFactor is now the posterior-weighted average of the scaled rate-matrix sum (scaled by stationary freqs)
        // We can now use this as our normalization factor
        
        cout << "       [ Got scalar " << scalar << "; gives rateFactor " << theRateFactor << " ]" << endl;
    }

 
    // Rescale RAR distribution
    mModel.scaleDiscreteRates( theRateFactor ) ;
    
    // Method 1: Adjust mu, rates, branchlengths
    double mu = mModel.returnUniformizationConstants()->mDataArray[0] / theRateFactor;
    mModel.setUniformization(0, mu );
    mModel.scaleRateMatrices( theRateFactor );
    mTree.root()->TransformBranchLength_B1u( mTree.root(), ( 1.0 / theRateFactor), mu );

    // Set the uniformization constant to within 1% of optimal, for the current generation
    if (options.__currentGen > 1000) {
        mModel.forceUpdateUniformization(0);
    }
    
    // Method 2: Adjust rates, branchlengths (not mu; conserve virtual sub probability)
    // mModel.scaleRateMatrices( theRateFactor );
    // double mu = mModel.returnUniformizationConstants()->mDataArray[0];
    
    // This will alter the likelihood (but perhaps better constrain the rate model)
    // double mu = mModel.returnOptimalUniformization(0); //mModel.returnUniformizationConstants()->mDataArray[0];
    // mModel.setUniformization(0, mu);
    
    // [ Need to re-enable the alternate branch-length scaling in TransformBranchLength_B1u() for this to work ]
    // mTree.root()->TransformBranchLength_B1u( mTree.root(), ( theRateFactor), mu );

    
    
    /*
    // Re-apply everything
    int temp = options.updatebls;
    options.updatebls = 0;
    Gibbs_updateAllAncestralAndTransientStates_full( mModel );
    options.updatebls = temp;
    */
    
    // Check:
    // cout << "Vsub integrated likleihood after transform: " << setprecision(10) << mTree.root()->_b1u_completeData_recursiveLikelihood( 0, mModel, m_vecSiteAssignment, theData) << endl;

    // Check2: 
    // cout << "Likelihood after: " << setprecision(10) << CalculateCompleteDataLogLikelihood(mModel) << endl;
}

void MCMC::_util_optimizeBranchLengths_capNodes(void) {
   
    // THIS FUNCTION IS DEPRECATED
    cout << "       **** WARNING: _util_optimizeBranchLengths_capNodes() function is deprecated ****" << endl;
    return;
    // mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
    
    // [ Feb 5 block - not sure we want to do this]
    numUpdated += Gibbs_updateAllAncestralAndTransientStates_full(  mModel );

    /*
    tree_countChanges( mModel );
    
    // Run the bl updater...
    int rannode;
    double treeLengthDelta = 0.6;
    int goodtimes=0;
    double lastTreeLength = mTree.root()->GetData().treeLength;
    
    // Resample branchlengths until the difference in treelength is small and stays small
    while ( (goodtimes < 3) ) {
        rannode = (int)( num->Random() * options.m_nNumBranches);
        
        // Resample the substitutions
        int version = 1;
        
        tree_updateBranchLengths( rannode, mModel, version, 0.2 );
        
        std::cout << "		" << updateCount++ << " [ Branch-lengths updated. Tree-length: " << setprecision(4) << mTree.root()->GetData().treeLength << "; numNodes: " << ( mTree.root()->CountLeaves() + mTree.root()->CountInternalNodes() ) << "; inflation factor: " << setprecision(3) << (double)((double)mTree.root()->CountBranches()/(double)options.__initialNumBranches) << "X (max. segment-size: " << options.maxBL <<") ]" << endl;        
        
        if (options.updatebls) mTree.root()->NormalizeBranchSegmentLengths(theData, mTree.root() );
        
        // [ Feb 5 block - not sure we want to do this] 
        numUpdated += Gibbs_updateAllAncestralAndTransientStates_full(  mModel );

        // [added this too]
        tree_countChanges( mModel );

        treeLengthDelta = abs( mTree.root()->GetData().treeLength - lastTreeLength );
        cout << "		[ TREELENGTH CHANGE WAS " << treeLengthDelta << " ]" << endl;
        if (treeLengthDelta <= 1.0) { goodtimes++; }
        else { goodtimes = 0; }
        
        lastTreeLength = mTree.root()->GetData().treeLength;
        
    }	
    // Need to re-subdivide the tree...
    mTree.subdivideLongBranches( mTree.root(), options.maxBL, mModel.returnUniformizationConstants(), 1 );
   
    // Subdivide can leave branch-specific changes un-counted
    tree_countChanges( mModel, 1 );
     */
}

void MCMC::_util_initializeRARmodelFromChain(void) {
    if ( ! ((resetAfterConjugatePhase == 0) && (mModel.getRARdistributed()) ) ) return;
    
    std::cout << "		[ Setting initial RAR rate distribution ]" << endl; 
    
    mModel.updateUniformization();
    likelihood = CalculateCompleteDataLogLikelihood(mModel);
    
    static int *oldAssignments;
    static int assignedAlready = 0;
    if (assignedAlready == 0) {
        oldAssignments = new int[ mModel.GetNumModels() * mModel.GetNumStates() * mModel.GetNumStates() ];
        assignedAlready = 1;
    } 
    
    double distParm1, distParm2;
    distParm1 = mModel.getGammaScale(0);
    distParm2 = mModel.getGammaShape(0);
    
    mModel.forceUpdateStationaryFreqs();
    
    mModel.backupRateAssignments(oldAssignments);
    mModel.setDistributionParmsFromCounts( mTree );
    
    mModel.forceUpdateStationaryFreqs();
    
    int tempCalcProbDist = mTree.root()->getCalculateFullProbabilityDistribution();
    mTree.root()->setCalculateFullProbabilityDistribution(0);
    mTree.setprobmatrixalongbranch(mModel, mModel.GetFreqModels());
    mTree.root()->setCalculateFullProbabilityDistribution(tempCalcProbDist);
    
    // Set the uniformization constant to the maximum possible row-sum
    mModel.setMaximumUniformization_RAR();
    
    options.capUniformizationRate = 1;
    resetAfterConjugatePhase = 1;
}

int MCMC::_util_sampleFromCumulative( double *theVector, int length ) {
	// This function assumes:
	//    1) LOG likelihoods are NOT HEATED
	//    2) Assumed theVector contains log-likelihood values
	
	// WARNING: this function will destroy the incoming vector...
	
	// Sampling from:  exp(value - max) / sum_i(exp(val_i - max))
	
	int i, theMax;
	double sum = 0.0, max = -std::numeric_limits<double>::max( );
	
	for (i=0; i<length; i++) {
		// Heat the relative log-likelihoods
		theVector[i] *= options.__globalLikelihoodTemp;
		if (theVector[i] > max) { max = theVector[i]; theMax = i; }
	}
	
	for (i=0; i<length; i++) {
		theVector[i] = exp(theVector[i] - max);
		sum += theVector[i];
	}
	
	for (i=0; i<length; i++) {
		theVector[i] /= sum;
		if (i>0) theVector[i] += theVector[i-1];
	}
	
	double rannum = num->Random();
	
	for (i=0; i<length;i++) {
		if (rannum <= theVector[i]) return i;
	}
    
    return i;
}


int MCMC::_util_sampleFromCumulative_raw( DoubleVector *inVector, int length ) {
    
    static DoubleVector theVector;
    if (theVector.size() != length) theVector.resize( length, 0.0 );
    
	int i;
	double sum = 0.0; // max = -std::numeric_limits<double>::max( );
	
	for (i=0; i<length; i++) {
        if ((*inVector)[i] >= 0) {
            sum += (*inVector)[i];
        } else {
            //sum -= (*inVector)[i];
        }
    }
	
	for (i=0; i<length; i++) {
        if ((*inVector)[i] >= 0) {
            theVector[i] = (*inVector)[i] / sum;
        } else {
            theVector[i] = 0.0;
        }
        
        if (i>0) theVector[i] += theVector[i-1];
	}
    
    /*cout << "Sampling distribution: " << endl;
    for (i=0; i<length; i++) {
        cout << "   State " << i << ": " << theVector[i] << endl;
    }*/
	
	double rannum = num->Random();
	
	for (i=0; i<length;i++) {
		if (rannum <= theVector[i]) return i;
	}
    
    return i;
}

int MCMC::_util_sampleFromCumulative_raw( DoubleVector *inVector, int excludeState, int length ) {
    
    static DoubleVector theVector;
    if (theVector.size() != length) theVector.resize( length, 0.0 );
    
	int i;
	double sum = 0.0; // max = -std::numeric_limits<double>::max( );
	
	for (i=0; i<length; i++) {
        if ( ((*inVector)[i] >= 0) && (i != excludeState) ) {
            sum += (*inVector)[i]; // * (*freqs)[i];
        } else {
            //sum -= (*inVector)[i];
        }
    }
	
	for (i=0; i<length; i++) {
        if ( ( (*inVector)[i] >= 0) && (i != excludeState) ) {
            theVector[i] = (*inVector)[i] / sum; // * (*freqs)[i] / sum;
 //           theVector[i] = (*inVector)[i]  / sum;
        } else {
            theVector[i] = 0.0;
        }
        
        if (i>0) theVector[i] += theVector[i-1];
	}
    
    /*cout << "Sampling distribution: " << endl;
     for (i=0; i<length; i++) {
     cout << "   State " << i << ": " << theVector[i] << endl;
     }*/
	
	double rannum = num->Random();
	
	for (i=0; i<length;i++) {
		if (rannum <= theVector[i]) return i;
	}
    
    return i;
}


int MCMC::_util_sampleChangeFromCumulative_raw( DoubleVector *inVector, int length, int ignoreState ) {
    static DoubleVector theVector; // = *(inVector);
    
    if (theVector.size() != length) theVector.resize( length, 0.0 );
    
	int i, theMax;
	double sum = 0.0, max = -std::numeric_limits<double>::max( );
	
	for (i=0; i<length; i++) {
        if (i != ignoreState) {
            sum += (*inVector)[i];
        }
	}
	
	for (i=0; i<length; i++) {
		theVector[i] = (*inVector)[i] / sum;
        if (i==ignoreState) theVector[i] = 0.0;
        
		if (i>0) theVector[i] += theVector[i-1];
	}
	
	double rannum = num->Random();
	
	for (i=0; i<length;i++) {
		if (rannum <= theVector[i]) return i;
	}
    
    return i;
}

